/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <fb.h>
#include <rabbitsign.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <errno.h>
#include "fbgui.h"
#include "creator.h"
#include "msgbox.h"

static fbsize default_blocksizes[] = { 765, 1023, 2048, 4096, 8192 };
#define NUM_LEVELS G_N_ELEMENTS(default_blocksizes)

typedef struct _fbbuildresult {
  unsigned set : 1;
  unsigned error : 1;
  int title_len;
  int name_len;
  fbsize compressed_size;
  fbsize total_size;
  fbsize ram_size;
} fbbuildresult;

typedef struct _fbbuildwindow {
  fbeditwindow* ew;

  fbbook* fb;

  GtkWidget* dlg;
  GtkWidget* ent_title;
  GtkWidget* ent_name;
  GtkWidget* rb_small;
  GtkWidget* rb_large;
  GtkWidget* rb_custom;
  GtkWidget* fcb_fontfile;
  GtkWidget* cb_hyphenate;
  GtkWidget* cmb_hyphlang;
  GtkWidget* scl_complvl;
  GtkWidget* lbl_before;
  GtkWidget* lbl_after;
  GtkWidget* lbl_total;
  GtkWidget* lbl_pages;
  GtkWidget* lbl_ram;

  int curlevel;
  fbbuildresult results[NUM_LEVELS];

  GThread* calc_thread;
  GMutex* fb_mutex;

  int calc_level;
  gboolean calc_valid;
  fbbuildresult calc_result;

  struct hyphlanginfo* hyphlangs;
  guint nhyphlangs;
} fbbuildwindow;


/**************** List of languages for hyphenation ****************/

struct hyphlanginfo {
  char* code;
  char* lang;
  char* variant;
};

static void check_hyph_lang(fbbuildwindow* bw, const char* code,
			    const char* lang, const char* variant)
{
  guint i;
  char* fpath;
  char* fname;

  for (i = 0; i < bw->nhyphlangs; i++)
    if (!strcmp(bw->hyphlangs[i].lang, lang)
	&& !strcmp(bw->hyphlangs[i].variant, variant))
      return;

  fname = g_strdup_printf("hyph_%s.dic", code);
  if ((fpath = fb_find_common_file(NULL, fname))) {
    fbfree(fpath);

    bw->hyphlangs = g_renew(struct hyphlanginfo, bw->hyphlangs,
			    bw->nhyphlangs + 1);

    bw->hyphlangs[bw->nhyphlangs].code = g_strdup(code);
    bw->hyphlangs[bw->nhyphlangs].lang = g_strdup(lang);
    bw->hyphlangs[bw->nhyphlangs].variant = g_strdup(variant);
    (bw->nhyphlangs)++;
  }

  g_free(fname);
}

static int cmp_hyph_lang(const void* pa, const void* pb)
{
  const struct hyphlanginfo *a, *b;
  int r;

  a = pa;
  b = pb;
  if ((r = strcmp(a->lang, b->lang)))
    return r;
  return strcmp(a->variant, b->variant);
}

static void get_hyph_langs(fbbuildwindow* bw)
{
  FILE* f;
  char buf[256], code[256];
  char *lang, *variant;
  guint i;
  int newlang;

  check_hyph_lang(bw, "en_US", "English", "United States");

  f = fb_open_common_file(NULL, "hyphlang.txt", "rt");
  if (!f)
    return;

  while (fgets(buf, sizeof(buf), f)) {
    if (0 < sscanf(buf, "%s%n", code, &i)) {
      lang = g_strstrip(buf + i);
      if ((variant = strchr(lang, '('))) {
	*variant = 0;
	variant++;
	i = strlen(variant);
	if (variant[i - 1] == ')')
	  variant[i - 1] = 0;
	lang = g_strstrip(lang);
      }
      else {
	variant = lang + strlen(lang);
      }

      check_hyph_lang(bw, code, lang, variant);
    }
  }

  fclose(f);

  if (bw->nhyphlangs == 0)
    return;

  qsort(bw->hyphlangs, bw->nhyphlangs,
	sizeof(struct hyphlanginfo), &cmp_hyph_lang);

  newlang = 1;
  for (i = 0; i < bw->nhyphlangs - 1; i++) {
    if (strcmp(bw->hyphlangs[i].lang, bw->hyphlangs[i + 1].lang)) {
      if (newlang)
	bw->hyphlangs[i].variant[0] = 0;
      newlang = 1;
    }
    else
      newlang = 0;
  }
  if (newlang && i < bw->nhyphlangs)
    bw->hyphlangs[i].variant[0] = 0;

}

static void free_hyph_langs(fbbuildwindow* bw)
{
  guint i;
  for (i = 0; i < bw->nhyphlangs; i++) {
    g_free(bw->hyphlangs[i].code);
    g_free(bw->hyphlangs[i].lang);
    g_free(bw->hyphlangs[i].variant);
  }
  g_free(bw->hyphlangs);
  bw->hyphlangs = NULL;
  bw->nhyphlangs = 0;
}

/**************** Size Calculation ****************/

static void update_display(fbbuildwindow* bw);

static gboolean calc_sizes_finished(gpointer data)
{
  fbbuildwindow* bw = data;

  bw->calc_thread = NULL;
  if (bw->calc_valid)
    bw->results[bw->calc_level] = bw->calc_result;
  update_display(bw);

  return FALSE;
}

static gpointer calc_sizes(gpointer data)
{
  fbbuildwindow* bw = data;
  int level;
  fbbuildresult res;
  fbsize msize;
  fboutbuf* buffer;

  level = bw->calc_level = bw->curlevel;

  g_mutex_lock(bw->fb_mutex);

  if (!bw->fb) {
    g_mutex_unlock(bw->fb_mutex);
    return NULL;
  }

  fb_chunk_clear(bw->fb->title);
  if (bw->ew->booktitle)
    fb_chunk_append_text_utf8(bw->fb, bw->fb->title, bw->ew->booktitle,
			      strlen(bw->ew->booktitle));

  fbfree(bw->fb->fontfile);
  bw->fb->fontfile = NULL;
  if (bw->ew->fontfile) {
    bw->fb->fontfile = fbnew(char, strlen(bw->ew->fontfile) + 1);
    strcpy(bw->fb->fontfile, bw->ew->fontfile);
  }

  fbfree(bw->fb->hyphlang);
  bw->fb->hyphlang = NULL;
  if (bw->ew->hyphlang) {
    bw->fb->hyphlang = fbnew(char, strlen(bw->ew->hyphlang) + 1);
    strcpy(bw->fb->hyphlang, bw->ew->hyphlang);
  }

  bw->fb->hyphenate = bw->ew->hyphenate ? 1 : 0;

  bw->fb->blocksize_limit = default_blocksizes[level];

  buffer = fb_outbuf_new();

  if (!fb_book_write_app(buffer, bw->fb)) {
    res.compressed_size = (bw->fb->packedbitlen + 7) / 8;
    res.total_size = buffer->length;
    res.title_len = bw->fb->title->length;
    res.name_len = bw->fb->appname->length;

    if (bw->fb->blocksize_actual > FB_MAX_STATIC_BLOCK_SIZE) {
      msize = bw->fb->blocksize_actual * 2 + 5 + 8;
      res.ram_size = msize - (msize % 9);
    }
    else
      res.ram_size = 0;

    res.set = 1;
    res.error = 0;
  }
  else {
    res.compressed_size = 0;
    res.total_size = 0;
    res.title_len = 0;
    res.name_len = 0;
    res.ram_size = 0;
    res.set = 1;
    res.error = 1;
  }
  fb_outbuf_free(buffer);
  clear_fb_errors(bw->ew);

  bw->calc_result = res;

  /* wake up the main thread and tell it the good news */
  g_idle_add(calc_sizes_finished, bw);
  g_main_context_wakeup(g_main_context_default());

  g_mutex_unlock(bw->fb_mutex);

  return NULL;
}

static void update_display(fbbuildwindow* bw)
{
  fbbuildresult* res = &bw->results[bw->curlevel];
  char buf[64];

  gtk_dialog_set_response_sensitive(GTK_DIALOG(bw->dlg),
				    GTK_RESPONSE_ACCEPT,
				    (bw->ew->appname && bw->ew->appname[0]
				     && bw->ew->fontfile && bw->ew->fontfile[0])
				    ? TRUE : FALSE);

  if (!res->set) {
    if (!bw->calc_thread) {
      bw->calc_valid = TRUE;
      if ((bw->calc_thread = g_thread_create(calc_sizes, bw, FALSE, NULL))) {
	if (bw->lbl_after) {
	  gtk_label_set_text(GTK_LABEL(bw->lbl_after), "");
	  gtk_label_set_text(GTK_LABEL(bw->lbl_total), "");
	  gtk_label_set_text(GTK_LABEL(bw->lbl_pages), "");
	  gtk_label_set_text(GTK_LABEL(bw->lbl_ram), "");
	}
      }
      else {
	calc_sizes(bw);
      }
    }
    return;
  }

  res->total_size -= res->name_len + res->title_len;
  res->title_len = bw->ew->booktitle ? strlen(bw->ew->booktitle) : 0;
  res->name_len = bw->ew->appname ? strlen(bw->ew->appname) : 0;
  res->total_size += res->name_len + res->title_len;

  if (bw->lbl_after) {
    sprintf(buf, "%ld", (long int) res->compressed_size);
    gtk_label_set_text(GTK_LABEL(bw->lbl_after), buf);
    sprintf(buf, "%ld", (long int) res->total_size);
    gtk_label_set_text(GTK_LABEL(bw->lbl_total), buf);
    sprintf(buf, "%ld", (long int) ((res->total_size + 0x3fff) / 0x4000));
    gtk_label_set_text(GTK_LABEL(bw->lbl_pages), buf);
    sprintf(buf, "%ld", (long int) res->ram_size);
    gtk_label_set_text(GTK_LABEL(bw->lbl_ram), buf);
  }
}

static void invalidate_results(fbbuildwindow* bw)
{
  gsize i;

  for (i = 0; i < NUM_LEVELS; i++)
    bw->results[i].set = 0;
  bw->calc_valid = FALSE;
}

static void queue_rs_error(const RSKey* key G_GNUC_UNUSED,
			   const RSProgram* prgm G_GNUC_UNUSED,
			   const char* msg,
			   void* data)
{
  char** p = data;

  if (!*p)
    *p = g_strdup(msg);
  else
    *p = g_strconcat(*p, "\n", msg, NULL);
}

static RSProgram* sign_application(fboutbuf* buffer, fbbuildwindow* bw)
{
  RSProgram* prgm;
  RSKey* key;
  char* errmsg = NULL;

  rs_set_error_func(&queue_rs_error, &errmsg);

  key = rs_key_new();
  prgm = rs_program_new_with_data(RS_CALC_TI83P, RS_DATA_APP,
				  buffer->data, buffer->length,
				  buffer->length_a);
  if (prgm)
    buffer->data = NULL;

  if (!key || !prgm
      || rs_key_find_for_id(key, rs_program_get_key_id(prgm), 0)
      || rs_repair_ti8x_app(prgm, (RS_FIX_PAGE_COUNT | RS_ZEALOUSLY_PAD_APP))
      || rs_sign_ti8x_app(prgm, key, 0)) {
    messagebox01(bw->dlg, GTK_MESSAGE_ERROR,
		 "Unable to sign application",
		 "An error occurred while signing the application:\n\n%s",
		 errmsg);
    g_free(errmsg);
    rs_set_error_func(NULL, NULL);
    rs_key_free(key);
    rs_program_free(prgm);
    return NULL;
  }

  g_free(errmsg);
  rs_set_error_func(NULL, NULL);
  rs_key_free(key);
  return prgm;
}

static gboolean save_final_output(RSProgram* prgm, fbbuildwindow* bw)
{
  gchar* suggest_name;
  gchar* suggest_dir;
  gchar* fname;
  FILE* appfile;

  fname = g_strconcat(bw->ew->appname, ".8xk", NULL);
  suggest_name = g_filename_from_utf8(fname, -1, NULL, NULL, NULL);
  g_free(fname);

  if (bw->ew->filename)
    suggest_dir = g_path_get_dirname(bw->ew->filename);
  else
    suggest_dir = NULL;

  fname = run_file_chooser("Save Application",
			   GTK_WINDOW(bw->dlg),
			   GTK_FILE_CHOOSER_ACTION_SAVE,
			   NULL,
			   suggest_name,
			   suggest_dir,
			   "Flash Application|*.8xk|");

  g_free(suggest_name);
  g_free(suggest_dir);

  if (!fname)
    return FALSE;

  appfile = g_fopen(fname, "wb");
  if (!appfile) {
    display_io_error(bw->dlg, fname, FALSE, errno);
    g_free(fname);
    return FALSE;
  }

  if (rs_write_ti8x_file(prgm, appfile, 0, 0, 0, 0)) {
    fclose(appfile);
    display_io_error(bw->dlg, fname, FALSE, errno);
    g_free(fname);
    return FALSE;
  }

  if (fclose(appfile)) {
    display_io_error(bw->dlg, fname, FALSE, errno);
    g_free(fname);
    return FALSE;
  }

  g_free(fname);
  return TRUE;
}


static void title_changed(GtkEditable* ed, gpointer data)
{
  fbbuildwindow* bw = data;
  const char* text = gtk_entry_get_text(GTK_ENTRY(ed));

  if (bw->ew->booktitle && bw->ew->booktitle[0]) {
    if (!text[0])
      invalidate_results(bw);
  }
  else if (text[0])
    invalidate_results(bw);

  g_free(bw->ew->booktitle);
  bw->ew->booktitle = g_strdup(text);
  update_display(bw);
}

static void name_changed(GtkEditable* ed, gpointer data)
{
  fbbuildwindow* bw = data;

  g_free(bw->ew->appname);
  bw->ew->appname = g_strdup(gtk_entry_get_text(GTK_ENTRY(ed)));
  update_display(bw);
}

static void level_changed(GtkAdjustment* adj, gpointer data)
{
  fbbuildwindow* bw = data;
  int val;

  val = gtk_adjustment_get_value(adj) + 0.5;
  bw->curlevel = val - 1;
  bw->ew->blocksize = default_blocksizes[bw->curlevel];
  update_display(bw);
}

static void font_type_changed(GtkToggleButton* tb G_GNUC_UNUSED,
			      gpointer data)
{
  fbbuildwindow* bw = data;
  int new, old;

  if (bw->ew->fontfile
      && !strcmp(bw->ew->fontfile, FB_DEFAULT_SMALL_FONT))
    old = 0;
  else if (bw->ew->fontfile
	   && !strcmp(bw->ew->fontfile, FB_DEFAULT_LARGE_FONT))
    old = 1;
  else
    old = 2;

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bw->rb_small)))
    new = 0;
  else if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bw->rb_large)))
    new = 1;
  else
    new = 2;

  if (new == old)
    return;

  g_free(bw->ew->fontfile);
  bw->ew->fontfile = NULL;

  switch (new) {
  case 0:
    gtk_widget_set_sensitive(bw->fcb_fontfile, FALSE);
    bw->ew->fontfile = g_strdup(FB_DEFAULT_SMALL_FONT);
    break;

  case 1:
    gtk_widget_set_sensitive(bw->fcb_fontfile, FALSE);
    bw->ew->fontfile = g_strdup(FB_DEFAULT_LARGE_FONT);
    break;

  case 2:
    gtk_widget_set_sensitive(bw->fcb_fontfile, TRUE);
    bw->ew->fontfile = file_button_get_selection(bw->fcb_fontfile);
    break;
  }

  invalidate_results(bw);
  update_display(bw);
}

static void font_file_changed(GtkWidget* fcb, gpointer data)
{
  fbbuildwindow* bw = data;

  if (!gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bw->rb_custom)))
    return;

  g_free(bw->ew->fontfile);
  bw->ew->fontfile = file_button_get_selection(fcb);

  invalidate_results(bw);
  update_display(bw);
}

static void hyphenate_changed(GtkWidget* w G_GNUC_UNUSED,
			      gpointer data)
{
  fbbuildwindow* bw = data;
  int n;

  n = gtk_combo_box_get_active(GTK_COMBO_BOX(bw->cmb_hyphlang));
  if (n > -1) {
    g_free(bw->ew->hyphlang);
    bw->ew->hyphlang = g_strdup(bw->hyphlangs[n].code);
  }
  else {
    g_free(bw->ew->hyphlang);
    bw->ew->hyphlang = NULL;
  }

  if (gtk_toggle_button_get_active(GTK_TOGGLE_BUTTON(bw->cb_hyphenate))) {
    bw->ew->hyphenate = 1;
    gtk_widget_set_sensitive(bw->cmb_hyphlang, TRUE);
  }
  else {
    bw->ew->hyphenate = 0;
    gtk_widget_set_sensitive(bw->cmb_hyphlang, FALSE);
  }

  invalidate_results(bw);
  update_display(bw);
}

static GtkWidget* new_gnome_style_frame(const gchar* label, GtkWidget* contents)
{
  GtkWidget *frame, *align;
  char *str;

  str = g_strconcat("<b>", label, "</b>", NULL);
  frame = gtk_frame_new(str);
  g_free(str);

  g_object_set(gtk_frame_get_label_widget(GTK_FRAME(frame)),
	       "use-markup", TRUE, NULL);
  gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);

  align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
  gtk_alignment_set_padding(GTK_ALIGNMENT(align), 0, 0, 12, 0);
  gtk_widget_show(align);
  gtk_container_add(GTK_CONTAINER(frame), align);
  gtk_container_add(GTK_CONTAINER(align), contents);
  gtk_widget_show(frame);

  return frame;
}

void run_builder_dialog(fbeditwindow* ew)
{
  GtkWidget *button;
  GtkWidget *vbox1, *vbox2, *hbox, *table, *frame, *label;
  GtkAdjustment *adj;
  fbbuildwindow* bw;
  fbsize before_size;
  fbsize i;
  char buf[64];
  fboutbuf* buffer;
  RSProgram* prgm;
  char *fpath, *suggest_dir, *p;
  const char* deflang;

  bw = g_new0(fbbuildwindow, 1);
  bw->ew = ew;
  bw->fb = create_fb_from_edit(ew, FALSE);

  bw->fb_mutex = g_mutex_new();

  for (i = 0; i < NUM_LEVELS - 1; i++)
    if (ew->blocksize <= default_blocksizes[i])
      break;
  bw->curlevel = i;

  bw->dlg = gtk_dialog_new_with_buttons("Build Application",
					GTK_WINDOW(ew->window),
					GTK_DIALOG_MODAL,
					GTK_STOCK_CLOSE, GTK_RESPONSE_REJECT,
					NULL);
  gtk_window_set_default_size(GTK_WINDOW(bw->dlg), 400, 300);
  bw->ew->dlg_build = bw->dlg;

  button = gtk_button_new_with_mnemonic("_Generate");
  GTK_WIDGET_SET_FLAGS(button, GTK_CAN_DEFAULT);
  gtk_button_set_image(GTK_BUTTON(button),
		       gtk_image_new_from_stock(GTK_STOCK_EXECUTE,
						GTK_ICON_SIZE_BUTTON));
  gtk_dialog_add_action_widget(GTK_DIALOG(bw->dlg), button, GTK_RESPONSE_ACCEPT);
  gtk_widget_show(button);

  gtk_dialog_set_alternative_button_order(GTK_DIALOG(bw->dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_REJECT,
					  -1);

  gtk_dialog_set_default_response(GTK_DIALOG(bw->dlg), GTK_RESPONSE_ACCEPT);

  if (!ew->appname || !ew->appname[0] || !ew->fontfile || !ew->fontfile[0])
    gtk_dialog_set_response_sensitive(GTK_DIALOG(bw->dlg),
				      GTK_RESPONSE_ACCEPT, FALSE);

  vbox1 = gtk_vbox_new(FALSE, 12);
  gtk_container_set_border_width(GTK_CONTAINER(vbox1), 6);
  {
    table = gtk_table_new(2, 3, FALSE);
    gtk_table_set_row_spacings(GTK_TABLE(table), 6);
    gtk_table_set_col_spacings(GTK_TABLE(table), 12);
    {
      bw->ent_title = gtk_entry_new();
      if (ew->booktitle)
	gtk_entry_set_text(GTK_ENTRY(bw->ent_title), ew->booktitle);
      gtk_entry_set_activates_default(GTK_ENTRY(bw->ent_title), TRUE);
      g_signal_connect(bw->ent_title, "changed", G_CALLBACK(title_changed), bw);
      gtk_table_attach_defaults(GTK_TABLE(table), bw->ent_title, 1, 3, 0, 1);
      gtk_widget_show(bw->ent_title);

      label = gtk_label_new_with_mnemonic("Book _Title:");
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), bw->ent_title);
      gtk_misc_set_alignment(GTK_MISC(label), LABEL_X_ALIGN, 0.5);
      gtk_table_attach(GTK_TABLE(table), label, 0, 1, 0, 1,
		       GTK_FILL, GTK_FILL, 0, 0);
      gtk_widget_show(label);

      bw->ent_name = gtk_entry_new();
      if (ew->appname)
	gtk_entry_set_text(GTK_ENTRY(bw->ent_name), ew->appname);
      gtk_entry_set_activates_default(GTK_ENTRY(bw->ent_name), TRUE);
      gtk_entry_set_max_length(GTK_ENTRY(bw->ent_name), 8);
      gtk_entry_set_width_chars(GTK_ENTRY(bw->ent_name), 12);
      g_signal_connect(bw->ent_name, "changed", G_CALLBACK(name_changed), bw);
      gtk_table_attach(GTK_TABLE(table), bw->ent_name, 1, 2, 1, 2,
		       GTK_FILL, GTK_FILL, 0, 0);
      gtk_widget_show(bw->ent_name);

      label = gtk_label_new_with_mnemonic("Application _Name:");
      gtk_label_set_mnemonic_widget(GTK_LABEL(label), bw->ent_name);
      gtk_misc_set_alignment(GTK_MISC(label), LABEL_X_ALIGN, 0.5);
      gtk_table_attach(GTK_TABLE(table), label, 0, 1, 1, 2,
		       GTK_FILL, GTK_FILL, 0, 0);
      gtk_widget_show(label);

      label = gtk_label_new("(8 characters maximum)");
      gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
      gtk_table_attach(GTK_TABLE(table), label, 2, 3, 1, 2,
		       GTK_EXPAND | GTK_FILL, GTK_FILL, 0, 0);
      gtk_widget_show(label);
    }
    gtk_widget_show(table);
    frame = new_gnome_style_frame("Book Information", table);
    gtk_box_pack_start(GTK_BOX(vbox1), frame, FALSE, FALSE, 0);

    vbox2 = gtk_vbox_new(TRUE, 2);
    {
      bw->rb_small = gtk_radio_button_new_with_mnemonic
	(NULL, "_Small (5 pixels, proportional)");
      gtk_box_pack_start(GTK_BOX(vbox2), bw->rb_small, FALSE, TRUE, 0);
      gtk_widget_show(bw->rb_small);

      bw->rb_large = gtk_radio_button_new_with_mnemonic_from_widget
	(GTK_RADIO_BUTTON(bw->rb_small), "_Large (7 pixels, monospace)");
      gtk_box_pack_start(GTK_BOX(vbox2), bw->rb_large, FALSE, TRUE, 0);
      gtk_widget_show(bw->rb_large);

      hbox = gtk_hbox_new(FALSE, 12);
      {
	bw->rb_custom = gtk_radio_button_new_with_mnemonic_from_widget
	  (GTK_RADIO_BUTTON(bw->rb_small), "C_ustom font:");
	gtk_box_pack_start(GTK_BOX(hbox), bw->rb_custom, FALSE, FALSE, 0);
	gtk_widget_show(bw->rb_custom);

	if (ew->fontfile && strcmp(ew->fontfile, FB_DEFAULT_SMALL_FONT)
	    && strcmp(ew->fontfile, FB_DEFAULT_LARGE_FONT))
	  fpath = fb_find_common_file(bw->fb, ew->fontfile);
	else
	  fpath = NULL;

	if (ew->filename)
	  suggest_dir = g_path_get_dirname(ew->filename);
	else
	  suggest_dir = NULL;

	bw->fcb_fontfile = make_file_button("Choose font file",
					    fpath, suggest_dir,
					    "Flashbook Font Files|*.fbf|");

	fbfree(fpath);
	g_free(suggest_dir);

	gtk_box_pack_start(GTK_BOX(hbox), bw->fcb_fontfile, TRUE, TRUE, 0);
	gtk_widget_show(bw->fcb_fontfile);
      }
      gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, TRUE, 0);
      gtk_widget_show(hbox);

      if (ew->fontfile && !strcmp(ew->fontfile, FB_DEFAULT_SMALL_FONT)) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bw->rb_small), TRUE);
	gtk_widget_set_sensitive(bw->fcb_fontfile, FALSE);
      }
      else if (ew->fontfile && !strcmp(ew->fontfile, FB_DEFAULT_LARGE_FONT)) {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bw->rb_large), TRUE);
	gtk_widget_set_sensitive(bw->fcb_fontfile, FALSE);
      }
      else {
	gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bw->rb_custom), TRUE);
      }

      g_signal_connect(bw->rb_small, "toggled",
		       G_CALLBACK(font_type_changed), bw);
      g_signal_connect(bw->rb_large, "toggled",
		       G_CALLBACK(font_type_changed), bw);
      g_signal_connect(bw->rb_custom, "toggled",
		       G_CALLBACK(font_type_changed), bw);
      g_signal_connect(bw->fcb_fontfile, "selection-changed",
		       G_CALLBACK(font_file_changed), bw);
    }
    gtk_widget_show(vbox2);
    frame = new_gnome_style_frame("Font", vbox2);
    gtk_box_pack_start(GTK_BOX(vbox1), frame, FALSE, FALSE, 0);

    get_hyph_langs(bw);

    vbox2 = gtk_vbox_new(FALSE, 3);
    {
      bw->cb_hyphenate = gtk_check_button_new_with_mnemonic
	("Automatically _hyphenate text");
      gtk_box_pack_start(GTK_BOX(vbox2), bw->cb_hyphenate, FALSE, TRUE, 0);
      gtk_widget_show(bw->cb_hyphenate);

      hbox = gtk_hbox_new(FALSE, 0);
      {
	label = gtk_label_new_with_mnemonic("L_anguage:");
	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 12);
	gtk_widget_show(label);

	bw->cmb_hyphlang = gtk_combo_box_new_text();
	gtk_label_set_mnemonic_widget(GTK_LABEL(label), bw->cmb_hyphlang);

	if (ew->hyphlang)
	  deflang = ew->hyphlang;
	else
	  deflang = "en_US";

	for (i = 0; i < bw->nhyphlangs; i++) {
	  if (bw->hyphlangs[i].variant[0]
	      && ((i != 0 && !strcmp(bw->hyphlangs[i].lang,
				     bw->hyphlangs[i - 1].lang))
		  || (i != bw->nhyphlangs - 1
		      && !strcmp(bw->hyphlangs[i].lang,
				 bw->hyphlangs[i + 1].lang)))) {
	    p = g_strdup_printf("%s (%s)", bw->hyphlangs[i].lang,
				bw->hyphlangs[i].variant);
	    gtk_combo_box_append_text(GTK_COMBO_BOX(bw->cmb_hyphlang), p);
	    g_free(p);
	  }
	  else {
	    gtk_combo_box_append_text(GTK_COMBO_BOX(bw->cmb_hyphlang),
				      bw->hyphlangs[i].lang);
	  }

	  if (i == 0 || !strcmp(deflang, bw->hyphlangs[i].code))
	    gtk_combo_box_set_active(GTK_COMBO_BOX(bw->cmb_hyphlang), i);
	}

	gtk_box_pack_start(GTK_BOX(hbox), bw->cmb_hyphlang, FALSE, FALSE, 0);
	gtk_widget_show(bw->cmb_hyphlang);

	if (ew->hyphenate)
	  gtk_toggle_button_set_active(GTK_TOGGLE_BUTTON(bw->cb_hyphenate), TRUE);
	else
	  gtk_widget_set_sensitive(bw->cmb_hyphlang, FALSE);

	g_signal_connect(bw->cb_hyphenate, "toggled",
			 G_CALLBACK(hyphenate_changed), bw);
	g_signal_connect(bw->cmb_hyphlang, "changed",
			 G_CALLBACK(hyphenate_changed), bw);

      }
      gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, TRUE, 0);
      gtk_widget_show(hbox);
    }
    gtk_widget_show(vbox2);
    frame = new_gnome_style_frame("Hyphenation", vbox2);
    gtk_box_pack_start(GTK_BOX(vbox1), frame, FALSE, FALSE, 0);
    if (bw->nhyphlangs == 0)
      gtk_widget_hide(frame);

    vbox2 = gtk_vbox_new(FALSE, 6);
    {
      hbox = gtk_hbox_new(FALSE, 12);
      {
	bw->scl_complvl = gtk_hscale_new_with_range(1, NUM_LEVELS, 1);
	gtk_scale_set_digits(GTK_SCALE(bw->scl_complvl), 0);
	gtk_widget_show(bw->scl_complvl);

	adj = gtk_range_get_adjustment(GTK_RANGE(bw->scl_complvl));
	gtk_adjustment_set_value(adj, bw->curlevel + 1);
	g_signal_connect(adj, "value-changed", G_CALLBACK(level_changed), bw);

	label = gtk_label_new_with_mnemonic("C_ompression Level:");
	gtk_label_set_mnemonic_widget(GTK_LABEL(label), bw->scl_complvl);
	gtk_widget_show(label);

	gtk_box_pack_start(GTK_BOX(hbox), label, FALSE, FALSE, 0);
	gtk_box_pack_start(GTK_BOX(hbox), bw->scl_complvl, TRUE, TRUE, 0);
      }
      gtk_box_pack_start(GTK_BOX(vbox2), hbox, FALSE, FALSE, 0);
      gtk_widget_show(hbox);

      table = gtk_table_new(5, 2, FALSE);
      {
	label = gtk_label_new("Size of text before compression:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 0, 1);
	gtk_widget_show(label);

	before_size = bw->fb->body->nutf8chars;
	for (i = 0; i < bw->fb->nsections; i++)
	  before_size += bw->fb->sections[i].title->nutf8chars;

	sprintf(buf, "%lu", (unsigned long int) before_size);
	bw->lbl_before = gtk_label_new(buf);
	gtk_misc_set_alignment(GTK_MISC(bw->lbl_before), 1.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), bw->lbl_before, 1, 2, 0, 1);
	gtk_widget_show(bw->lbl_before);

	label = gtk_label_new("Size of text after compression:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 1, 2);
	gtk_widget_show(label);

	bw->lbl_after = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(bw->lbl_after), 1.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), bw->lbl_after, 1, 2, 1, 2);
	gtk_widget_show(bw->lbl_after);

	label = gtk_label_new("Total size of application:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 2, 3);
	gtk_widget_show(label);

	bw->lbl_total = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(bw->lbl_total), 1.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), bw->lbl_total, 1, 2, 2, 3);
	gtk_widget_show(bw->lbl_total);

	label = gtk_label_new("Number of application slots used:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 3, 4);
	gtk_widget_show(label);

	bw->lbl_pages = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(bw->lbl_pages), 1.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), bw->lbl_pages, 1, 2, 3, 4);
	gtk_widget_show(bw->lbl_pages);

	label = gtk_label_new("RAM required to decompress:");
	gtk_misc_set_alignment(GTK_MISC(label), 0.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), label, 0, 1, 4, 5);
	gtk_widget_show(label);

	bw->lbl_ram = gtk_label_new("");
	gtk_misc_set_alignment(GTK_MISC(bw->lbl_ram), 1.0, 0.5);
	gtk_table_attach_defaults(GTK_TABLE(table), bw->lbl_ram, 1, 2, 4, 5);
	gtk_widget_show(bw->lbl_ram);
      }
      gtk_box_pack_start(GTK_BOX(vbox2), table, FALSE, FALSE, 0);
      gtk_widget_show(table);
    }
    gtk_widget_show(vbox2);
    frame = new_gnome_style_frame("Compression", vbox2);
    gtk_box_pack_start(GTK_BOX(vbox1), frame, FALSE, FALSE, 0);
  }
  gtk_box_pack_start(GTK_BOX(GTK_DIALOG(bw->dlg)->vbox), vbox1, TRUE, TRUE, 0);
  gtk_widget_show(vbox1);
  update_display(bw);

  while (1) {
    if (gtk_dialog_run(GTK_DIALOG(bw->dlg)) != GTK_RESPONSE_ACCEPT)
      break;

    g_mutex_lock(bw->fb_mutex);

    fb_chunk_clear(bw->fb->title);
    if (bw->ew->booktitle)
      fb_chunk_append_text_utf8(bw->fb, bw->fb->title, bw->ew->booktitle,
				strlen(bw->ew->booktitle));

    fb_chunk_clear(bw->fb->appname);
    if (bw->ew->appname)
      fb_chunk_append_text_utf8(bw->fb, bw->fb->appname, bw->ew->appname,
				strlen(bw->ew->appname));

    fbfree(bw->fb->fontfile);
    bw->fb->fontfile = NULL;
    if (bw->ew->fontfile) {
      bw->fb->fontfile = fbnew(char, strlen(bw->ew->fontfile) + 1);
      strcpy(bw->fb->fontfile, bw->ew->fontfile);
    }

    fbfree(bw->fb->hyphlang);
    bw->fb->hyphlang = NULL;
    if (bw->ew->hyphlang) {
      bw->fb->hyphlang = fbnew(char, strlen(bw->ew->hyphlang) + 1);
      strcpy(bw->fb->hyphlang, bw->ew->hyphlang);
    }

    bw->fb->hyphenate = bw->ew->hyphenate ? 1 : 0;

    bw->fb->blocksize_limit = default_blocksizes[bw->curlevel];

    buffer = fb_outbuf_new();

    if (display_fb_error(bw->ew, fb_book_write_app(buffer, bw->fb))) {
      fb_outbuf_free(buffer);
      g_mutex_unlock(bw->fb_mutex);
      continue;
    }

    g_mutex_unlock(bw->fb_mutex);

    prgm = sign_application(buffer, bw);
    fb_outbuf_free(buffer);
    if (!prgm)
      continue;

    if (save_final_output(prgm, bw)) {
      rs_program_free(prgm);
      break;
    }

    rs_program_free(prgm);
  }

  g_mutex_lock(bw->fb_mutex);
  fb_book_free(bw->fb);
  bw->fb = NULL;
  g_mutex_unlock(bw->fb_mutex);

  g_mutex_free(bw->fb_mutex);

  g_idle_remove_by_data(bw);

  free_hyph_langs(bw);

  bw->ew->dlg_build = NULL;
  gtk_widget_destroy(bw->dlg);
  g_free(bw);
}
