/*
 * Flashbook GUI
 * Copyright (c) 2008-2010 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <fb.h>
#include <rabbitsign.h>
#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <errno.h>
#include "fbpreview.h"
#include "fbtextbuffer.h"
#include "fbgui.h"
#include "creator.h"
#include "msgbox.h"

#if GLIB_CHECK_VERSION(2, 14, 0)
# define USE_REGEX
#endif

#ifdef HAVE_LIBIGE_MAC_INTEGRATION
# include <ige-mac-menu.h>
# include <ige-mac-bundle.h>
# define USE_MDI
# define USE_MAC_MENU
#endif

#ifdef USE_UNIFIED_MDI
# include "fontedit.h"
#endif

#ifdef TEST_MAC_MDI_MODE
# define ige_mac_menu_set_menu_bar(mb) do {			\
    GtkWidget* win = gtk_window_new(GTK_WINDOW_TOPLEVEL);	\
    gtk_container_add(GTK_CONTAINER(win), GTK_WIDGET(mb));	\
    gtk_widget_show(win);					\
  } while (0)

# define IgeMacMenuGroup int
# define ige_mac_menu_set_quit_menu_item(mi)
# define ige_mac_menu_add_app_menu_group() (0)
# define ige_mac_menu_add_app_menu_item(a, b, c)
# define ige_mac_menu_connect_window_key_handler(w)

# define USE_MDI
# define USE_MAC_MENU
#endif


/**************** Error reporting ****************/

static void queueerror(fbbook* fb, int errcode G_GNUC_UNUSED,
		       const char* message)
{
  char* p;
  fbeditwindow* ew = fb->userdata;

  if (ew->nerrors > 10)
    return;
  else if (ew->nerrors == 10)
    message = "(more errors follow...)";

  if (ew->error_log) {
    p = g_strconcat(ew->error_log, "\n", message, NULL);
    g_free(ew->error_log);
  }
  else
    p = g_strdup(message);
  ew->error_log = p;
  ew->nerrors++;
}

static void ignoreerror(fbbook* fb G_GNUC_UNUSED,
			int errcode G_GNUC_UNUSED,
			const char* message G_GNUC_UNUSED)
{
}

static fbbook* fb_book_new_for_edit(fbeditwindow* ew,
				    gboolean ignore_err)
{
  fbbook* fb = fb_book_new();
  if (ignore_err)
    fb->errfunc = &ignoreerror;
  else
    fb->errfunc = &queueerror;
  fb->userdata = ew;
  return fb;
}

void clear_fb_errors(fbeditwindow* ew)
{
  g_free(ew->error_log);
  ew->error_log = NULL;
  ew->nerrors = 0;
}

static const struct {
  int code;
  const char* title;
  const char* detail;
} error_info[] =
  {{ ERR_MISSING_FILE,          "A required file is missing",
     "One or more required files cannot be found.  This probably"
     " indicates that Flashbook has not been installed correctly.\n\n%s" },

   { ERR_FILE_IO,               "File I/O error",
     "An error occurred while trying to write the output file.\n\n%s" },

   { ERR_XML_PARSER,            "XML parsing error",
     "The given input file is not a valid Flashbook file.\n\n%s" },

   { ERR_UNKNOWN_XML_ROOT,      "XML parsing error",
     "The given input file is not a valid Flashbook file.\n\n%s" },

   { ERR_INVALID_NUMERIC_VALUE, "XML parsing error",
     "The given input file is not a valid Flashbook file.\n\n%s" },
   
   { ERR_NO_APPNAME,            "No application name specified",
     "You must specify a name for the application.  This is the name"
     " which will appear in the APPLICATIONS menu; it must be 8"
     " or fewer characters in length and use only characters from"
     " the calculator's built-in character set." },

   { ERR_TOO_MANY_CHARS,        "Book contains too many different characters",
     "The book contains more than 255 distinct characters (including"
     " space and newline characters.)  You may only use 255 distinct"
     " characters in any single book." },

   { ERR_BITMAPS_NOT_FOUND,     "Characters missing from font",
     "One or more of the characters used in the book is not in the"
     " given font.  You must either add these characters to the"
     " font or remove them from the text."
     "\n\n%s" },

   { ERR_METADATA_TOO_LARGE,    "Too much metadata",
     "The combined size of the reader code and metadata for the book"
     " is too large (greater than 16k.)  This most likely means that"
     " you tried to convert a text file that is way too large to fit"
     " on a calculator.  If you do intend to create a very large"
     " eBook, you probably need to increase the compression level." },

   { ERR_ASSEMBLER,             "Assembler error",
     "Unable to assemble the reader core:\n\n%s" },
   
   { ERR_HYPHEN_INIT,           "Unable to load hyphenation file",
     "The hyphenation pattern file for the selected language is not"
     " present or cannot be parsed." },
   
   { ERR_NO_HEADER,             "Unable to unpack",
     "The given Flash Application does not appear to contain any"
     " Flashbook data." },

   { ERR_UNKNOWN_HEADER,        "Unable to unpack",
     "The given Flash Application uses features that are not"
     " supported by this version of Flashbook.\n\n%s" },

   { ERR_INVALID_HEADER,        "Unable to unpack",
     "The given Flash Application appears to have been corrupted"
     " or be missing data." },

   { ERR_INVALID_HUFFMAN_TREE,  "Unable to unpack",
     "The given Flash Application appears to have been corrupted"
     " or be missing data." },

   { ERR_INVALID_POINTER,       "Unable to unpack",
     "The given Flash Application appears to have been corrupted"
     " or be missing data." },

   { ERR_RUNAWAY_DATA,          "Unable to unpack",
     "The given Flash Application appears to have been corrupted"
     " or be missing data." }};

static const char* get_fb_error_message(int errnum)
{
  gsize i;
  for (i = 0; i < G_N_ELEMENTS(error_info); i++)
    if (error_info[i].code == errnum)
      return error_info[i].title;
  return NULL;
}

static const char* get_fb_error_description(int errnum)
{
  gsize i;
  for (i = 0; i < G_N_ELEMENTS(error_info); i++)
    if (error_info[i].code == errnum)
      return error_info[i].detail;
  return NULL;
}

int display_fb_error(fbeditwindow* ew, int errnum)
{
  const char *msg, *desc;

  if (errnum) {
    msg = get_fb_error_message(errnum);
    desc = get_fb_error_description(errnum);

    if (msg && desc) {
      messagebox01(ew->window, GTK_MESSAGE_ERROR, msg, desc, ew->error_log);
    }
    else {
      messagebox02(ew->window, GTK_MESSAGE_ERROR,
		   "Internal error",
		   "An error of type %d occurred.  "
		   "This is a bug; please report it.  "
		   "Message log:\n\n%s",
		   errnum, ew->error_log);
    }
  }
  clear_fb_errors(ew);
  return errnum;
}


/**************** Create fbbook from edit ****************/

fbbook* create_fb_from_edit(fbeditwindow* ew, gboolean ignore_err)
{
  fbbook* fb = fb_book_new_for_edit(ew, ignore_err);
  GtkTextIter lstart, lend;
  char *s, *t;
  fbsection* sect;

  if (ew->appname && ew->appname[0])
    fb_chunk_append_text_utf8(fb, fb->appname, ew->appname, strlen(ew->appname));
  else
    fb_chunk_append_text_utf8(fb, fb->appname, "Untitled", 8);

  if (ew->booktitle && ew->booktitle[0])
    fb_chunk_append_text_utf8(fb, fb->title, ew->booktitle, strlen(ew->booktitle));

  if (ew->fontfile && ew->fontfile[0]) {
    fbfree(fb->fontfile);
    fb->fontfile = fbnew(char, strlen(ew->fontfile) + 1);
    strcpy(fb->fontfile, ew->fontfile);
  }

  if (ew->hyphlang) {
    fbfree(fb->hyphlang);
    fb->hyphlang = fbnew(char, strlen(ew->hyphlang) + 1);
    strcpy(fb->hyphlang, ew->hyphlang);
  }

  fb->hyphenate = ew->hyphenate ? 1 : 0;

  fb->blocksize_limit = ew->blocksize;

  if (ew->filename)
    fb->searchdir = fb_get_dirname(ew->filename);

  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &lstart, 0);
  lend = lstart;
  while (!gtk_text_iter_is_end(&lstart)) {
    gtk_text_iter_forward_line(&lend);

    s = gtk_text_iter_get_slice(&lstart, &lend);

    if (gtk_text_iter_has_tag(&lstart, ew->tag_section)) {
      t = s + strlen(s);
      while (t != s && (t[-1] == '\n' || t[-1] == '\r'))
	t--;
      *t = 0;

      sect = fb_book_append_section(fb);
      fb_chunk_append_text_utf8(fb, sect->title, s, strlen(s));
    }
    else {
      fb_chunk_append_text_utf8(fb, fb->body, s, strlen(s));
    }

    g_free(s);

    lstart = lend;
  }

  return fb;
}


/**************** Title bar and status bar ****************/

static void update_titlebar(fbeditwindow* ew)
{
  set_window_title(GTK_WINDOW(ew->window), ew->displayname,
		   ew->filename, gtk_text_buffer_get_modified(ew->textbuf));
}

static void update_statusbar(fbeditwindow* ew)
{
  char buf[32];
  char* p;
  gdouble nk = gtk_text_buffer_get_char_count(ew->textbuf) / 1024.0;

  gtk_statusbar_remove(GTK_STATUSBAR(ew->status),
		       ew->scid_text_info,
		       ew->smid_text_info);

  if (ew->previewerror)
    p = g_strdup_printf("Error: %s", get_fb_error_message(ew->previewerror));
  else if (nk < 0.1)
    p = g_strdup("");
  else {
    if (nk < 100.0)
      g_snprintf(buf, sizeof(buf), "%.1fk", nk);
    else if (nk > 1024.0)
      g_snprintf(buf, sizeof(buf), "%.1fM", nk / 1024.0);
    else
      g_snprintf(buf, sizeof(buf), "%.0fk", nk);
    p = g_strdup_printf("%s bytes uncompressed", buf);

    /* Yes, of course that's not technically true in reference to the
       text buffer, or to the size of the source file, but it is the
       size of the text once it's been converted into book encoding
       (one byte per character), which is what we really want to know
       here. */
  }
  ew->smid_text_info = gtk_statusbar_push(GTK_STATUSBAR(ew->status),
					  ew->scid_text_info, p);
  g_free(p);
}


/**************** Opening/saving Flashbook files ****************/

#define READ_DATA(bbb, nnn) do {				\
    if ((nnn) > fread((bbb), 1, (nnn), f))			\
      return FALSE;						\
    fsize -= (nnn);						\
  } while (0)

#define READ_WORD(vvv) do {		\
    READ_DATA(ibuf, 2);			\
    (vvv) = (ibuf[0] | (ibuf[1] << 8));	\
  } while (0)

#define SKIP_DATA(nnn) fseek(f, (nnn), SEEK_CUR)

static gboolean import_ti_var_file(fbbook* fb, FILE* f)
{
  char buf[45], vname[9];
  guchar ibuf[2];
  unsigned int fsize = 0;
  unsigned int hsize, vsize, tsize;
  gboolean empty = TRUE;
  fbsection* sect;
  unsigned int i, n;

  fb->encoding_id = ENCODING_TI83P_LARGE;

  if (45 > fread(buf, 1, 45, f))
    return FALSE;

  READ_WORD(fsize);

  while (fsize > 0) {

    READ_WORD(hsize);		/* size of variable header */

    if (hsize < 11 || hsize > 32)
      return FALSE;

    READ_DATA(buf, hsize);	/* read variable header*/
    READ_WORD(vsize);		/* get size of variable data */

    strncpy(vname, buf + 3, 8);
    vname[8] = 0;

    if ((buf[2] & 0x1f) == 0x15 && vsize >= 26) {
      /* Notefolio file */

      READ_DATA(buf, 26);	/* read variable length + Notefolio header */

      if (strncmp(buf + 2, "\xf3\x47\xbf\xaf", 4)) {
	SKIP_DATA(vsize - 26);
      }
      else {
	if (empty) {
	  /* set app name to name of this file */
	  for (i = 0; vname[i]; i++) {
	    n = g_unichar_to_utf8(fb_book_get_system_char(fb, vname[i]), buf);
	    fb_chunk_append_text_utf8(fb, fb->appname, buf, n);
	  }
	}
	if (!empty || vsize - 26 < fsize) {
	  /* insert a section header */
	  sect = fb_book_append_section(fb);
	  for (i = 0; vname[i]; i++) {
	    n = g_unichar_to_utf8(fb_book_get_system_char(fb, vname[i]), buf);
	    fb_chunk_append_text_utf8(fb, sect->title, buf, n);
	  }
	}

	/* Get length of text from Notefolio header */
	tsize = (buf[18] | (buf[19] << 8)) - 26;
	if (tsize > vsize - 26)
	  tsize = vsize - 26;

	for (i = 0; i < tsize; i++) {
	  READ_DATA(ibuf, 1);

	  /* Note: the new characters that were introduced with OS
	     1.15 and 1.16 are not included in the base character map
	     (so that Flashbook eBooks can be used on older OSes.)  We
	     do, however, want to interpret them here. */

	  if (ibuf[0] == 0xf1)	/* special byte value used by
				   Notefolio to mark soft line
				   breaks */
	    fb_chunk_append_text_utf8(fb, fb->body, " ", 1);
	  else if (ibuf[0] == 0xf2) /* 1.15+ */
	    fb_chunk_append_text_utf8(fb, fb->body, "$", 1);  
	  else if (ibuf[0] == 0xf3) /* 1.16+: U+2B06
				       UPWARDS BLACK ARROW */
	    fb_chunk_append_text_utf8(fb, fb->body, "\342\254\206", 3);
	  else if (ibuf[0] == 0xf4) /* 1.16+: U+00DF
				       LATIN SMALL LETTER SHARP S */
	    fb_chunk_append_text_utf8(fb, fb->body, "\303\237", 2);
	  else {
	    n = g_unichar_to_utf8(fb_book_get_base_char(fb, ibuf[0]), buf);
	    fb_chunk_append_text_utf8(fb, fb->body, buf, n);
	  }
	}
	SKIP_DATA(vsize - tsize - 26);
	fb_chunk_append_text_utf8(fb, fb->body, "\n", 1);
	empty = FALSE;
      }
    }
    else if ((buf[2] & 0x1f) == 0x06 && vsize >= 14) {
      /* QWERTY */

      READ_DATA(buf, 14);
      if (buf[13] != 0 || strcmp(buf + 2, "QWERTY Text")) {
	SKIP_DATA(vsize - 14);
      }
      else {
	if (empty) {
	  /* set app name to name of this file */
	  for (i = 0; vname[i]; i++) {
	    n = g_unichar_to_utf8(fb_book_get_system_char(fb, vname[i]), buf);
	    fb_chunk_append_text_utf8(fb, fb->appname, buf, n);
	  }
	}
	if (!empty || vsize - 14 < fsize) {
	  /* insert a section header */
	  sect = fb_book_append_section(fb);
	  for (i = 0; vname[i]; i++) {
	    n = g_unichar_to_utf8(fb_book_get_system_char(fb, vname[i]), buf);
	    fb_chunk_append_text_utf8(fb, sect->title, buf, n);
	  }
	}

	for (i = 0; i < vsize - 14; i++) {
	  READ_DATA(ibuf, 1);
	  if (ibuf[0] & 0x80)
	    fb_chunk_append_text_utf8(fb, fb->body, (char*) ibuf, 1);
	  else if (ibuf[0] != '\r') {
	    /* QWERTY only supports ASCII; guess that non-ASCII
	       characters are Latin-1 */
	    n = g_unichar_to_utf8(ibuf[0], buf);
	    fb_chunk_append_text_utf8(fb, fb->body, buf, n);
	  }
	}
	fb_chunk_append_text_utf8(fb, fb->body, "\n", 1);
	empty = FALSE;
      }
    }
    else {
      SKIP_DATA(vsize);
    }
  }

  return !empty;
}

static gboolean open_file(fbeditwindow* ew, const gchar* fname, gboolean replace)
{
  FILE* f;
  char sig[8];
  fbbook* fb;
  RSProgram* prgm;
  GtkTextIter start, end;
  fbsize i, j;
  const char *p, *q;
  char* name;
  char* cwd;

  f = g_fopen(fname, "rb");
  if (!f) {
    display_io_error(ew->window, fname, TRUE, errno);
    return FALSE;
  }

  if (fread(sig, 1, 8, f) < 8)
    sig[0] = 0;

  if (!strncmp(sig, "**TIFL**", 8)) {
    fseek(f, 0L, SEEK_SET);
    prgm = rs_program_new();
    if (rs_read_program_file(prgm, f, fname, 0)) {
      name = filename_to_display(fname);
      messagebox01(ew->window, GTK_MESSAGE_ERROR,
		   "Invalid application file",
		   "The file %s is not a valid TI Flash application.", name);
      g_free(name);
      rs_program_free(prgm);
      fclose(f);
      return FALSE;
    }
    fclose(f);
    fb = fb_book_new_for_edit(ew, FALSE);
    if (display_fb_error(ew, fb_book_read_app(fb, prgm->data,
					      prgm->length, 0x4000))) {
      fb_book_free(fb);
      rs_program_free(prgm);
      return FALSE;
    }
    rs_program_free(prgm);

    fb_chunk_convert_to_utf8(fb, fb->appname, 1);
    fb_chunk_convert_to_utf8(fb, fb->title, 0);
    fb_chunk_convert_to_utf8(fb, fb->body, 0);
    for (i = 0; i < fb->nsections; i++)
      fb_chunk_convert_to_utf8(fb, fb->sections[i].title, 0);

    if (replace) {
      g_free(ew->filename);
      ew->filename = NULL;
      g_free(ew->displayname);
      if (fb->appname->utf8size)
	ew->displayname = g_strndup((char*) fb->appname->utf8,
				    fb->appname->utf8size);
      else
	ew->displayname = g_strdup("Untitled");
    }
  }
  else if (!strncmp(sig, "**TI83F*", 8) || !strncmp(sig, "**TI83**", 8)) {
    fb = fb_book_new_for_edit(ew, FALSE);
    if (!import_ti_var_file(fb, f)) {
      name = filename_to_display(fname);
      messagebox01(ew->window, GTK_MESSAGE_ERROR,
		   "Invalid TI variable file",
		   "The file %s is not a recognized text variable format.",
		   name);
      g_free(name);
      fclose(f);
      fb_book_free(fb);
      return FALSE;
    }

    if (replace) {
      g_free(ew->filename);
      ew->filename = NULL;
      g_free(ew->displayname);
      if (fb->appname->utf8size)
	ew->displayname = g_strndup((char*) fb->appname->utf8,
				    fb->appname->utf8size);
      else
	ew->displayname = g_strdup("Untitled");
    }
  }
  else {
    /* make sure to open text files in text mode */
    fclose(f);
    f = g_fopen(fname, "rt");
    if (!f) {
      display_io_error(ew->window, fname, TRUE, errno);
      return FALSE;
    }
    fb = fb_book_new_for_edit(ew, FALSE);
    if (fb_book_read_xml(fb, f, fname)) {
      /* not valid XML -> fall back to importing as a text file */
      clear_fb_errors(ew);
      fclose(f);
      fb_book_free(fb);
      if (!import_text_file(ew, fname, replace))
	return FALSE;

      update_titlebar(ew);
      update_statusbar(ew);
      return TRUE;
    }
    fclose(f);

    if (replace) {
      g_free(ew->filename);
      if (g_path_is_absolute(fname))
	ew->filename = g_strdup(fname);
      else {
	cwd = g_get_current_dir();
	ew->filename = g_build_filename(cwd, fname, NULL);
	g_free(cwd);
      }

      g_free(ew->displayname);
      ew->displayname = filename_to_display(fname);
    }
  }

  if (replace) {
    g_free(ew->booktitle);
    ew->booktitle = g_strndup((char*) fb->title->utf8, fb->title->utf8size);

    g_free(ew->appname);
    ew->appname = g_strndup((char*) fb->appname->utf8, fb->appname->utf8size);

    g_free(ew->fontfile);
    if (fb->fontfile)
      ew->fontfile = g_strdup(fb->fontfile);
    else
      ew->fontfile = g_strdup(FB_DEFAULT_SMALL_FONT);

    g_free(ew->hyphlang);
    if (fb->hyphlang)
      ew->hyphlang = g_strdup(fb->hyphlang);
    else
      ew->hyphlang = NULL;

    ew->hyphenate = fb->hyphenate ? TRUE : FALSE;

    ew->blocksize = fb->blocksize_limit;

    fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), FALSE);
    gtk_text_buffer_set_text(ew->textbuf, "", -1);
  }

  gtk_text_buffer_begin_user_action(ew->textbuf);

  gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);
  gtk_text_buffer_delete_interactive(ew->textbuf, &start, &end,
				     gtk_text_view_get_editable
				     (GTK_TEXT_VIEW(ew->textedit)));

  if (!replace) {
    if (fb->title->utf8size) {
      gtk_text_buffer_insert_with_tags(ew->textbuf, &start,
				       (gchar*) fb->title->utf8,
				       fb->title->utf8size,
				       ew->tag_section,
				       NULL);
      gtk_text_buffer_insert_with_tags(ew->textbuf, &start, "\n", 1,
				       ew->tag_section, NULL);
    }
  }

  i = j = 0;
  p = (const char*) fb->body->utf8;
  while (i < fb->body->nutf8chars) {
    while (j < fb->nsections && fb->sections[j].pos_utf8 <= i) {
      if (fb->sections[j].title->utf8size > 0) {
	gtk_text_buffer_insert_with_tags(ew->textbuf, &start,
					 (gchar*) fb->sections[j].title->utf8,
					 fb->sections[j].title->utf8size,
					 ew->tag_section,
					 NULL);
      }
      gtk_text_buffer_insert_with_tags(ew->textbuf, &start, "\n", 1,
				       ew->tag_section, NULL);
      j++;
    }
    q = p;
    while (i < fb->body->nutf8chars
	   && (j >= fb->nsections || i < fb->sections[j].pos_utf8)) {
      p = g_utf8_next_char(p);
      i++;
    }
    gtk_text_buffer_insert(ew->textbuf, &start, q, p - q);
  }

  fb_book_free(fb);

  gtk_text_buffer_end_user_action(ew->textbuf);

  if (replace) {
    fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), TRUE);
    gtk_text_buffer_set_modified(ew->textbuf, FALSE);
  }
  update_titlebar(ew);
  update_statusbar(ew);

  if (ew->preview)
    fb_preview_refresh(FB_PREVIEW(ew->preview));

  return TRUE;
}

static gboolean save_file(fbeditwindow* ew, const gchar* fname)
{
  FILE* f;
  GtkTextIter lstart, lend;
  char *p, *q, *r, *s, *t;

  f = g_fopen(fname, "wt");
  if (!f) {
    display_io_error(ew->window, fname, FALSE, errno);
    return FALSE;
  }

  fputs("<?xml version=\"1.0\" encoding=\"UTF-8\"?>\n", f);
  fprintf(f, "<!-- Created by Flashbook Creator %s -->\n",
	  PACKAGE_VERSION);
  fprintf(f, "<flashbook>\n");

  if (ew->booktitle && ew->booktitle[0]) {
    s = g_markup_printf_escaped("<title>%s</title>\n", ew->booktitle);
    fputs(s, f);
    g_free(s);
  }

  if (ew->appname && ew->appname[0]) {
    s = g_markup_printf_escaped("<appname>%s</appname>\n", ew->appname);
    fputs(s, f);
    g_free(s);
  }

  if (ew->fontfile && ew->fontfile[0]) {
    p = g_path_get_dirname(ew->fontfile);
    q = g_path_get_dirname(fname);

    if (!strcmp(p, q)) {
      r = g_path_get_basename(ew->fontfile);
      s = fb_escape_filename(r, -1);
      g_free(r);
    }
    else {
      s = fb_escape_filename(ew->fontfile, -1);
    }

    fprintf(f, "<font>%s</font>\n", s);
    g_free(p);
    g_free(q);
    fbfree(s);
  }

  if (ew->hyphenate || ew->hyphlang) {
    fprintf(f, "<hyphenation enabled=\"%c\"", ew->hyphenate ? '1' : '0');
    if (ew->hyphlang)
      fprintf(f, " lang=\"%s\"", ew->hyphlang);
    fprintf(f, "/>\n");
  }

  if (ew->blocksize != FB_DEFAULT_BLOCK_SIZE) {
    s = g_strdup_printf("<blocksize>%lu</blocksize>\n",
			(unsigned long) ew->blocksize);
    fputs(s, f);
    g_free(s);
  }

  fputs("<body>", f);

  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &lstart, 0);
  lend = lstart;
  while (!gtk_text_iter_is_end(&lstart)) {
    gtk_text_iter_forward_line(&lend);

    s = gtk_text_iter_get_slice(&lstart, &lend);
    t = g_markup_escape_text(s, -1);
    g_free(s);

    if (gtk_text_iter_has_tag(&lstart, ew->tag_section)) {
      s = t + strlen(t);
      while (s != t && (s[-1] == '\n' || s[-1] == '\r'))
	s--;
      *s = 0;

      if (*t)
	fprintf(f, "<section title=\"%s\"/>", t);
      else
	fputs("<section/>", f);
    }
    else {
      fputs(t, f);
    }
    g_free(t);

    lstart = lend;
  }

  fputs("</body>\n</flashbook>\n", f);

  if (ferror(f)) {
    fclose(f);
    display_io_error(ew->window, fname, FALSE, errno);
    return FALSE;
  }
  if (fclose(f)) {
    display_io_error(ew->window, fname, FALSE, errno);
    return FALSE;
  }

  if (fname != ew->filename) {
    g_free(ew->filename);
    ew->filename = g_strdup(fname);
    g_free(ew->displayname);
    ew->displayname = filename_to_display(fname);
  }
  gtk_text_buffer_set_modified(ew->textbuf, FALSE);
  update_titlebar(ew);

  if (ew->preview)
    fb_preview_refresh(FB_PREVIEW(ew->preview));

  return TRUE;
}

static gboolean do_save(fbeditwindow* ew, gboolean saveas)
{
  gchar* suggest_name;
  gchar* fname;

  if (ew->filename && !saveas)
    return save_file(ew, ew->filename);

  if (ew->appname && ew->appname[0]) {
    fname = g_strconcat(ew->appname, ".ebk", NULL);
    suggest_name = g_filename_from_utf8(fname, -1, NULL, NULL, NULL);
    g_free(fname);
  }
  else
    suggest_name = g_strdup("untitled.ebk");

  fname = run_file_chooser("Save Flashbook file",
			   GTK_WINDOW(ew->window),
			   GTK_FILE_CHOOSER_ACTION_SAVE,
			   ew->filename,
			   suggest_name,
			   NULL,
			   "eBook Source File|*.ebk|");

  g_free(suggest_name);

  if (!fname)
    return FALSE;

  if (!save_file(ew, fname)) {
    g_free(fname);
    return FALSE;
  }

  g_free(fname);
  return TRUE;
}

static gboolean prompt_save(fbeditwindow* ew)
{
  GtkWidget* dlg;  

  if (!ew->textbuf || !gtk_text_buffer_get_modified(ew->textbuf))
    return TRUE;

  dlg = gtk_message_dialog_new(GTK_WINDOW(ew->window), GTK_DIALOG_MODAL,
			       GTK_MESSAGE_QUESTION, GTK_BUTTONS_NONE,
			       "Save changes to file %s before closing?",
			       ew->displayname);
  gtk_message_dialog_format_secondary_markup
    (GTK_MESSAGE_DIALOG(dlg),
     "Your changes will be lost if you don't save them.");

  gtk_dialog_add_buttons(GTK_DIALOG(dlg),
			 "Close _Without Saving", GTK_RESPONSE_REJECT,
			 GTK_STOCK_CANCEL, GTK_RESPONSE_CANCEL,
			 GTK_STOCK_SAVE, GTK_RESPONSE_ACCEPT,
			 NULL);
  gtk_dialog_set_alternative_button_order(GTK_DIALOG(dlg),
					  GTK_RESPONSE_ACCEPT,
					  GTK_RESPONSE_REJECT,
					  GTK_RESPONSE_CANCEL,
					  -1);
  gtk_dialog_set_default_response(GTK_DIALOG(dlg), GTK_RESPONSE_ACCEPT);

  switch (gtk_dialog_run(GTK_DIALOG(dlg))) {
  case GTK_RESPONSE_ACCEPT:
    gtk_widget_destroy(dlg);
    return do_save(ew, FALSE);

  case GTK_RESPONSE_REJECT:
    gtk_widget_destroy(dlg);
    return TRUE;

  default:
    gtk_widget_destroy(dlg);
    return FALSE;
  }
}


/**************** Window management ****************/

/* Possible window management modes

   normal SDI: one window only

   non-Mac MDI: each window has own menubar + own set of actions; with
   the exception of the Quit action, each window acts independently

   Mac MDI: menubar and actions are shared; we must track the
   currently active window and keep action states updated accordingly
 */

#ifdef USE_MDI
static fbeditwindow* mdi_window_list;

#ifdef USE_MAC_MENU
static fbeditwindow* mdi_current_window;
static gboolean mdi_fontedit_active;
#endif

static fbeditwindow* create_window();
#endif

static fbeditwindow* get_new_window(fbeditwindow* ew)
{
#ifdef USE_MDI
  ew = create_window();
  ew->mdiprev = NULL;
  ew->mdinext = mdi_window_list;
  mdi_window_list = ew;
  if (ew->mdinext)
    ew->mdinext->mdiprev = ew;

  return ew;

#else
  if (!prompt_save(ew))
    return NULL;
  else
    return ew;
#endif
}

static void cancel_new_window(fbeditwindow* ew G_GNUC_UNUSED)
{
#ifdef USE_MDI
  gtk_widget_destroy(ew->window);
#endif
}

static void show_new_window(fbeditwindow* ew G_GNUC_UNUSED)
{
#ifdef USE_MDI
  gtk_widget_show(ew->window);
#endif
}

static fbeditwindow* get_current_window(fbeditwindow* ew G_GNUC_UNUSED)
{
#ifdef USE_MAC_MENU
  return mdi_current_window;
#else
  return ew;
#endif
}


/**************** Editor actions ****************/

static void action_new(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_new_window(data);

  if (!ew)
    return;

  fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), FALSE);
  gtk_text_buffer_set_text(ew->textbuf, "", -1);
  fb_text_buffer_set_recording(FB_TEXT_BUFFER(ew->textbuf), TRUE);
  gtk_text_buffer_set_modified(ew->textbuf, FALSE);

  g_free(ew->filename);
  ew->filename = NULL;
  g_free(ew->displayname);
  ew->displayname = g_strdup("Untitled");
  g_free(ew->booktitle);
  ew->booktitle = NULL;
  g_free(ew->appname);
  ew->appname = NULL;
  g_free(ew->fontfile);
  ew->fontfile = g_strdup(FB_DEFAULT_SMALL_FONT);
  g_free(ew->hyphlang);
  ew->hyphlang = NULL;
  ew->hyphenate = FALSE;

  update_titlebar(ew);
  update_statusbar(ew);
  if (ew->preview)
    fb_preview_refresh(FB_PREVIEW(ew->preview));

  show_new_window(ew);
}

#ifdef USE_UNIFIED_MDI
# ifdef USE_MAC_MENU
static gboolean mac_menu_focus_in();
static gboolean mac_menu_focus_out();
# else
#  define mac_menu_focus_in NULL
#  define mac_menu_focus_out NULL
# endif

static void action_new_font(GtkAction* action G_GNUC_UNUSED,
			    gpointer data G_GNUC_UNUSED)
{
  fontedit_init(mac_menu_focus_in, mac_menu_focus_out);
  fontedit_new();
}
#endif

static void action_open(GtkAction* action G_GNUC_UNUSED, gpointer data)
{ 
  fbeditwindow* ew = get_new_window(data);
  gchar* fname;
  gchar* suggest_dir;

  if (!ew)
    return;

  if (ew->filename)
    suggest_dir = g_path_get_dirname(ew->filename);
  else
    suggest_dir = NULL;

  fname = run_file_chooser("Open Flashbook file",
			   GTK_WINDOW(ew->window),
			   GTK_FILE_CHOOSER_ACTION_OPEN,
			   NULL,
			   NULL,
			   suggest_dir,
			   "eBook Source Files|*.ebk|"
			   "Flash Applications|*.8xk|"
			   "Notefolio Files|*.8xv|"
			   "QWERTY Files|*.83p;*.8xp;*.8xz|"
			   "Text Files|*.txt|"
#ifdef USE_UNIFIED_MDI
			   "Flashbook Font Files|*.fbf|"
#endif
			   "All Files|*|");

  g_free(suggest_dir);

#ifdef USE_UNIFIED_MDI
  if (fname) {
    const char* ext = strrchr(fname, '.');
    if (ext && !g_ascii_strcasecmp(ext, ".fbf")) {
      fontedit_init(mac_menu_focus_in, mac_menu_focus_out);
      fontedit_open(fname);
      cancel_new_window(ew);
      g_free(fname);
      return;
    }
  }
#endif

  if (fname) {
    if (open_file(ew, fname, TRUE))
      show_new_window(ew);
    else
      cancel_new_window(ew);
    g_free(fname);
  }
}

static void action_import(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_new_window(data);
  gchar* fname;
  gchar* suggest_dir;

  if (!ew)
    return;

  if (ew->filename)
    suggest_dir = g_path_get_dirname(ew->filename);
  else
    suggest_dir = NULL;

  fname = run_file_chooser("Import text file",
			   GTK_WINDOW(ew->window),
			   GTK_FILE_CHOOSER_ACTION_OPEN,
			   NULL,
			   NULL,
			   suggest_dir,
			   "Text Files|*.txt|"
			   "All Files|*|");

  g_free(suggest_dir);

  if (fname) {
    if (import_text_file(ew, fname, TRUE)) {
      update_titlebar(ew);
      update_statusbar(ew);
      show_new_window(ew);
    }
    else {
      cancel_new_window(ew);
    }
    g_free(fname);
  }
}

static void action_save(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

#if defined(USE_MAC_MENU) && defined(USE_UNIFIED_MDI)
  if (mdi_fontedit_active) {
    fontedit_save();
    return;
  }
#endif

  if (ew)
    do_save(ew, FALSE);
}

static void action_save_as(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

#if defined(USE_MAC_MENU) && defined(USE_UNIFIED_MDI)
  if (mdi_fontedit_active) {
    fontedit_save_as();
    return;
  }
#endif

  if (ew)
    do_save(ew, TRUE);
}

static void action_build(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

  if (!ew)
    return;

  if (ew->dlg_build) {
    gtk_window_present(GTK_WINDOW(ew->dlg_build));
    return;
  }

  run_builder_dialog(ew);

  /* refresh preview -- font may have changed */
  if (ew->preview)
    fb_preview_refresh(FB_PREVIEW(ew->preview));
}

#ifdef USE_MDI
static void action_close(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

# if defined(USE_MAC_MENU) && defined(USE_UNIFIED_MDI)
  if (mdi_fontedit_active) {
    fontedit_close();
    return;
  }
# endif

  if (!ew)
    return;

  if (!prompt_save(ew))
    return;

  if (ew->window)
    gtk_widget_destroy(ew->window);
}
#endif

static void action_quit(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = data;

#ifdef USE_MDI

# ifdef USE_UNIFIED_MDI
  if (!fontedit_prompt_save())
    return;
# endif

  for (ew = mdi_window_list; ew; ew = ew->mdinext) {
    if (!prompt_save(ew))
      return;
  }

# ifdef USE_UNIFIED_MDI
  fontedit_quit();
# endif

  while (mdi_window_list) {
    ew = mdi_window_list->mdinext;
    gtk_widget_destroy(mdi_window_list->window);
    mdi_window_list = ew;
  }
  gtk_main_quit();
#else
  if (!prompt_save(ew))
    return;

  gtk_widget_destroy(ew->window);
#endif
}

static void action_undo(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkTextMark* mark;

  if (!ew)
    return;

  fb_text_buffer_undo(FB_TEXT_BUFFER(ew->textbuf));
  mark = gtk_text_buffer_get_insert(ew->textbuf);
  gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(ew->textedit), mark);
}

static void action_redo(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkTextMark* mark;

  if (!ew)
    return;

  fb_text_buffer_redo(FB_TEXT_BUFFER(ew->textbuf));
  mark = gtk_text_buffer_get_insert(ew->textbuf);
  gtk_text_view_scroll_mark_onscreen(GTK_TEXT_VIEW(ew->textedit), mark);
}

static void action_cut(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

  if (!ew)
    return;

  gtk_text_buffer_cut_clipboard(ew->textbuf,
				gtk_clipboard_get(GDK_SELECTION_CLIPBOARD),
				TRUE);
}

static void action_copy(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

  if (!ew)
    return;

  gtk_text_buffer_copy_clipboard(ew->textbuf,
				 gtk_clipboard_get(GDK_SELECTION_CLIPBOARD));
}

static void action_paste(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

  if (!ew)
    return;

  gtk_text_buffer_paste_clipboard(ew->textbuf,
				  gtk_clipboard_get(GDK_SELECTION_CLIPBOARD),
				  NULL, TRUE);
}

static void action_delete(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkTextIter start, end;

  if (!ew)
    return;

  gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);
  gtk_text_buffer_delete_interactive(ew->textbuf, &start, &end,
				     gtk_text_view_get_editable
				     (GTK_TEXT_VIEW(ew->textedit)));
}

static void action_select_all(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkTextIter start, end;

  if (!ew)
    return;

  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &start, 0);
  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &end, -1);
  gtk_text_buffer_select_range(ew->textbuf, &start, &end);
}


static void action_find(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);

  if (!ew)
    return;

  if (ew->dlg_search)
    gtk_window_present(GTK_WINDOW(ew->dlg_search));
  else
    start_search_dialog(ew);
}

static void action_new_section(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkTextIter start, end;
  int ostart, oend;

  if (!ew || !gtk_text_view_get_editable(GTK_TEXT_VIEW(ew->textedit)))
    return;

  gtk_text_buffer_begin_user_action(ew->textbuf);

  gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);

  if (!gtk_text_iter_starts_line(&start)) {
    gtk_text_buffer_insert_interactive(ew->textbuf, &start, "\n", 1, TRUE);
    gtk_text_buffer_get_selection_bounds(ew->textbuf, &start, &end);
  }

  ostart = gtk_text_iter_get_offset(&start);
  oend = gtk_text_iter_get_offset(&end);
  gtk_text_buffer_insert_interactive(ew->textbuf, &end, "\n", 1, TRUE);
  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &start, ostart);
  gtk_text_buffer_apply_tag(ew->textbuf, ew->tag_section, &start, &end);
  gtk_text_buffer_get_iter_at_offset(ew->textbuf, &end, oend);
  gtk_text_buffer_select_range(ew->textbuf, &start, &end);

  gtk_text_buffer_end_user_action(ew->textbuf);
}

static void action_insert_file(GtkAction* action G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  gchar* fname;
  gchar* suggest_dir;

  if (!ew)
    return;

  if (ew->filename)
    suggest_dir = g_path_get_dirname(ew->filename);
  else
    suggest_dir = NULL;

  fname = run_file_chooser("Insert file",
			   GTK_WINDOW(ew->window),
			   GTK_FILE_CHOOSER_ACTION_OPEN,
			   NULL,
			   NULL,
			   suggest_dir,
			   "eBook Source Files|*.ebk|"
			   "Flash Applications|*.8xk|"
			   "Notefolio Files|*.8xv|"
			   "QWERTY Files|*.83p;*.8xp;*.8xz|"
			   "Text Files|*.txt|"
			   "All Files|*|");
  g_free(suggest_dir);

  if (fname) {
    open_file(ew, fname, FALSE);
    update_titlebar(ew);
    update_statusbar(ew);
    g_free(fname);
  }
}

static void action_about(GtkAction* action G_GNUC_UNUSED,
			 gpointer data G_GNUC_UNUSED)
{
#ifdef USE_MAC_MENU
  show_about_dialog(NULL);
#else
  fbeditwindow* ew = get_current_window(data);
  show_about_dialog(ew ? GTK_WINDOW(ew->window) : NULL);
#endif
}

static fbbook* preview_callback(gpointer data)
{
  fbeditwindow* ew = data;
  fbbook* fb = create_fb_from_edit(ew, TRUE);
  int err, err1;

  err1 = fb_book_assign_font(fb);
  fb_chunk_convert_to_ti(fb, fb->body, 0);
  err = fb_chunk_split_lines(fb, fb->body);

  if (!err)
    err = err1;

  if (err != ew->previewerror) {
    ew->previewerror = err;
    update_statusbar(ew);
  }

  return fb;
}

static void action_show_preview(GtkToggleAction* action, gpointer data)
{
  fbeditwindow* ew = get_current_window(data);
  GtkWidget *frame, *box;

  if (!ew)
    return;

  if (!gtk_toggle_action_get_active(action)) {
    if (ew->previewpane)
      gtk_widget_destroy(ew->previewpane);
    ew->preview = ew->previewpane = NULL;
    ew->previewerror = 0;
    update_statusbar(ew);
  }
  else if (!ew->preview) {
    ew->previewpane = frame = gtk_frame_new("<b>Preview</b>");
    g_object_set(gtk_frame_get_label_widget(GTK_FRAME(frame)),
		 "use-markup", TRUE, NULL);
    gtk_frame_set_shadow_type(GTK_FRAME(frame), GTK_SHADOW_NONE);
    gtk_container_set_border_width(GTK_CONTAINER(frame), 3);
    {
      box = fb_preview_box_new(preview_callback, ew, &ew->preview);
      gtk_container_add(GTK_CONTAINER(frame), box);
      gtk_widget_show(box);
    }
    gtk_box_pack_start(GTK_BOX(ew->mainbox), frame, FALSE, FALSE, 0);
    gtk_widget_show(frame);
  }
}

static void text_inserted(GtkTextBuffer* textbuf,
			  GtkTextIter* iter,
			  const gchar* text,
			  gint len,
			  gpointer data)
{
  fbeditwindow* ew = data;
  GtkTextIter lstart, lend;
  int i;
  int o, n, nt;

  if (gtk_text_iter_begins_tag(iter, ew->tag_section)
      && !gtk_text_iter_starts_line(iter)) {
    lstart = *iter;
    gtk_text_iter_set_line_offset(&lstart, 0);
    gtk_text_buffer_apply_tag(textbuf, ew->tag_section, &lstart, iter);
  }
  else if (gtk_text_iter_has_tag(iter, ew->tag_section)) {

    /* If we are inserting within a section tag, we want all text up to
       and including the first newline to be section, and subsequent
       text (even potentially including part of the old title) to be
       body. */

    if (len < 0)
      len = strlen(text);
    for (i = 0; i < len; i++) {
      if (text[i] == '\n') {
	i++;

	o = gtk_text_iter_get_offset(iter);

	nt = g_utf8_pointer_to_offset(text, text + i);
	n = nt + g_utf8_pointer_to_offset(text + i, text + len);

	gtk_text_buffer_get_iter_at_offset(textbuf, &lstart, o - n);
	gtk_text_buffer_get_iter_at_offset(textbuf, &lend, o - n + nt);

	gtk_text_buffer_apply_tag(textbuf, ew->tag_section, &lstart, &lend);

	lstart = *iter;
	gtk_text_iter_set_line_offset(&lstart, 0);
	gtk_text_iter_forward_line(&lstart);
	gtk_text_buffer_remove_tag(textbuf, ew->tag_section, &lend, &lstart);

	break;
      }
    }
  }
}

static void text_deleted(GtkTextBuffer* textbuf,
			 GtkTextIter* start,
			 GtkTextIter* end G_GNUC_UNUSED, /* after deletion,
							    start == end */
			 gpointer data)
{
  fbeditwindow* ew = data;
  GtkTextIter lstart;

  if (gtk_text_iter_begins_tag(start, ew->tag_section)) {
    if (!gtk_text_iter_starts_line(start)) {
      lstart = *start;
      gtk_text_iter_set_line_offset(&lstart, 0);
      gtk_text_buffer_apply_tag(textbuf, ew->tag_section, &lstart, start);
    }
  }
  else if (gtk_text_iter_ends_tag(start, ew->tag_section)) {
    if (!gtk_text_iter_starts_line(start)) {
      lstart = *start;
      gtk_text_iter_set_line_offset(&lstart, 0);
      gtk_text_iter_forward_line(&lstart);
      gtk_text_buffer_apply_tag(textbuf, ew->tag_section, start, &lstart);
    }
  }
}

static void text_mark_set(GtkTextBuffer* textbuf,
			  GtkTextIter* location G_GNUC_UNUSED,
			  GtkTextMark* mark G_GNUC_UNUSED,
			  gpointer data)
{
  fbeditwindow* ew = data;
  gboolean sel = gtk_text_buffer_get_selection_bounds(textbuf, NULL, NULL);

#ifdef USE_MAC_MENU
  if (ew != mdi_current_window) return;
#endif
  gtk_action_set_sensitive(ew->action_cut, sel);
  gtk_action_set_sensitive(ew->action_copy, sel);
  gtk_action_set_sensitive(ew->action_delete, sel);
}

static void text_modified_changed(GtkTextBuffer* textbuf G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = data;
  update_titlebar(ew);
}

static void text_changed(GtkTextBuffer* textbuf G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = data;
  update_statusbar(ew);
  text_mark_set(ew->textbuf, NULL, NULL, ew);
  if (ew->preview)
    fb_preview_refresh(FB_PREVIEW(ew->preview));
}

static void undo_state_changed(FBTextBuffer* textbuf, gpointer data)
{
  fbeditwindow* ew = data;

#ifdef USE_MAC_MENU
  if (ew != mdi_current_window) return;
#endif
  gtk_action_set_sensitive(ew->action_undo, fb_text_buffer_can_undo(textbuf));
  gtk_action_set_sensitive(ew->action_redo, fb_text_buffer_can_redo(textbuf));
}

static void style_update(GtkWidget* widget,
			 GtkStyle* oldstyle G_GNUC_UNUSED,
			 gpointer data)
{
  fbeditwindow* ew = data;

  g_object_set(ew->tag_section,
	       "background-gdk", &widget->style->bg[GTK_STATE_NORMAL],
	       "foreground-gdk", &widget->style->fg[GTK_STATE_NORMAL],
	       NULL);

  /* paragraph-background[-gdk] isn't available in GTK+ 2.6 */

  if (gtk_major_version > 2
      || (gtk_major_version == 2 && gtk_minor_version >= 8))
    g_object_set(ew->tag_section,
		 "paragraph-background-gdk",
		 &widget->style->bg[GTK_STATE_NORMAL],
		 NULL);
}

static gboolean window_delete(GtkWidget* widget G_GNUC_UNUSED,
			      GdkEvent* event G_GNUC_UNUSED,
			      gpointer data)
{
  fbeditwindow* ew = data;
  return (prompt_save(ew) ? FALSE : TRUE);
}

#ifndef USE_MAC_MENU
static void menu_item_select(GtkItem* item, gpointer data)
{
  fbeditwindow* ew = data;
  gtk_statusbar_push(GTK_STATUSBAR(ew->status), ew->scid_menu_hint,
		     g_object_get_data(G_OBJECT(item), "fbedit-tooltip"));
}

static void menu_item_deselect(GtkItem* item G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = data;
  gtk_statusbar_pop(GTK_STATUSBAR(ew->status), ew->scid_menu_hint);  
}

static void uimgr_connect_proxy(GtkUIManager* uimgr G_GNUC_UNUSED,
				GtkAction* action, GtkWidget* proxy,
				gpointer data)
{
  gchar* text;
  if (GTK_IS_MENU_ITEM(proxy)) {
    g_object_get(action, "tooltip", &text, NULL);
    if (text) {
      g_object_set_data(G_OBJECT(proxy), "fbedit-tooltip", text);
      g_signal_connect(proxy, "select", G_CALLBACK(menu_item_select), data);
      g_signal_connect(proxy, "deselect", G_CALLBACK(menu_item_deselect), data);
    }
  }
}

static void uimgr_disconnect_proxy(GtkUIManager* uimgr G_GNUC_UNUSED,
				   GtkAction* action G_GNUC_UNUSED,
				   GtkWidget* proxy, gpointer data)
{
  gchar* text;
  if (GTK_IS_MENU_ITEM(proxy)) {
    g_signal_handlers_disconnect_by_func(proxy, G_CALLBACK(menu_item_select), data);
    g_signal_handlers_disconnect_by_func(proxy, G_CALLBACK(menu_item_deselect), data);
    text = g_object_get_data(G_OBJECT(proxy), "fbedit-tooltip");
    g_free(text);
  }
}
#endif


#ifndef GTK_STOCK_SELECT_ALL
#define GTK_STOCK_SELECT_ALL "gtk-select-all"
#endif

/* Global actions (always available) */

static const GtkActionEntry global_action_ents[] =
  {{ "file-menu", 0, "_File", 0, 0, 0 },

#ifdef USE_UNIFIED_MDI
   
   { "new-menu", GTK_STOCK_NEW, "_New", "", 0, 0 },

   { "new",      0,                "_eBook",                "<Control>N",
     "Create a new book",
     G_CALLBACK(action_new) },

   { "new-font", 0,                "_Flashbook Font...",    0,
     "Create a new Flashbook font file",
     G_CALLBACK(action_new_font) },

#else

   { "new",     GTK_STOCK_NEW,     0,                       0,
     "Create a new book",
     G_CALLBACK(action_new) },

#endif

   { "open",    GTK_STOCK_OPEN,    "_Open...",              0,
     "Open an existing book",
     G_CALLBACK(action_open) },

   { "import",  GTK_STOCK_OPEN,    "_Import Text File...",  "",
     "Import text from a file",
     G_CALLBACK(action_import) },

   { "quit",    GTK_STOCK_QUIT,    0,                       0,
     "Quit Flashbook Creator",
     G_CALLBACK(action_quit) },

   { "edit-menu", 0, "_Edit", 0, 0, 0 },

   { "view-menu", 0, "_View", 0, 0, 0 },

   { "help-menu", 0, "_Help", 0, 0, 0 },

   { "about",   GTK_STOCK_ABOUT,   0,                       0,
     "About Flashbook Creator",
     G_CALLBACK(action_about) }};


/* Document actions (on the Mac, available only when a document window
   is selected) */

static const GtkActionEntry document_action_ents[] =
  {{ "save",    GTK_STOCK_SAVE,    0,                       0,
     "Save the current book",
     G_CALLBACK(action_save) },

#ifdef USE_MDI
  { "close",   GTK_STOCK_CLOSE,   0,                       0,
    "Close the current book",
    G_CALLBACK(action_close) },
#endif

   { "save-as", GTK_STOCK_SAVE_AS, "Save _As...",    "<Shift><Control>S",
     "Save the current book to a new file",
     G_CALLBACK(action_save_as) }};


/* Creator actions (using the unified GUI, available only when a
   Creator window is selected) */

static const GtkActionEntry creator_action_ents[] =
  {{ "build",   GTK_STOCK_EXECUTE, "_Build Application...", "<Control>B",
     "Build the current book into a Flash application",
     G_CALLBACK(action_build) },

   { "undo",    GTK_STOCK_UNDO,    0,                       "<Control>Z",
     "Undo the previous action",
     G_CALLBACK(action_undo) },

   { "redo",    GTK_STOCK_REDO,    0,                "<Shift><Control>Z",
     "Redo the previous action",
     G_CALLBACK(action_redo) },

   { "cut",     GTK_STOCK_CUT,     0,                       0,
     "Cut the selected text to the clipboard",
     G_CALLBACK(action_cut) },

   { "copy",    GTK_STOCK_COPY,    0,                       0,
     "Copy the selected text to the clipboard",
     G_CALLBACK(action_copy) },

   { "paste",   GTK_STOCK_PASTE,   0,                       0,
     "Paste text from the clipboard",
     G_CALLBACK(action_paste) },

   { "delete",  GTK_STOCK_DELETE,  0,                       0,
     "Delete the selected text",
     G_CALLBACK(action_delete) },

   { "select-all", GTK_STOCK_SELECT_ALL, "Select _All",     "<Control>A",
     "Select all text",
     G_CALLBACK(action_select_all) },

   { "find", GTK_STOCK_FIND_AND_REPLACE, "_Find and Replace...", "<Control>F",
     "Search and/or replace text strings in the book",
     G_CALLBACK(action_find) },

   { "new-section", 0,             "Insert _Section Break", "<Control>T",
     "Insert a section break at the current position",
     G_CALLBACK(action_new_section) },

   { "insert-file", 0,             "_Insert File...",       0,
     "Insert text from a file",
     G_CALLBACK(action_insert_file) }};


static const GtkToggleActionEntry creator_toggle_action_ents[] =
  {{ "show-preview", 0,            "_Preview",              "<Control>P",
     "Show a preview of the current book",
     G_CALLBACK(action_show_preview), FALSE }};


static const char uidesc[] =
  "<menubar name=\"menu-bar\">"
  " <menu action=\"file-menu\">"
#ifdef USE_UNIFIED_MDI
  "  <menu action=\"new-menu\">"
  "   <menuitem action=\"new\"/>"
  "   <menuitem action=\"new-font\"/>"
  "  </menu>"
#else
  "  <menuitem action=\"new\"/>"
#endif
  "  <menuitem action=\"open\"/>"
  "  <menuitem action=\"import\"/>"
  "  <separator/>"
#ifdef USE_MDI
  "  <menuitem action=\"close\"/>"
#endif
  "  <menuitem action=\"save\"/>"
  "  <menuitem action=\"save-as\"/>"
  "  <separator/>"
  "  <menuitem action=\"build\"/>"
  "  <separator/>"
  "  <menuitem action=\"quit\"/>"
  " </menu>"
  " <menu action=\"edit-menu\">"
  "  <menuitem action=\"undo\"/>"
  "  <menuitem action=\"redo\"/>"
  "  <separator/>"
  "  <menuitem action=\"cut\"/>"
  "  <menuitem action=\"copy\"/>"
  "  <menuitem action=\"paste\"/>"
  "  <menuitem action=\"delete\"/>"
  "  <menuitem action=\"select-all\"/>"
  "  <separator/>"
  "  <menuitem action=\"find\"/>"
  "  <separator/>"
  "  <menuitem action=\"new-section\"/>"
  "  <menuitem action=\"insert-file\"/>"
  " </menu>"
  " <menu action=\"view-menu\">"
  "  <menuitem action=\"show-preview\"/>"
  " </menu>"
  " <menu action=\"help-menu\">"
  "  <menuitem action=\"about\" />"
  " </menu>"
  "</menubar>";


#ifdef USE_MAC_MENU

static GtkAction *mac_menu_action_undo, *mac_menu_action_redo,
  *mac_menu_action_cut, *mac_menu_action_copy, *mac_menu_action_delete;
static GtkToggleAction *mac_menu_action_preview;

static GtkActionGroup *mac_menu_creator_actions, *mac_menu_document_actions;

static void mac_menu_init()
{
  GtkActionGroup *global_actions;
  GtkUIManager *uimgr;
  GtkWidget *menubar, *quititem, *aboutitem;
  GError* err = NULL;
  IgeMacMenuGroup* appgrp;
  char* lbl;

  uimgr = gtk_ui_manager_new();

  global_actions = gtk_action_group_new("Main");
  gtk_action_group_add_actions(global_actions, global_action_ents,
			       G_N_ELEMENTS(global_action_ents), NULL);

  mac_menu_document_actions = gtk_action_group_new("Document");
  gtk_action_group_add_actions(mac_menu_document_actions,
			       document_action_ents,
			       G_N_ELEMENTS(document_action_ents), NULL);

  mac_menu_creator_actions = gtk_action_group_new("Creator");
  gtk_action_group_add_actions(mac_menu_creator_actions,
			       creator_action_ents,
			       G_N_ELEMENTS(creator_action_ents), NULL);
  gtk_action_group_add_toggle_actions
    (mac_menu_creator_actions, creator_toggle_action_ents,
     G_N_ELEMENTS(creator_toggle_action_ents), NULL);

  mac_menu_action_undo
    = gtk_action_group_get_action(mac_menu_creator_actions, "undo");
  mac_menu_action_redo
    = gtk_action_group_get_action(mac_menu_creator_actions, "redo");
  mac_menu_action_cut
    = gtk_action_group_get_action(mac_menu_creator_actions, "cut");
  mac_menu_action_copy
    = gtk_action_group_get_action(mac_menu_creator_actions, "copy");
  mac_menu_action_delete
    = gtk_action_group_get_action(mac_menu_creator_actions, "delete");
  mac_menu_action_preview = GTK_TOGGLE_ACTION
    (gtk_action_group_get_action(mac_menu_creator_actions, "show-preview"));

  gtk_ui_manager_insert_action_group(uimgr, global_actions, 0);
  gtk_ui_manager_insert_action_group(uimgr, mac_menu_document_actions, 0);
  gtk_ui_manager_insert_action_group(uimgr, mac_menu_creator_actions, 0);

  if (!gtk_ui_manager_add_ui_from_string(uimgr, uidesc, -1, &err)) {
    g_error("Failed to create menus: %s", err->message);
  }
  menubar = gtk_ui_manager_get_widget(uimgr, "/menu-bar");

  ige_mac_menu_set_menu_bar(GTK_MENU_SHELL(menubar));

  quititem = gtk_ui_manager_get_widget(uimgr, "/menu-bar/file-menu/quit");
  ige_mac_menu_set_quit_menu_item(GTK_MENU_ITEM(quititem));

  aboutitem = gtk_ui_manager_get_widget(uimgr, "/menu-bar/help-menu/about");
  appgrp = ige_mac_menu_add_app_menu_group();
  lbl = g_strdup_printf("About %s", g_get_application_name());
  ige_mac_menu_add_app_menu_item(appgrp, GTK_MENU_ITEM(aboutitem), lbl);
  g_free(lbl);
}

static void mac_menu_set_creator_focus(fbeditwindow* ew)
{
  mdi_current_window = ew;
  mdi_fontedit_active = FALSE;

  gtk_action_group_set_sensitive(mac_menu_document_actions, TRUE);
  gtk_action_group_set_sensitive(mac_menu_creator_actions, TRUE);

  /* update Undo/Redo */
  undo_state_changed(FB_TEXT_BUFFER(ew->textbuf), ew);
  /* update Cut/Copy/Delete */
  text_mark_set(ew->textbuf, NULL, NULL, ew);

  mdi_current_window = NULL;
  if (ew->preview)
    gtk_toggle_action_set_active(mac_menu_action_preview, TRUE);
  else
    gtk_toggle_action_set_active(mac_menu_action_preview, FALSE);
  mdi_current_window = ew;
}

static void mac_menu_set_fontedit_focus()
{
  mdi_current_window = NULL;
  mdi_fontedit_active = TRUE;

  gtk_action_group_set_sensitive(mac_menu_document_actions, TRUE);
  gtk_action_group_set_sensitive(mac_menu_creator_actions, FALSE);
}

static void mac_menu_set_null_focus()
{
  mdi_current_window = NULL;
  mdi_fontedit_active = FALSE;

  gtk_action_group_set_sensitive(mac_menu_document_actions, FALSE);
  gtk_action_group_set_sensitive(mac_menu_creator_actions, FALSE);
}

static gboolean mac_menu_focus_in(GtkWidget* w G_GNUC_UNUSED,
				  GdkEvent* ev G_GNUC_UNUSED,
				  gpointer data)
{
  fbeditwindow* ew = data;
  if (ew)
    mac_menu_set_creator_focus(ew);
  else
    mac_menu_set_fontedit_focus();
  return FALSE;
}

static gboolean mac_menu_focus_out(GtkWidget* w G_GNUC_UNUSED,
				   GdkEvent* ev G_GNUC_UNUSED,
				   gpointer data)
{
  fbeditwindow* ew = data;
  if (ew == mdi_current_window) {
# ifndef TEST_MAC_MDI_MODE
    mac_menu_set_null_focus();
# endif
  }
  return FALSE;
}

#endif	/* USE_MAC_MENU */


static void destroy_window(GtkObject* obj G_GNUC_UNUSED, gpointer data)
{
  fbeditwindow* ew = data;

#ifdef USE_MAC_MENU
  if (ew == mdi_current_window)
    mac_menu_set_null_focus();
#endif

#ifdef USE_MDI
  if (ew->mdiprev)
    ew->mdiprev->mdinext = ew->mdinext;
  if (ew->mdinext)
    ew->mdinext->mdiprev = ew->mdiprev;
  if (mdi_window_list == ew)
    mdi_window_list = ew->mdinext;

# ifndef USE_MAC_MENU
  if (!mdi_window_list)
    gtk_main_quit();
# endif

#else
  gtk_main_quit();
#endif

  g_free(ew->filename);
  g_free(ew->displayname);
  g_free(ew->booktitle);
  g_free(ew->appname);
  g_free(ew->fontfile);
  g_free(ew->hyphlang);
  g_free(ew->error_log);
  g_free(ew);
}

static fbeditwindow* create_window()
{
  fbeditwindow *ew;
  GtkWidget *vbox, *sw;
#ifndef USE_MAC_MENU
  GtkWidget *menubar;
  GtkActionGroup *actgrp;
  GtkUIManager *uimgr;
  GtkAccelGroup *accelgrp;
  GError* err = NULL;
#endif

  ew = g_new0(fbeditwindow, 1);
  ew->filename = NULL;
  ew->displayname = g_strdup("Untitled");
  ew->fontfile = g_strdup(FB_DEFAULT_SMALL_FONT);
  ew->blocksize = FB_DEFAULT_BLOCK_SIZE;
  ew->hyphlang = NULL;
  ew->hyphenate = FALSE;

  ew->mdiprev = NULL;
  ew->mdinext = NULL;

  ew->window = gtk_window_new(GTK_WINDOW_TOPLEVEL);
  gtk_window_set_default_size(GTK_WINDOW(ew->window), 600, 400);
  {
    vbox = gtk_vbox_new(FALSE, 0);
    {
#ifdef USE_MAC_MENU
      ew->action_undo = mac_menu_action_undo;
      ew->action_redo = mac_menu_action_redo;
      ew->action_cut = mac_menu_action_cut;
      ew->action_copy = mac_menu_action_copy;
      ew->action_delete = mac_menu_action_delete;
      g_signal_connect(ew->window, "focus-in-event",
		       G_CALLBACK(mac_menu_focus_in), ew);
      g_signal_connect(ew->window, "focus-out-event",
		       G_CALLBACK(mac_menu_focus_out), ew);
      ige_mac_menu_connect_window_key_handler(GTK_WINDOW(ew->window));
#else
      actgrp = gtk_action_group_new("Menu");
      gtk_action_group_add_actions(actgrp, global_action_ents,
				   G_N_ELEMENTS(global_action_ents), ew);
      gtk_action_group_add_actions(actgrp, document_action_ents,
				   G_N_ELEMENTS(document_action_ents), ew);
      gtk_action_group_add_actions(actgrp, creator_action_ents,
				   G_N_ELEMENTS(creator_action_ents), ew);
      gtk_action_group_add_toggle_actions
	(actgrp, creator_toggle_action_ents,
	 G_N_ELEMENTS(creator_toggle_action_ents), ew);

      uimgr = gtk_ui_manager_new();
      g_signal_connect(uimgr, "connect-proxy",
		       G_CALLBACK(uimgr_connect_proxy), ew);
      g_signal_connect(uimgr, "disconnect-proxy",
		       G_CALLBACK(uimgr_disconnect_proxy), ew);

      ew->action_undo = gtk_action_group_get_action(actgrp, "undo");
      ew->action_redo = gtk_action_group_get_action(actgrp, "redo");
      ew->action_cut = gtk_action_group_get_action(actgrp, "cut");
      ew->action_copy = gtk_action_group_get_action(actgrp, "copy");
      ew->action_delete = gtk_action_group_get_action(actgrp, "delete");

      gtk_ui_manager_insert_action_group(uimgr, actgrp, 0);
      accelgrp = gtk_ui_manager_get_accel_group(uimgr);
      gtk_window_add_accel_group(GTK_WINDOW(ew->window), accelgrp);
      if (!gtk_ui_manager_add_ui_from_string(uimgr, uidesc, -1, &err)) {
	g_error("Failed to create menus: %s", err->message);
      }
      menubar = gtk_ui_manager_get_widget(uimgr, "/menu-bar");
      gtk_box_pack_start(GTK_BOX(vbox), menubar, FALSE, FALSE, 0);
#endif

      ew->mainbox = gtk_hbox_new(FALSE, 0);
      {
	sw = gtk_scrolled_window_new(NULL, NULL);
	gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				       GTK_POLICY_AUTOMATIC,
				       GTK_POLICY_AUTOMATIC);
	gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
					    GTK_SHADOW_IN);
	{
	  ew->textbuf = fb_text_buffer_new();
	  ew->textedit = gtk_text_view_new_with_buffer(ew->textbuf);
	  ew->tag_section = gtk_text_buffer_create_tag
	    (ew->textbuf, "section",
	     "style", PANGO_STYLE_ITALIC,
	     "scale", PANGO_SCALE_LARGE,
	     "justification", GTK_JUSTIFY_CENTER,
	     "pixels-above-lines", 3,
	     "pixels-below-lines", 3,
	     NULL);

	  gtk_text_view_set_wrap_mode(GTK_TEXT_VIEW(ew->textedit),
				      GTK_WRAP_WORD);
	  gtk_text_view_set_pixels_above_lines(GTK_TEXT_VIEW(ew->textedit),
					       2);
	  gtk_text_view_set_pixels_below_lines(GTK_TEXT_VIEW(ew->textedit),
					       2);

#if GTK_CHECK_VERSION(2, 10, 0)
	  gtk_text_buffer_register_serialize_tagset(ew->textbuf,
						    "flashbook-1.0");
	  gtk_text_buffer_register_deserialize_tagset(ew->textbuf,
						      "flashbook-1.0");
#endif

	  g_signal_connect(ew->textbuf, "undo-state-changed",
			   G_CALLBACK(undo_state_changed), ew);
	  undo_state_changed(FB_TEXT_BUFFER(ew->textbuf), ew);

	  g_signal_connect(ew->textbuf, "mark-set",
			   G_CALLBACK(text_mark_set), ew);
	  text_mark_set(ew->textbuf, NULL, NULL, ew);

	  g_signal_connect_after(ew->textbuf, "insert-text",
				 G_CALLBACK(text_inserted), ew);
	  g_signal_connect_after(ew->textbuf, "delete-range",
				 G_CALLBACK(text_deleted), ew);
	  g_signal_connect(ew->textbuf, "modified-changed",
			   G_CALLBACK(text_modified_changed), ew);
	  g_signal_connect(ew->textbuf, "changed",
			   G_CALLBACK(text_changed), ew);

	  g_signal_connect(ew->textedit, "style-set",
			   G_CALLBACK(style_update), ew);

	  style_update(ew->textedit, NULL, ew);

	  gtk_container_add(GTK_CONTAINER(sw), ew->textedit);
	  gtk_widget_show(ew->textedit);
	}
	gtk_box_pack_end(GTK_BOX(ew->mainbox), sw, TRUE, TRUE, 0);
	gtk_widget_show(sw);
      }
      gtk_box_pack_start(GTK_BOX(vbox), ew->mainbox, TRUE, TRUE, 0);
      gtk_widget_show(ew->mainbox);

      ew->status = gtk_statusbar_new();
      gtk_box_pack_start(GTK_BOX(vbox), ew->status, FALSE, FALSE, 0);
      gtk_widget_show(ew->status);

      ew->scid_menu_hint = gtk_statusbar_get_context_id(GTK_STATUSBAR(ew->status),
							"menu-hint");
      ew->scid_text_info = gtk_statusbar_get_context_id(GTK_STATUSBAR(ew->status),
							"text-info");
      ew->smid_text_info = gtk_statusbar_push(GTK_STATUSBAR(ew->status),
					      ew->scid_text_info, "");
    }
    gtk_container_add(GTK_CONTAINER(ew->window), vbox);
    gtk_widget_show(vbox);
  }
  g_signal_connect(ew->window, "destroy", G_CALLBACK(destroy_window), ew);
  g_signal_connect(ew->window, "delete-event", G_CALLBACK(window_delete), ew);

  update_titlebar(ew);
  update_statusbar(ew);

  return ew;
}

int main(int argc, char** argv)
{
  fbeditwindow* ew;
  char* fname;

  fix_main_args(&argc, &argv);

  gtk_init(&argc, &argv);
#ifdef USE_UNIFIED_MDI
  g_set_application_name("Flashbook");
#else
  g_set_application_name("Flashbook Creator");
#endif
  g_thread_init(NULL);
  rs_set_progname(argv[0]);
  rs_set_verbose(0);

  if ((fname = fb_find_common_file(NULL, "fbcreator.png"))) {
    gtk_window_set_default_icon_from_file(fname, NULL);
    fbfree(fname);
  }

  load_menurc_file("creator");

#ifdef USE_MAC_MENU
  mac_menu_init();
#endif

  ew = create_window();

#ifdef USE_MDI
  mdi_window_list = ew;
#endif

  gtk_widget_show(ew->window);

  if (argc > 1)
    open_file(ew, argv[1], TRUE);

  gtk_main();

  save_menurc_file("creator");

  return 0;
}
