/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <fb.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include "fbpreview.h"

#ifndef GDK_WINDOWING_QUARTZ
/* cache a copy of the font as a server-side bitmap; for whatever
   reason, this doesn't work with current GDK (2.14) on Quartz */
# define USE_FONT_BITMAP
#endif

static void fb_preview_finalize(GObject *obj);
static void fb_preview_destroy(GtkObject *object);
static void fb_preview_set_adjustments(FBPreview *preview,
				       GtkAdjustment *hadj,
				       GtkAdjustment *vadj);
static void fb_preview_realize(GtkWidget *widget);
static void fb_preview_unrealize(GtkWidget *widget);
static void fb_preview_size_request(GtkWidget *widget,
				    GtkRequisition *requisition);
static void fb_preview_size_allocate(GtkWidget *widget,
				     GtkAllocation *allocation);
static gboolean fb_preview_expose(GtkWidget *widget,
				  GdkEventExpose *event);
static gboolean fb_preview_button_press(GtkWidget *widget,
					GdkEventButton *event);
static gboolean fb_preview_key_press(GtkWidget *widget,
				     GdkEventKey *event);
static void fb_preview_update_combo(FBPreview *preview);
static void fb_preview_adjustment_changed(GtkAdjustment *adj,
					  gpointer data);


/* GObject stuff... */

G_DEFINE_TYPE(FBPreview, fb_preview, GTK_TYPE_WIDGET);

static void
g_cclosure_user_marshal_VOID__OBJECT_OBJECT (GClosure     *closure,
					     GValue       *return_value
					                   G_GNUC_UNUSED,
                                             guint         n_param_values,
                                             const GValue *param_values,
					     gpointer      invocation_hint
					                   G_GNUC_UNUSED,
                                             gpointer      marshal_data)
{
  typedef void (*GMarshalFunc_VOID__OBJECT_OBJECT) (gpointer     data1,
                                                    gpointer     arg_1,
                                                    gpointer     arg_2,
                                                    gpointer     data2);
  register GMarshalFunc_VOID__OBJECT_OBJECT callback;
  register GCClosure *cc = (GCClosure*) closure;
  register gpointer data1, data2;

  g_return_if_fail (n_param_values == 3);

  if (G_CCLOSURE_SWAP_DATA (closure))
    {
      data1 = closure->data;
      data2 = g_value_peek_pointer (param_values + 0);
    }
  else
    {
      data1 = g_value_peek_pointer (param_values + 0);
      data2 = closure->data;
    }
  callback = (GMarshalFunc_VOID__OBJECT_OBJECT) (marshal_data ? marshal_data : cc->callback);

  callback (data1,
            g_value_get_object (param_values + 1),
            g_value_get_object (param_values + 2),
            data2);
}

static void
fb_preview_class_init (FBPreviewClass *class)
{
  GObjectClass *o_class;
  GtkObjectClass *object_class;
  GtkWidgetClass *widget_class;

  fb_preview_parent_class = g_type_class_peek_parent(class);

  o_class = G_OBJECT_CLASS(class);
  o_class->finalize = fb_preview_finalize;

  object_class = GTK_OBJECT_CLASS(class);
  object_class->destroy = fb_preview_destroy;

  widget_class = GTK_WIDGET_CLASS(class);
  widget_class->realize = fb_preview_realize;
  widget_class->unrealize = fb_preview_unrealize;
  widget_class->size_request = fb_preview_size_request;
  widget_class->size_allocate = fb_preview_size_allocate;
  widget_class->expose_event = fb_preview_expose;
  widget_class->key_press_event = fb_preview_key_press;
  widget_class->button_press_event = fb_preview_button_press;

  widget_class->set_scroll_adjustments_signal =
    g_signal_new("set-scroll-adjustments",
		 G_TYPE_FROM_CLASS(o_class),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(FBPreviewClass, set_scroll_adjustments),
		 NULL, NULL,
		 g_cclosure_user_marshal_VOID__OBJECT_OBJECT,
		 G_TYPE_NONE, 2,
		 GTK_TYPE_ADJUSTMENT,
		 GTK_TYPE_ADJUSTMENT);

  class->set_scroll_adjustments = fb_preview_set_adjustments;
}


static void
fb_preview_init (FBPreview *preview)
{
  preview->refresh = NULL;
  preview->refresh_data = NULL;
  preview->fb = NULL;
  preview->hadjustment = NULL;
  preview->vadjustment = NULL;
  preview->need_refresh = TRUE;

  GTK_WIDGET_SET_FLAGS(GTK_WIDGET(preview), GTK_CAN_FOCUS);
  gtk_widget_add_events(GTK_WIDGET(preview), (GDK_KEY_PRESS_MASK
					      | GDK_SCROLL_MASK
					      | GDK_BUTTON_PRESS_MASK));
}

static void
fb_preview_finalize (GObject *obj)
{
  FBPreview *preview = FB_PREVIEW(obj);

  preview->refresh = NULL;
  preview->refresh_data = NULL;
  if (preview->fb)
    fb_book_free(preview->fb);
  preview->fb = NULL;
  preview->hadjustment = NULL;
  preview->vadjustment = NULL;
  preview->fontbitmap = NULL;
}

static void
fb_preview_destroy (GtkObject *object)
{
  FBPreview* preview = FB_PREVIEW(object);

  if (preview->hadjustment)
    g_object_unref(preview->hadjustment);
  preview->hadjustment = NULL;
  if (preview->vadjustment)
    g_object_unref(preview->vadjustment);
  preview->vadjustment = NULL;
  if (preview->fontbitmap)
    g_object_unref(preview->fontbitmap);
  preview->fontbitmap = NULL;

  if (GTK_OBJECT_CLASS(fb_preview_parent_class)->destroy)
    (*GTK_OBJECT_CLASS(fb_preview_parent_class)->destroy)(object);
}

static void
fb_preview_set_adjustments (FBPreview *preview,
			    GtkAdjustment *hadj,
			    GtkAdjustment *vadj)
{
  gboolean need_adjust = FALSE;

  if (!hadj)
    hadj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 0.0, 1.0, 1.0, 1.0));
  else
    g_return_if_fail(GTK_IS_ADJUSTMENT(hadj));

  if (!vadj)
    vadj = GTK_ADJUSTMENT(gtk_adjustment_new(0.0, 0.0, 0.0, 1.0, 1.0, 1.0));
  else
    g_return_if_fail(GTK_IS_ADJUSTMENT(vadj));

  if (preview->hadjustment != hadj) {
    if (preview->hadjustment) {
      g_signal_handlers_disconnect_by_func(preview->hadjustment,
					   fb_preview_adjustment_changed,
					   preview);
      g_object_unref(preview->hadjustment);
    }
    preview->hadjustment = hadj;
    g_object_ref(hadj);
    gtk_object_sink(GTK_OBJECT(hadj));

    g_signal_connect(hadj, "value-changed",
		     G_CALLBACK(fb_preview_adjustment_changed),
		     preview);
    need_adjust = TRUE;
  }

  if (preview->vadjustment != vadj) {
    if (preview->vadjustment) {
      g_signal_handlers_disconnect_by_func(preview->vadjustment,
					   fb_preview_adjustment_changed,
					   preview);
      g_object_unref(preview->vadjustment);
    }
    preview->vadjustment = vadj;
    g_object_ref(vadj);
    gtk_object_sink(GTK_OBJECT(vadj));

    g_signal_connect(vadj, "value-changed",
		     G_CALLBACK(fb_preview_adjustment_changed),
		     preview);
    need_adjust = TRUE;
  }

  if (need_adjust)
    fb_preview_adjustment_changed(NULL, preview);
}

static void
fb_preview_realize (GtkWidget *widget)
{
  GdkWindowAttr attributes;
  gint attributes_mask;

  g_return_if_fail(FB_IS_PREVIEW(widget));
  GTK_WIDGET_SET_FLAGS(widget, GTK_REALIZED);

  attributes.window_type = GDK_WINDOW_CHILD;
  attributes.x = widget->allocation.x;
  attributes.y = widget->allocation.y;
  attributes.width = widget->allocation.width;
  attributes.height = widget->allocation.height;
  attributes.wclass = GDK_INPUT_OUTPUT;
  attributes.visual = gtk_widget_get_visual(widget);
  attributes.colormap = gtk_widget_get_colormap(widget);
  attributes.event_mask = gtk_widget_get_events(widget) | GDK_EXPOSURE_MASK;
  attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP;

  widget->window = gdk_window_new(gtk_widget_get_parent_window(widget),
				  &attributes, attributes_mask);
  gdk_window_set_user_data(widget->window, widget);

  widget->style = gtk_style_attach(widget->style, widget->window);

  gdk_window_set_background(widget->window, &widget->style->bg[GTK_STATE_NORMAL]);
}

static void
fb_preview_unrealize (GtkWidget *widget)
{
  FBPreview* preview;
  g_return_if_fail(FB_IS_PREVIEW(widget));
  preview = FB_PREVIEW(widget);

  if (preview->fontbitmap)
    g_object_unref(preview->fontbitmap);
  preview->fontbitmap = NULL;

  if (GTK_WIDGET_CLASS(fb_preview_parent_class)->unrealize)
    (*GTK_WIDGET_CLASS(fb_preview_parent_class)->unrealize)(widget);
}

static void
fb_preview_size_request (GtkWidget *widget G_GNUC_UNUSED,
			 GtkRequisition *requisition)
{
  requisition->width = 192;
  requisition->height = 128;
}

static void
fb_preview_size_allocate (GtkWidget *widget,
			  GtkAllocation *allocation)
{
  FBPreview* preview;

  g_return_if_fail(FB_IS_PREVIEW(widget));
  g_return_if_fail(allocation != NULL);
  preview = FB_PREVIEW(widget);

  widget->allocation = *allocation;

  if (GTK_WIDGET_REALIZED(widget))
    gdk_window_move_resize(widget->window, allocation->x, allocation->y,
			   allocation->width, allocation->height);

  gtk_widget_queue_draw(GTK_WIDGET(preview));
}

#ifdef USE_FONT_BITMAP
static void
set_big_pixel(gchar* imgbits, int x, int y, int mag)
{
  int i, j;

  x *= mag;
  y *= mag;

  for (i = 0; i < mag; i++)
    for (j = 0; j < mag; j++)
      imgbits[((y + i) * mag) + ((x + j) / 8)] |= (1 << ((x + j) % 8));
}
#endif

static gboolean
fb_preview_expose (GtkWidget *widget,
		   GdkEventExpose *event)
{
  FBPreview* preview;
  gboolean need_adjust = FALSE;
  fbbook* fb;
  fbsize linenum, pos;
  int x, y;
  int mag = 2;
#ifdef USE_FONT_BITMAP
  gchar* imgbits;
  int i;
#endif
  GdkGC* gc;
  int j, k, c;
  GtkListStore* store;
  GtkTreeIter iter;
  fbsize n;
  gchar* str;

  g_return_val_if_fail(FB_IS_PREVIEW(widget), FALSE);
  preview = FB_PREVIEW(widget);
  fb = preview->fb;

  if (preview->need_refresh || !fb || !preview->fontbitmap) {
    if (!preview->refresh)
      return FALSE;

    if (fb)
      fb_book_free(fb);
    fb = preview->fb = (*preview->refresh)(preview->refresh_data);
    if (!fb)
      return FALSE;

    if (preview->combobox) {
      preview->setadjust = TRUE;
      store = GTK_LIST_STORE(gtk_combo_box_get_model(preview->combobox));
      gtk_list_store_clear(store);
      if (fb->nsections > 0) {
	gtk_widget_show(GTK_WIDGET(preview->combobox));

	if (fb->sections[0].pos_conv != 0) {
	  gtk_list_store_append(store, &iter);
	  gtk_list_store_set(store, &iter, 0, "Section 0",
			     1, PANGO_STYLE_ITALIC, -1);
	}

	for (n = 0; n < fb->nsections; n++) {
	  gtk_list_store_append(store, &iter);

	  if (fb->sections[n].title->utf8
	      && fb->sections[n].title->utf8size > 0) {
	    str = g_strndup((const char*) fb->sections[n].title->utf8,
			    fb->sections[n].title->utf8size);
	    gtk_list_store_set(store, &iter, 0, str,
			       1, PANGO_STYLE_NORMAL, -1);
	  }
	  else {
	    str = g_strdup_printf("Section %ld", (long int) n + 1);
	    gtk_list_store_set(store, &iter, 0, str,
			       1, PANGO_STYLE_ITALIC, -1);
	  }

	  g_free(str);
	}
      }
      else
	gtk_widget_hide(GTK_WIDGET(preview->combobox));

      preview->setadjust = FALSE;
      fb_preview_update_combo(preview);
    }

    if (fb->body->nlines != (preview->vadjustment->upper - 0.5)) {
      preview->vadjustment->upper = fb->body->nlines + 1;
      need_adjust = TRUE;
    }

#ifdef USE_FONT_BITMAP

    imgbits = g_new0(gchar, mag * mag * fb->fontheight * 256);

    for (i = 0; i < 256; i++)
      if (fb->fontchars[i].set)
	for (j = 0; j < fb->fontheight; j++)
	  for (k = 0; k < 8; k++)
	    if (fb->fontchars[i].bitmap[j] & (0x80 >> k))
	      set_big_pixel(imgbits, k, i * fb->fontheight + j, mag);

    if (preview->fontbitmap)
      g_object_unref(preview->fontbitmap);

    preview->fontbitmap = gdk_bitmap_create_from_data
      (GDK_DRAWABLE(widget->window), imgbits, 8 * mag, fb->fontheight * 256 * mag);

    g_free(imgbits);

#endif

    preview->need_refresh = FALSE;
  }

  if (widget->allocation.height
      != (gint) (preview->vadjustment->page_size * fb->rowheight * mag + 0.5)) {
    preview->vadjustment->page_size = ((gdouble) widget->allocation.height
				       / (fb->rowheight * mag));
    preview->vadjustment->page_increment = preview->vadjustment->page_size - 3.0;
    need_adjust = TRUE;
  }

  if (preview->vadjustment->step_increment != 2.0) {
    preview->vadjustment->step_increment = 2.0;
    need_adjust = TRUE;
  }

  if (need_adjust) {
    gtk_adjustment_changed(preview->vadjustment);
    gtk_adjustment_set_value(preview->vadjustment,
			     CLAMP(preview->vadjustment->value,
				   preview->vadjustment->lower,
				   preview->vadjustment->upper
				   - preview->vadjustment->page_size));
  }

  linenum = preview->vadjustment->value; /* rounds down */
  y = (gint) ((linenum - preview->vadjustment->value) * fb->rowheight * mag);

  while (y + fb->fontheight * mag <= event->area.y) {
    y += fb->rowheight * mag;
    linenum++;
  }

  gc = widget->style->fg_gc[widget->state];

#ifdef USE_FONT_BITMAP

  gdk_gc_set_clip_mask(gc, preview->fontbitmap);

  while (y < event->area.y + event->area.height) {
    x = 0;
    if (linenum < fb->body->nlines) {
      for (pos = fb->body->linestart[linenum];
	   pos < fb->body->lineend[linenum];
	   pos++) {
	c = fb->body->converted[pos];
	x += fb->fontchars[c].width * mag;
	gdk_gc_set_clip_origin(gc, x - 8 * mag, y - fb->fontheight * c * mag);
	gdk_draw_rectangle(GDK_DRAWABLE(widget->window), gc, TRUE,
			   x - 8 * mag, y, 8 * mag, fb->fontheight * mag);
      }
    }
    y += fb->rowheight * mag;
    linenum++;
  }

  gdk_gc_set_clip_mask(gc, NULL);

#else

  while (y < event->area.y + event->area.height) {
    x = 0;
    if (linenum < fb->body->nlines) {
      for (pos = fb->body->linestart[linenum];
	   pos < fb->body->lineend[linenum];
	   pos++) {
	c = fb->body->converted[pos];
	x += fb->fontchars[c].width * mag;
	for (j = 0; j < fb->fontheight; j++)
	  for (k = 0; k < 8; k++)
	    if (fb->fontchars[c].bitmap[j] & (1 << k))
	      gdk_draw_rectangle(GDK_DRAWABLE(widget->window), gc, TRUE,
				 x - k * mag, y + j * mag, mag, mag);
      }
    }
    y += fb->rowheight * mag;
    linenum++;
  }

#endif

  return FALSE;
}

static gboolean
fb_preview_button_press (GtkWidget *widget,
			 GdkEventButton *event G_GNUC_UNUSED)
{
  gtk_widget_grab_focus(widget);
  return FALSE;
}

static fbsize
fb_preview_get_position (FBPreview *preview)
{
  g_return_val_if_fail(preview->fb != NULL, 0);

  if ((fbsize) preview->vadjustment->value < preview->fb->body->nlines)
    return preview->fb->body->linestart[(fbsize) preview->vadjustment->value];
  else
    return preview->fb->body->linestart[preview->fb->body->nlines - 1];
}

static void
fb_preview_set_position (FBPreview *preview, fbsize pos)
{
  fbsize i = 0;

  g_return_if_fail(preview->fb != NULL);

  while (i < preview->fb->body->nlines
	 && preview->fb->body->linestart[i] < pos)
    i++;

  gtk_adjustment_set_value(preview->vadjustment, i);
}

static void
fb_preview_previous_section (FBPreview *preview)
{
  fbsize pos;
  fbsize i;

  if (!preview->fb)
    return;

  pos = fb_preview_get_position(preview);

  i = 0;
  while (i < preview->fb->nsections
	 && preview->fb->sections[i].pos_conv < pos)
    i++;
  if (i > 0)
    fb_preview_set_position(preview, preview->fb->sections[i - 1].pos_conv);
  else
    fb_preview_set_position(preview, 0);
}

static void
fb_preview_next_section (FBPreview *preview)
{
  fbsize pos;
  fbsize i;

  if (!preview->fb)
    return;

  pos = fb_preview_get_position(preview);

  i = 0;
  while (i < preview->fb->nsections
	 && preview->fb->sections[i].pos_conv <= pos)
    i++;
  if (i == preview->fb->nsections)
    return;

  fb_preview_set_position(preview, preview->fb->sections[i].pos_conv);
}

static gboolean fb_preview_key_press(GtkWidget *widget,
				     GdkEventKey *event)
{
  FBPreview* preview = FB_PREVIEW(widget);
  gdouble v;

  /* The key mapping here is intended to resemble the mapping of the
     calculator keys, where 2nd = Control or Alt, and Alpha = Shift */

  if (event->state & (GDK_CONTROL_MASK | GDK_MOD1_MASK)) {
    if (event->keyval == GDK_Left) {
      fb_preview_previous_section(preview);
      return TRUE;
    }
    else if (event->keyval == GDK_Right) {
      fb_preview_next_section(preview); 
      return TRUE;
    }
    else
      return FALSE;
  }

  v = preview->vadjustment->value;

  if (event->keyval == GDK_Page_Up
      || (event->keyval == GDK_Up && (event->state & GDK_SHIFT_MASK))
      || event->keyval == GDK_Left)
    v -= preview->vadjustment->page_increment;
  else if (event->keyval == GDK_Page_Down
	   || (event->keyval == GDK_Down && (event->state & GDK_SHIFT_MASK))
	   || event->keyval == GDK_Right)
    v += preview->vadjustment->page_increment;
  else if (event->keyval == GDK_Up)
    v -= preview->vadjustment->step_increment;
  else if (event->keyval == GDK_Down)
    v += preview->vadjustment->step_increment;
  else if (event->keyval == GDK_Home)
    v = preview->vadjustment->lower;
  else if (event->keyval == GDK_End)
    v = preview->vadjustment->upper;
  else
    return FALSE;

  v = CLAMP(v, preview->vadjustment->lower,
	    preview->vadjustment->upper - preview->vadjustment->page_size);
  gtk_adjustment_set_value(preview->vadjustment, v);
  return TRUE;
}

static void
fb_preview_update_combo (FBPreview *preview)
{
  fbsize pos, i;

  if (preview->combobox && preview->fb && !preview->setadjust) {
    pos = fb_preview_get_position(preview);
    i = 0;
    while (i < preview->fb->nsections
	   && preview->fb->sections[i].pos_conv <= pos)
      i++;

    if (preview->fb->nsections > 0
	&& preview->fb->sections[0].pos_conv == 0)
      i--;

    preview->setadjust = TRUE;
    gtk_combo_box_set_active(preview->combobox, i);
    preview->setadjust = FALSE;
  }
}

static void
fb_preview_adjustment_changed (GtkAdjustment *adj,
			       gpointer data)
{
  FBPreview* preview = data;

  if (adj == preview->vadjustment) {
    gtk_widget_queue_draw(GTK_WIDGET(preview));
    fb_preview_update_combo(preview);
  }
}


GtkWidget *
fb_preview_new (FBPreviewRefreshFunc fnc,
		gpointer data)
{
  FBPreview *preview;

  preview = g_object_new(FB_TYPE_PREVIEW, NULL);
  preview->refresh = fnc;
  preview->refresh_data = data;
  return GTK_WIDGET(preview);
}

void
fb_preview_refresh (FBPreview* preview)
{
  g_return_if_fail(FB_IS_PREVIEW(preview));
  preview->need_refresh = TRUE;
  gtk_widget_queue_draw(GTK_WIDGET(preview));
}


/**************** Complete preview window ****************/

static void
combo_changed(GtkComboBox* combo, gpointer data)
{
  FBPreview *preview = data;
  gint n;

  if (preview->fb && preview->vadjustment && !preview->setadjust) {
    n = gtk_combo_box_get_active(combo);

    preview->setadjust = TRUE;

    if (preview->fb->nsections > 0
	&& preview->fb->sections[0].pos_conv != 0)
      n--;

    if (n >= 0 && n < (gint) preview->fb->nsections)
      fb_preview_set_position(preview, preview->fb->sections[n].pos_conv);
    else if (n == -1)
      fb_preview_set_position(preview, (fbsize) 0);

    preview->setadjust = FALSE;
  }
}

GtkWidget *
fb_preview_box_new (FBPreviewRefreshFunc fnc,
		    gpointer data,
		    GtkWidget **preview_widget)
{
  GtkWidget *vbox, *sw, *preview, *align, *combo;
  GtkListStore *store;
  GtkCellRenderer *renderer;

  vbox = gtk_vbox_new(FALSE, 0);
  {
    align = gtk_alignment_new(0.5, 0.5, 1.0, 1.0);
    gtk_alignment_set_padding(GTK_ALIGNMENT(align), 6, 6, 6, 6);
    {
      store = gtk_list_store_new(2, G_TYPE_STRING, PANGO_TYPE_STYLE);
      combo = gtk_combo_box_new_with_model(GTK_TREE_MODEL(store));

      renderer = gtk_cell_renderer_text_new();
      g_object_set(renderer, "ellipsize", PANGO_ELLIPSIZE_END, NULL);
      gtk_cell_layout_pack_start(GTK_CELL_LAYOUT(combo), renderer, TRUE);
      gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(combo), renderer,
				    "text", 0);
      gtk_cell_layout_add_attribute(GTK_CELL_LAYOUT(combo), renderer,
				    "style", 1);

      gtk_container_add(GTK_CONTAINER(align), combo);
    }
    gtk_box_pack_start(GTK_BOX(vbox), align, FALSE, FALSE, 0);
    gtk_widget_show(align);

    sw = gtk_scrolled_window_new(NULL, NULL);
    gtk_scrolled_window_set_policy(GTK_SCROLLED_WINDOW(sw),
				   GTK_POLICY_NEVER,
				   GTK_POLICY_ALWAYS);
    gtk_scrolled_window_set_shadow_type(GTK_SCROLLED_WINDOW(sw),
					GTK_SHADOW_IN);
    {
      preview = fb_preview_new(fnc, data);
      gtk_container_add(GTK_CONTAINER(sw), preview);
      gtk_widget_show(preview);
    }
    gtk_box_pack_start(GTK_BOX(vbox), sw, TRUE, TRUE, 0);
    gtk_widget_show(sw);
  }

  FB_PREVIEW(preview)->combobox = GTK_COMBO_BOX(combo);

  g_signal_connect(combo, "changed", G_CALLBACK(combo_changed), preview);

  if (preview_widget)
    *preview_widget = preview;

  return vbox;
}
