/*
 * Flashbook GUI
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <gtk/gtktextbuffer.h>
#include "fbtextbuffer.h"

struct _FBUndoAction {
  enum {
    INSERT_TEXT,
    DELETE_RANGE,
    APPLY_TAG,
    REMOVE_TAG,
    BEGIN_USER_ACTION,
    END_USER_ACTION
  } type;

  unsigned modflag : 1;
  unsigned fresh : 1;

  gint start_offset;
  gint end_offset;
  GtkTextTag *tag;
  gchar *text;

  FBUndoAction* next;
};

static void fb_text_buffer_finalize(GObject *obj);

static void fb_text_buffer_insert_text(GtkTextBuffer *buffer,
					    GtkTextIter   *iter,
					    const gchar   *text,
					    gint           len);
static void fb_text_buffer_delete_range(GtkTextBuffer *buffer,
					     GtkTextIter   *start,
					     GtkTextIter   *end);
static void fb_text_buffer_modified_changed(GtkTextBuffer *buffer);
static void fb_text_buffer_apply_tag(GtkTextBuffer     *buffer,
					  GtkTextTag        *tag,
					  const GtkTextIter *start,
					  const GtkTextIter *end);
static void fb_text_buffer_remove_tag(GtkTextBuffer     *buffer,
					   GtkTextTag        *tag,
					   const GtkTextIter *start,
					   const GtkTextIter *end);
static void fb_text_buffer_begin_user_action(GtkTextBuffer *buffer);
static void fb_text_buffer_end_user_action(GtkTextBuffer *buffer);

static void fb_text_buffer_real_clear_history(FBTextBuffer *utb);
static void fb_text_buffer_real_undo(FBTextBuffer *utb);
static void fb_text_buffer_real_redo(FBTextBuffer *utb);

enum {
  UNDO_STATE_CHANGED,
  CLEAR_HISTORY,
  UNDO,
  REDO,
  LAST_SIGNAL
};

static guint signals[LAST_SIGNAL] = { 0 };

G_DEFINE_TYPE(FBTextBuffer, fb_text_buffer, GTK_TYPE_TEXT_BUFFER);

static void
fb_text_buffer_class_init (FBTextBufferClass *class)
{
  GObjectClass *o_class;
  GtkTextBufferClass *tb_class;

  fb_text_buffer_parent_class = g_type_class_peek_parent(class);

  o_class = G_OBJECT_CLASS(class);
  o_class->finalize = fb_text_buffer_finalize;

  tb_class = GTK_TEXT_BUFFER_CLASS(class);

  tb_class->insert_text = fb_text_buffer_insert_text;
  tb_class->delete_range = fb_text_buffer_delete_range;
  tb_class->modified_changed = fb_text_buffer_modified_changed;
  tb_class->apply_tag = fb_text_buffer_apply_tag;
  tb_class->remove_tag = fb_text_buffer_remove_tag;
  tb_class->begin_user_action = fb_text_buffer_begin_user_action;
  tb_class->end_user_action = fb_text_buffer_end_user_action;

  class->clear_history = fb_text_buffer_real_clear_history;
  class->undo = fb_text_buffer_real_undo;
  class->redo = fb_text_buffer_real_redo;

  signals[UNDO_STATE_CHANGED] =
    g_signal_new("undo-state-changed",
		 G_OBJECT_CLASS_TYPE(o_class),
		 G_SIGNAL_RUN_LAST,
		 G_STRUCT_OFFSET(FBTextBufferClass, undo_state_changed),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__VOID,
		 G_TYPE_NONE,
		 0);

  signals[CLEAR_HISTORY] =
    g_signal_new("clear-history",
		 G_OBJECT_CLASS_TYPE(o_class),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(FBTextBufferClass, clear_history),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__VOID,
		 G_TYPE_NONE,
		 0);

  signals[UNDO] =
    g_signal_new("undo",
		 G_OBJECT_CLASS_TYPE(o_class),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(FBTextBufferClass, undo),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__VOID,
		 G_TYPE_NONE,
		 0);

  signals[REDO] =
    g_signal_new("redo",
		 G_OBJECT_CLASS_TYPE(o_class),
		 G_SIGNAL_RUN_LAST | G_SIGNAL_ACTION,
		 G_STRUCT_OFFSET(FBTextBufferClass, redo),
		 NULL, NULL,
		 g_cclosure_marshal_VOID__VOID,
		 G_TYPE_NONE,
		 0);
}

static void
fb_text_buffer_init (FBTextBuffer *self)
{
  self->max_chars_per_undo = 50;
  self->undo_stack = NULL;
  self->redo_stack = NULL;
  self->recording = TRUE;
}

/**************** Undo Stack ****************/

static void
free_undo_stack (FBUndoAction *action)
{
  FBUndoAction* next;

  while (action) {
    if (action->tag)
      g_object_unref(action->tag);
    g_free(action->text);
    next = action->next;
    g_free(action);
    action = next;
  }
}

static void
push_undo_action (FBTextBuffer *utb,
		  gint          type,
		  gint          start_offset,
		  gint          end_offset,
		  GtkTextTag   *tag,
		  gchar        *text)
{
  FBUndoAction *act = g_new(FBUndoAction, 1);

  act->type = type;
  act->start_offset = start_offset;
  act->end_offset = end_offset;
  act->tag = tag;
  act->text = text;
  act->fresh = 1;
  act->modflag = GTK_TEXT_BUFFER(utb)->modified;
  act->next = utb->undo_stack;
  utb->undo_stack = act;

  if (type != BEGIN_USER_ACTION) {
    if (act->next)
      act->next->fresh = 0;
    if (act->next && act->next->next)
      act->next->next->fresh = 0;
  }

  if (utb->redo_stack) {
    free_undo_stack(utb->redo_stack);
    utb->redo_stack = NULL;
  }

  g_signal_emit(utb, signals[UNDO_STATE_CHANGED], 0, NULL);
}

static void
execute_undo_action(FBTextBuffer  *utb,
		    FBUndoAction **fromstack,
		    FBUndoAction **tostack)
{
  GtkTextBuffer *buffer;
  GtkTextIter start, end;
  FBUndoAction *act;

  buffer = GTK_TEXT_BUFFER(utb);
  act = *fromstack;

  if (!act || act->type == BEGIN_USER_ACTION)
    return;

  *fromstack = act->next;
  act->next = *tostack;
  *tostack = act;
  act->fresh = 0;

  if (act->modflag != buffer->modified) {
    gtk_text_buffer_set_modified(buffer, (act->modflag ? TRUE : FALSE));
    act->modflag ^= 1;
  }

  switch (act->type) {
  case INSERT_TEXT:
    g_assert(act->tag == NULL);
    g_assert(act->text == NULL);
    g_assert(act->end_offset >= act->start_offset);

    act->type = DELETE_RANGE;
    gtk_text_buffer_get_iter_at_offset(buffer, &start, act->start_offset);
    gtk_text_buffer_get_iter_at_offset(buffer, &end, act->end_offset);
    gtk_text_buffer_select_range(buffer, &start, &start);
    act->text = gtk_text_buffer_get_slice(buffer, &start, &end, TRUE);

    gtk_text_buffer_delete(buffer, &start, &end);
    break;

  case DELETE_RANGE:
    g_assert(act->tag == NULL);
    g_assert(act->text != NULL);

    act->type = INSERT_TEXT;

    gtk_text_buffer_get_iter_at_offset(buffer, &start, act->start_offset);
    gtk_text_buffer_select_range(buffer, &start, &start);
    gtk_text_buffer_insert(buffer, &start, act->text, -1);

    g_free(act->text);
    act->text = NULL;
    break;

  case APPLY_TAG:
    g_assert(act->tag != NULL);
    g_assert(act->text == NULL);

    act->type = REMOVE_TAG;

    gtk_text_buffer_get_iter_at_offset(buffer, &start, act->start_offset);
    gtk_text_buffer_get_iter_at_offset(buffer, &end, act->end_offset);
    gtk_text_buffer_remove_tag(buffer, act->tag, &start, &end);
    break;

  case REMOVE_TAG:
    g_assert(act->tag != NULL);
    g_assert(act->text == NULL);

    act->type = APPLY_TAG;

    gtk_text_buffer_get_iter_at_offset(buffer, &start, act->start_offset);
    gtk_text_buffer_get_iter_at_offset(buffer, &end, act->end_offset);
    gtk_text_buffer_apply_tag(buffer, act->tag, &start, &end);
    break;

  case END_USER_ACTION:
    g_assert(act->tag == NULL);
    g_assert(act->text == NULL);

    act->type = BEGIN_USER_ACTION;

    gtk_text_buffer_begin_user_action(buffer);

    while (*fromstack && (*fromstack)->type != BEGIN_USER_ACTION)
      execute_undo_action(utb, fromstack, tostack);

    if (*fromstack) {
      act = *fromstack;
      act->type = END_USER_ACTION;
      *fromstack = act->next;
      act->next = *tostack;
      *tostack = act;
      act->fresh = 0;

      if (act->modflag != buffer->modified) {
	gtk_text_buffer_set_modified(buffer, (act->modflag ? TRUE : FALSE));
	act->modflag ^= 1;
      }

      gtk_text_buffer_get_iter_at_offset(buffer, &start, act->start_offset);
      gtk_text_buffer_place_cursor(buffer, &start);
    }
    else
      g_critical("Found END_USER_ACTION without corresponding BEGIN_USER_ACTION");

    gtk_text_buffer_end_user_action(buffer);
    break;

  case BEGIN_USER_ACTION:
  default:
    g_assert_not_reached();
  }
}


/**************** GtkTextBuffer virtual functions ****************/

static void
fb_text_buffer_finalize (GObject *obj)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(obj);

  free_undo_stack(utb->undo_stack);
  free_undo_stack(utb->redo_stack);
  utb->undo_stack = utb->redo_stack = NULL;

  if (G_OBJECT_CLASS(fb_text_buffer_parent_class)->finalize)
    (*G_OBJECT_CLASS(fb_text_buffer_parent_class)->finalize)(obj);
}

static gboolean
can_concatenate (FBTextBuffer *utb,
		 int           undotype)
{
  if (!utb->redo_stack
      && utb->undo_stack
      && utb->undo_stack->type == BEGIN_USER_ACTION
      && utb->undo_stack->next
      && utb->undo_stack->next->fresh
      && utb->undo_stack->next->type == END_USER_ACTION
      && utb->undo_stack->next->next
      && (int) utb->undo_stack->next->next->type == undotype
      && utb->undo_stack->next->next->next
      && utb->undo_stack->next->next->next->type == BEGIN_USER_ACTION
      && (utb->undo_stack->next->next->end_offset
	  - utb->undo_stack->next->next->start_offset
	  < utb->max_chars_per_undo))
    return TRUE;
  else
    return FALSE;
}

static void
fb_text_buffer_insert_text (GtkTextBuffer *buffer,
			    GtkTextIter   *iter,
			    const gchar   *text,
			    gint           len)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);
  FBUndoAction *act;
  gint start_offset, end_offset;

  if (utb->recording) {
    start_offset = gtk_text_iter_get_offset(iter);
    end_offset = start_offset + g_utf8_strlen(text, len);

    if (end_offset == start_offset + 1
	&& text[0] != '\n'
	&& can_concatenate(utb, INSERT_TEXT)
	&& utb->undo_stack->next->next->end_offset == start_offset) {

      act = utb->undo_stack->next->next;
      utb->undo_stack->next->next = NULL;
      free_undo_stack(utb->undo_stack);
      utb->undo_stack = act;

      act->end_offset = end_offset;
    }
    else {
      push_undo_action(utb, INSERT_TEXT, start_offset, end_offset, NULL, NULL);
    }
  }

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->insert_text)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->insert_text)
      (buffer, iter, text, len);
}

static void
fb_text_buffer_delete_range (GtkTextBuffer *buffer,
			     GtkTextIter   *start,
			     GtkTextIter   *end)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);
  FBUndoAction *act;
  gint start_offset, end_offset;
  gchar *text, *p;

  if (utb->recording) {
    /* remove all tags first (so that they will be recorded in the
       history) */
    gtk_text_buffer_remove_all_tags(buffer, start, end);

    start_offset = gtk_text_iter_get_offset(start);
    end_offset = gtk_text_iter_get_offset(end);
    text = gtk_text_buffer_get_slice(buffer, start, end, TRUE);

    if (end_offset == start_offset + 1
	&& text[0] != '\n'
	&& can_concatenate(utb, DELETE_RANGE)) {
      act = utb->undo_stack->next->next;

      if (act->start_offset == start_offset) {

	/* deleting forwards */

	utb->undo_stack->next->next = NULL;
	free_undo_stack(utb->undo_stack);
	utb->undo_stack = act;

	act->end_offset += end_offset - start_offset;
	p = g_strconcat(act->text, text, NULL);
	g_free(act->text);
	g_free(text);
	act->text = p;
      }
      else if (act->start_offset == end_offset) {

	/* deleting backwards */

	utb->undo_stack->next->next = NULL;
	free_undo_stack(utb->undo_stack);
	utb->undo_stack = act;

	act->start_offset = start_offset;
	p = g_strconcat(text, act->text, NULL);
	g_free(act->text);
	g_free(text);
	act->text = p;
      }
      else {
	push_undo_action(utb, DELETE_RANGE, start_offset, end_offset, NULL, text);
      }
    }
    else {
      push_undo_action(utb, DELETE_RANGE, start_offset, end_offset, NULL, text);
    }
  }

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->delete_range)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->delete_range)
      (buffer, start, end);
}

static void
fb_text_buffer_modified_changed (GtkTextBuffer *buffer)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);
  FBUndoAction *act;

  if (utb->recording) {
    if (!buffer->modified) {
      for (act = utb->undo_stack; act; act = act->next) {
	act->fresh = 0;
	act->modflag = 1;
      }
      for (act = utb->redo_stack; act; act = act->next) {
	act->fresh = 0;
	act->modflag = 1;
      }
    }
  }

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->modified_changed)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->modified_changed)
      (buffer);
}

static void
push_tag_pattern (FBTextBuffer      *utb,
		  gboolean           applying_tag,
		  GtkTextTag        *tag,
		  const GtkTextIter *start,
		  const GtkTextIter *end)
{
  GtkTextIter iter;
  gint offset, prev_offset;
  gboolean on;
  gint end_offset;

  end_offset = gtk_text_iter_get_offset(end);

  iter = *start;
  on = gtk_text_iter_has_tag(&iter, tag);
  offset = gtk_text_iter_get_offset(&iter);

  while (offset < end_offset) {
    prev_offset = offset;

    if (gtk_text_iter_forward_to_tag_toggle(&iter, tag)) {
      offset = gtk_text_iter_get_offset(&iter);
      if (offset > end_offset)
	offset = end_offset;
    }
    else {
      offset = end_offset;
    }

    if (on && !applying_tag) {
      g_object_ref(tag);
      push_undo_action(utb, REMOVE_TAG, prev_offset, offset, tag, NULL);
    }
    else if (!on && applying_tag) {
      g_object_ref(tag);
      push_undo_action(utb, APPLY_TAG, prev_offset, offset, tag, NULL);
    }
    on = !on;
  }
}

static void
fb_text_buffer_apply_tag (GtkTextBuffer     *buffer,
			  GtkTextTag        *tag,
			  const GtkTextIter *start,
			  const GtkTextIter *end)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);

  if (utb->recording)
    push_tag_pattern(utb, TRUE, tag, start, end);

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->apply_tag)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->apply_tag)
      (buffer, tag, start, end);
}

static void
fb_text_buffer_remove_tag (GtkTextBuffer     *buffer,
			   GtkTextTag        *tag,
			   const GtkTextIter *start,
			   const GtkTextIter *end)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);

  if (utb->recording)
    push_tag_pattern(utb, FALSE, tag, start, end);

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->remove_tag)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->remove_tag)
      (buffer, tag, start, end);
}

static void
fb_text_buffer_begin_user_action (GtkTextBuffer *buffer)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);
  GtkTextIter cursor;
  gint cursor_offset;

  if (utb->recording) {
    gtk_text_buffer_get_iter_at_mark(buffer, &cursor,
				     gtk_text_buffer_get_insert(buffer));
    cursor_offset = gtk_text_iter_get_offset(&cursor);
    push_undo_action(utb, BEGIN_USER_ACTION, cursor_offset,
		     cursor_offset, NULL, NULL);
  }

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->begin_user_action)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->begin_user_action)
      (buffer);
}

static void
fb_text_buffer_end_user_action (GtkTextBuffer *buffer)
{
  FBTextBuffer *utb = FB_TEXT_BUFFER(buffer);
  GtkTextIter cursor;
  gint cursor_offset;

  if (utb->recording) {
    gtk_text_buffer_get_iter_at_mark(buffer, &cursor,
				     gtk_text_buffer_get_insert(buffer));
    cursor_offset = gtk_text_iter_get_offset(&cursor);
    push_undo_action(utb, END_USER_ACTION, cursor_offset,
		     cursor_offset, NULL, NULL);
  }

  if (GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->end_user_action)
    (*GTK_TEXT_BUFFER_CLASS(fb_text_buffer_parent_class)->end_user_action)
      (buffer);
}


/**************** New virtual functions ****************/

static void
fb_text_buffer_real_clear_history (FBTextBuffer *utb)
{
  free_undo_stack(utb->undo_stack);
  free_undo_stack(utb->redo_stack);
  utb->undo_stack = utb->redo_stack = NULL;

  g_signal_emit(utb, signals[UNDO_STATE_CHANGED], 0, NULL);
}

static void
fb_text_buffer_real_undo (FBTextBuffer *utb)
{
  if (utb->undo_stack && utb->undo_stack->type != BEGIN_USER_ACTION
      && utb->recording) {
    utb->recording = FALSE;
    execute_undo_action(utb, &utb->undo_stack, &utb->redo_stack);
    utb->recording = TRUE;
    g_signal_emit(utb, signals[UNDO_STATE_CHANGED], 0, NULL);
  }
}

static void
fb_text_buffer_real_redo (FBTextBuffer *utb)
{
  if (utb->redo_stack && utb->redo_stack->type != BEGIN_USER_ACTION
      && utb->recording) {
    utb->recording = FALSE;
    execute_undo_action(utb, &utb->redo_stack, &utb->undo_stack);
    utb->recording = TRUE;
    g_signal_emit(utb, signals[UNDO_STATE_CHANGED], 0, NULL);
  }
}


/**************** API ****************/

GtkTextBuffer *
fb_text_buffer_new (void)
{
  FBTextBuffer *utb;
  utb = g_object_new(FB_TYPE_TEXT_BUFFER, NULL);
  return GTK_TEXT_BUFFER(utb);
}

void
fb_text_buffer_set_recording (FBTextBuffer* buffer,
			      gboolean      recording)
{
  g_return_if_fail(FB_IS_TEXT_BUFFER(buffer));

  if ((buffer->recording && recording) || (!buffer->recording && !recording))
    return;

  buffer->recording = recording;
  fb_text_buffer_clear_history(buffer);
}

void
fb_text_buffer_clear_history (FBTextBuffer* buffer)
{
  g_return_if_fail(FB_IS_TEXT_BUFFER(buffer));
  g_signal_emit(buffer, signals[CLEAR_HISTORY], 0, NULL);
}

gboolean
fb_text_buffer_can_undo (FBTextBuffer* buffer)
{
  g_return_val_if_fail(FB_IS_TEXT_BUFFER(buffer), FALSE);

  if (!buffer->recording || !buffer->undo_stack
      || buffer->undo_stack->type == BEGIN_USER_ACTION)
    return FALSE;
  else
    return TRUE;
}

gboolean
fb_text_buffer_can_redo (FBTextBuffer* buffer)
{
  g_return_val_if_fail(FB_IS_TEXT_BUFFER(buffer), FALSE);

  if (!buffer->recording || !buffer->redo_stack
      || buffer->redo_stack->type == BEGIN_USER_ACTION)
    return FALSE;
  else
    return TRUE;
}

void
fb_text_buffer_undo (FBTextBuffer* buffer)
{
  g_return_if_fail(FB_IS_TEXT_BUFFER(buffer));
  g_signal_emit(buffer, signals[UNDO], 0, NULL);
}

void
fb_text_buffer_redo (FBTextBuffer* buffer)
{
  g_return_if_fail(FB_IS_TEXT_BUFFER(buffer));
  g_signal_emit(buffer, signals[REDO], 0, NULL);
}
