/*
 * Flashbook GUI
 * Copyright (c) 2008-2010 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <gtk/gtk.h>
#include <glib/gstdio.h>
#include <fb.h>
#include "fbgui.h"
#include "msgbox.h"

#ifdef G_OS_WIN32
# define WIN32_LEAN_AND_MEAN
# include <windows.h>
# include <shellapi.h>
#endif

#if defined(HAVE_COCOA_COCOA_H) && defined(GDK_WINDOWING_QUARTZ)
# include <gdk/gdkquartz.h>
# import <Cocoa/Cocoa.h>
#endif

void fix_main_args(int* argc G_GNUC_UNUSED,
		   char*** argv G_GNUC_UNUSED)
{
#ifdef G_OS_WIN32
  wchar_t** wargv;
  int i;

  if (G_WIN32_HAVE_WIDECHAR_API()) {
    wargv = CommandLineToArgvW(GetCommandLineW(), &i);
    if (i != *argc) {
      /* could this actually happen? */
      *argc = i;
      *argv = g_new(char*, i);
    }

    for (i = 0; i < *argc; i++) {
      (*argv)[i] = g_utf16_to_utf8(wargv[i], -1, 0, 0, 0);
    }
  }
  else {
    for (i = 0; i < *argc; i++) {
      (*argv)[i] = g_locale_to_utf8((*argv)[i], -1, 0, 0, 0);
    }
  }
#endif
}

static char* get_menurc_filename(const char* progname, gboolean create)
{
  const char* confdir;
  char *subdir, *file, *path;

  confdir = g_get_user_config_dir();
  if (create)
    g_mkdir(confdir, 0755);

  subdir = g_build_filename(confdir, "flashbook", NULL);
  if (create)
    g_mkdir(subdir, 0755);

  file = g_strconcat(progname, ".menurc", NULL);
  path = g_build_filename(subdir, file, NULL);
  g_free(subdir);
  g_free(file);

  return path;
}

void load_menurc_file(const char* progname)
{
  char *name = get_menurc_filename(progname, FALSE);
  gtk_accel_map_load(name);
  g_free(name);
}

static void check_map_changed(gpointer data,
			      const gchar* path G_GNUC_UNUSED,
			      guint key G_GNUC_UNUSED,
			      GdkModifierType mods G_GNUC_UNUSED,
			      gboolean changed)
{
  gboolean *flag = data;
  if (changed)
    *flag = TRUE;
}

void save_menurc_file(const char* progname)
{
  gboolean changed = FALSE;
  char* name;

  gtk_accel_map_foreach(&changed, check_map_changed);
  if (!changed)
    return;

  name = get_menurc_filename(progname, TRUE);
  gtk_accel_map_save(name);
  g_free(name);
}

void show_about_dialog(GtkWindow* window)
{
  static const char copyright[] =
    "Copyright \302\251 2008" "\342\200\223" "2010"
    " Benjamin Moody <floppusmaximus@users.sf.net>";
  static const char comments[] =
    "This program is free software.  You are welcome to modify and/or"
    " redistribute it under certain conditions; see the license for more"
    " details.";
  static const char website[] = PACKAGE_WEBSITE;
  static const char version[] = PACKAGE_VERSION;
  static char* license = NULL;
  char* fname;

  if (!license) {
    fname = fb_find_common_file(NULL, "COPYING");
    if (!fname)
      fname = fb_find_common_file(NULL, "COPYING.txt");
    if (fname)
      g_file_get_contents(fname, &license, NULL, NULL);
    fbfree(fname);
  }

  gtk_show_about_dialog(window,
			"copyright", copyright,
			"comments", comments,
			"license", license,
			"website", website,
			"version", version,
			NULL);
}

char* filename_to_display(const char* fname)
{
  char* bname = g_path_get_basename(fname);
  char* utf8name = g_filename_to_utf8(bname, -1, NULL, NULL, NULL);
  g_free(bname);
  if (utf8name)
    return utf8name;
  else
    return g_strdup("<unknown filename>");
}

void display_io_error(GtkWidget* parent, const gchar* fname,
		      gboolean reading, int num)
{
  gchar *name;

  name = filename_to_display(fname);

  if (reading)
    messagebox11(parent, GTK_MESSAGE_ERROR,
		 "Could not open the file %s", name,
		 "An error occurred when trying to read the file: %s.",
		 g_strerror(num));
  else
    messagebox11(parent, GTK_MESSAGE_ERROR,
		 "Could not save the file %s", name,
		 "An error occurred when trying to write the file: %s.",
		 g_strerror(num));
  g_free(name);
}

void set_window_title(GtkWindow* window, const char* title,
		      const char* fname G_GNUC_UNUSED,
		      gboolean modified)
{
#if defined(HAVE_COCOA_COCOA_H) && defined(GDK_WINDOWING_QUARTZ)
  NSWindow *nswin;
  char* utf8name;

  if (!window || !GTK_WIDGET(window)->window) {
    gtk_window_set_title(window, title);
    return;
  }

  nswin = gdk_quartz_window_get_nswindow(GTK_WIDGET(window)->window);

  [nswin setDocumentEdited: (modified ? YES : NO)];
  [nswin setTitle: [NSString stringWithUTF8String: title]];

  if (fname) {
    utf8name = g_filename_to_utf8(fname, -1, NULL, NULL, NULL);
    [nswin setRepresentedFilename: [NSString stringWithUTF8String: utf8name]];
    g_free(utf8name);
  }

#else
  char* p = g_strdup_printf("%s%s - %s",
			    modified ? "*" : "",
			    title, g_get_application_name());
  gtk_window_set_title(window, p);
  g_free(p);
#endif
}
