/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include "fb.h"

#if !defined(memmove) && !defined(HAVE_MEMMOVE) && defined(HAVE_BCOPY)
# define memmove(d,s,l) bcopy((s),(d),(l))
#endif

/**************** Character lookup tables ****************/

/* Look up location of a character in a chartable. */
static inline int get_existing_index(const fbchartable* tbl, fbunichar c)
{
  int x, y, z;

  x = -1;
  y = tbl->nchars;
  while (x < y - 1) {
    z = (x + y) / 2;
    if (tbl->chars[z] == c)
      return z;
    else if (tbl->chars[z] < c)
      x = z;
    else
      y = z;
  }

  return -1;
}

/* Add a character to a chartable. */
static inline int insert_char(fbchartable* tbl, fbunichar c, int pos)
{
  if (tbl->nchars >= FB_MAX_CHARS) {
    tbl->overflow = 1;
    return -1;
  }

  if (pos < tbl->nchars) {
    memmove(tbl->chars + pos + 1, tbl->chars + pos,
	    (tbl->nchars - pos) * sizeof(fbunichar));
    memmove(tbl->values + pos + 1, tbl->values + pos,
	    (tbl->nchars - pos) * sizeof(fbsize));
  }
  tbl->chars[pos] = c;
  tbl->values[pos] = 0;
  tbl->nchars++;
  return pos;
}

/* Look up location of a character in a chartable, or add one. */
static inline int get_new_index(fbchartable* tbl, fbunichar c)
{
  int x, y, z;

  x = -1;
  y = tbl->nchars;
  while (x < y - 1) {
    z = (x + y) / 2;
    if (tbl->chars[z] == c)
      return z;
    else if (tbl->chars[z] < c)
      x = z;
    else
      y = z;
  }
  return insert_char(tbl, c, y);
}

/* Get the value of a character in a chartable. */
fbsize fb_chartable_get_value(const fbchartable* tbl, fbunichar c)
{
  int i = get_existing_index(tbl, c);
  if (i == -1)
    return 0;
  else
    return tbl->values[i];
}

/* Set the value of a character in a chartable. */
void fb_chartable_set_value(fbchartable* tbl, fbunichar c, fbsize value)
{
  int i = (value ? get_new_index(tbl, c) : get_existing_index(tbl, c));
  if (i != -1)
    tbl->values[i] = value;
}

/* Increment the value of a character in a chartable. */
void fb_chartable_inc_value(fbchartable* tbl, fbunichar c)
{
  int i = get_new_index(tbl, c);
  if (i != -1)
    tbl->values[i]++;
}

/* Merge two chartables by adding values. */
void fb_chartable_add_table(fbchartable* dest, const fbchartable* src)
{
  int i, j;

  if (src->overflow) {
    dest->overflow = 1;
    return;
  }

  for (i = 0; i < src->nchars; i++) {
    j = get_new_index(dest, src->chars[i]);
    if (j != -1)
      dest->values[j] += src->values[i];
    else
      return;
  }
}
