/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include "fb.h"

#if !defined(memmove) && !defined(HAVE_MEMMOVE) && defined(HAVE_BCOPY)
# define memmove(d,s,l) bcopy((s),(d),(l))
#endif

/**************** Character set conversion ****************/

static const fbunichar base_encoding_ti83p_large[256] =
  { 0,       0x1D45B, 0,       0,        0,       0x25B8, 0x2B06,   0x2B07,
    0x222B,  0xD7,    0,       0,        0xB7,    0,      0,        0x1D405,
    0x221A,  0,       0xB2,    0x2220,   0xB0,    0x2B3,  0,        0x2264,
    0x2260,  0x2265,  0x207B,  0,        0x2192,  0x23E8, 0x2191,   0x2193,
    ' ',     '!',     '"',     '#',      0x2074,  '%',    '&',      '\'',
    '(',     ')',     '*',     '+',      ',',     '-',    '.',      '/',
    '0',     '1',     '2',     '3',      '4',     '5',    '6',      '7',
    '8',     '9',     ':',     ';',      '<',     '=',    '>',      '?',
      
    '@',     'A',     'B',     'C',      'D',     'E',    'F',      'G',
    'H',     'I',     'J',     'K',      'L',     'M',    'N',      'O',
    'P',     'Q',     'R',     'S',      'T',     'U',    'V',      'W',
    'X',     'Y',     'Z',     0x3B8,    '\\',    ']',    '^',      '_',
    '`',     'a',     'b',     'c',      'd',     'e',    'f',      'g',
    'h',     'i',     'j',     'k',      'l',     'm',    'n',      'o',
    'p',     'q',     'r',     's',      't',     'u',    'v',      'w',
    'x',     'y',     'z',     '{',      '|',     '}',    '~',      0,   

    0x2080,  0x2081,  0x2082,  0x2083,   0x2084,  0x2085, 0x2086,   0x2087,
    0x2088,  0x2089,  0xC1,    0xC0,     0xC2,    0xC4,   0xE1,     0xE0,
    0xE2,    0xE4,    0xC9,    0xC8,     0xCA,    0xCB,   0xE9,     0xE8,
    0xEA,    0xEB,    0xCD,    0xCC,     0xCE,    0xCF,   0xED,     0xEC,
    0xEE,    0xEF,    0xD3,    0xD2,     0xD4,    0xD6,   0xF3,     0xF2,
    0xF4,    0xF6,    0xDA,    0xD9,     0xDB,    0xDC,   0xFA,     0xF9,
    0xFB,    0xFC,    0xC7,    0xE7,     0xD1,    0xF1,   0x2CA,    0x2CB,
    0xA8,    0xBF,    0xA1,    0x3B1,    0x3B2,   0x3B3,  0x2206,   0x3B4,

    0x3F5,   '[',     0x3BB,   0x3BC,    0x3C0,   0x3C1,  0x3A3,    0x3C3,
    0x3C4,   0x3D5,   0x3A9,   0,        0,       0,      0x2026,   0x25C2,
    0x25AA,  0,       0x2010,  0,        0,       0xB3,   '\n',     0,
    0,       0x3C7,   0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0 };

static const fbunichar base_encoding_ti83p_small[256] =
  { 0,       0x1D45B, 0,       0,        0,       0x25B8, 0x2007,   0x2B07,
    0x222B,  0xD7,    0,       0,        0xB7,    0,      0,        0x1D405,
    0x221A,  0,       0xB2,    0x2220,   0xB0,    0x2B3,  0,        0x2264,
    0x2260,  0x2265,  0x207B,  0,        0x2192,  0x23E8, 0x2191,   0x2193,
    ' ',     '!',     '"',     '#',      '$',     '%',    '&',      '\'',
    '(',     ')',     '*',     '+',      ',',     '-',    '.',      '/',
    '0',     '1',     '2',     '3',      '4',     '5',    '6',      '7',
    '8',     '9',     ':',     ';',      '<',     '=',    '>',      '?',

    '@',     'A',     'B',     'C',      'D',     'E',    'F',      'G',
    'H',     'I',     'J',     'K',      'L',     'M',    'N',      'O',
    'P',     'Q',     'R',     'S',      'T',     'U',    'V',      'W',
    'X',     'Y',     'Z',     0x3B8,    '\\',    ']',    '^',      '_',
    '`',     'a',     'b',     'c',      'd',     'e',    'f',      'g',
    'h',     'i',     'j',     'k',      'l',     'm',    'n',      'o',
    'p',     'q',     'r',     's',      't',     'u',    'v',      'w',
    'x',     'y',     'z',     '{',      '|',     '}',    '~',      0,   

    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0xC1,    0xC0,     0xC2,    0xC4,   0xE1,     0xE0,
    0xE2,    0xE4,    0xC8,    0xC9,     0xCA,    0xCB,   0xE9,     0xE8,
    0xEA,    0xEB,    0xCD,    0xCC,     0xCE,    0xCF,   0xED,     0xEC,
    0xEE,    0xEF,    0xD3,    0xD2,     0xD4,    0xD6,   0xF3,     0xF2,
    0xF4,    0xF6,    0xDA,    0xD9,     0xDB,    0xDC,   0xFA,     0xF9,
    0xFB,    0xFC,    0xC7,    0xE7,     0xD1,    0xF1,   0x2CA,    0x2CB,
    0xA8,    0xBF,    0xA1,    0x3B1,    0x3B2,   0x3B3,  0x2206,   0x3B4,

    0x3F5,   '[',     0x3BB,   0x3BC,    0x3C0,   0x3C1,  0x3A3,    0x3C3,
    0x3C4,   0x3D5,   0x3A9,   0,        0,       0,      0,        0x25C2,
    0x25AA,  0,       0x2010,  0,        0,       0xB3,   '\n',     0,
    0,       0x3C7,   0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0,
    0,       0,       0,       0,        0,       0,      0,        0 };

/* Get the base encoding map for the given ID.  This is the mapping to
   be used for builtin characters within the ebook. */
static const fbunichar* get_base_encoding(int id)
{
  switch (id) {
  case ENCODING_TI83P_LARGE:
    return base_encoding_ti83p_large;

  case ENCODING_TI83P_SMALL:
    return base_encoding_ti83p_small;

  default:
    return NULL;
  }
}

/* Get the system encoding map for the given ID.  This is the mapping
   to be used for the application name.

   Currently there's only one system encoding since only one system
   (TI-83+) is supported. */
static const fbunichar* get_system_encoding(int id FBATTR_UNUSED)
{
  return base_encoding_ti83p_large;
}

/* Add a character to the list of characters used in the chunk.  This
   does not actually assign a byte value to the character, even if the
   base encoding is known; the assignment of byte values is done by
   fb_book_assign_characters(). */
void fb_chunk_register_character(fbbook* fb FBATTR_UNUSED, fbchunk* chunk,
				 fbunichar c)
{
  if (c != 0)
    fb_chartable_inc_value(&chunk->charcounts, c);
}

/* Assign byte values to the characters used in the book (if possible.)
   Also calculate the encoding_table and total character counts. */
int fb_book_assign_characters(fbbook* fb)
{
  const fbunichar* map;
  fbchartable charcounts;
  int i, j;
  fbsize k, count;

  for (i = 0; i < 256; i++)
    fb->charcount[i] = 0;

  /* Calculate character counts. */

  fb_chartable_clear(&charcounts);
  fb_chartable_inc_value(&charcounts, ' ');
  fb_chartable_inc_value(&charcounts, '\n');
  if (fb->hyphenate)
    fb_chartable_inc_value(&charcounts, 0xAD);

  fb_chartable_add_table(&charcounts, &fb->title->charcounts);
  fb_chartable_add_table(&charcounts, &fb->body->charcounts);
  for (k = 0; k < fb->nsections; k++)
    fb_chartable_add_table(&charcounts, &fb->sections[k].title->charcounts);

  if (charcounts.overflow)
    return fb_error(fb, ERR_TOO_MANY_CHARS,
                    "Too many different characters used in book"
                    " (maximum is 255)");

  map = get_base_encoding(fb->encoding_id);

  /* Clear existing encoding. */

  fb_chartable_clear(&fb->encoding_table);

  for (i = 0; i < 256; i++) {
    fb->encoding[i] = 0;
    fb->fontchars[i].set = 0;
    fb->fontchars[i].width = 0;
    fb->fontchars[i].value = 0;
    fb->fontchars[i].custom = 0;
  }

  /* Assign byte values to characters used in the book.  Try to use
     the base map assignments first. */

  if (map) {
    for (i = 1; i < 256; i++) {
      if (map[i] && !fb->encoding[i]) {
	count = fb_chartable_get_value(&charcounts, map[i]);
	if (count) {
	  fb->encoding[i] = map[i];
	  fb_chartable_set_value(&fb->encoding_table, map[i], i);
	  fb_chartable_set_value(&charcounts, map[i], 0);
	  fb->charcount[i] = count;
	}
      }
    }
  }

  /* Assign the remaining characters to unused byte values. */

  i = 1;
  for (j = 0; j < charcounts.nchars; j++) {
    if (!charcounts.chars[j] || !charcounts.values[j])
      continue;

    while (i < 256 && fb->encoding[i])
      i++;
    if (i > 255)
      return fb_error(fb, ERR_TOO_MANY_CHARS,
                      "Internal error (this is a bug)");

    fb->encoding[i] = charcounts.chars[j];
    fb_chartable_set_value(&fb->encoding_table, charcounts.chars[j], i);
    fb->charcount[i] = charcounts.values[j];
  }

  /* Find newline and space characters. */

  for (i = 0; i < 256; i++) {
    if (fb->encoding[i] == '\n')
      fb->newlinechar = i;
    else if (fb->encoding[i] == ' ')
      fb->spacechar = i;
  }

  if (fb->encoding_table.overflow)
    return fb_error(fb, ERR_TOO_MANY_CHARS,
		    "Internal error (this is a bug)");

  return 0;
}

/* Append UTF-8 text to an fbchunk.  s is a pointer to the text; n is
   the size in bytes.  Note that the book must also be specified when
   appending to the title, body, or section title, so that characters
   are registered properly. */
void fb_chunk_append_text_utf8(fbbook* fb, fbchunk* chunk,
			       const char* s, fbsize n)
{
  unsigned char *p, *q;
  unsigned char *end;
  fbunichar c;

  p = (unsigned char*) s;
  end = p + n;

  if (chunk->converted) {
    fbfree(chunk->converted);
    chunk->converted = NULL;
    chunk->length = 0;
    chunk->length_a = 0;
  }

  while (p < end) {
    q = p;
    c = read_utf8(&p);

    if (c == '\r') {
      if (*p == '\n')
        continue;
      else
        c = '\n';
    }

    if (c != 0x0000 && ((c & 0xfffe) != 0xfffe)) {
      if (fb && chunk != fb->appname)
        fb_chunk_register_character(fb, chunk, c);
      chunk->nutf8chars++;

      if ((fbsize) (chunk->utf8size + p - q) >= chunk->utf8size_a) {
        chunk->utf8size_a = (chunk->utf8size + p - q) * 2;
        chunk->utf8 = fbrenew(unsigned char, chunk->utf8,
                              chunk->utf8size_a);
      }
      while (q < p)
        chunk->utf8[chunk->utf8size++] = *(q++);
    }
  }
}

/* Insert text (in the TI encoding) into an fbchunk.  Note that this
   affects only the converted, not the UTF-8 text. */
void fb_chunk_insert_text_ti(fbbook* fb, fbchunk* chunk, fbsize offset,
			     const fbchar* s, fbsize n)
{
  fbchar* buf;
  fbsize i;

  if (!chunk->converted || offset > chunk->length || n < 1)
    return;

  if (chunk->length + n > chunk->length_a) {
    chunk->length_a = chunk->length + n + 100;
    buf = fbnew(fbchar, chunk->length_a);
    
    if (offset > 0)
      memcpy(buf, chunk->converted, offset * sizeof(fbchar));
    memcpy(buf + offset, s, n * sizeof(fbchar));
    if (offset < chunk->length)
      memcpy(buf + offset + n, chunk->converted + offset,
	     (chunk->length - offset) * sizeof(fbchar));

    fbfree(chunk->converted);
    chunk->converted = buf;
  }
  else {
    memmove(chunk->converted + offset + n,
	    chunk->converted + offset,
	    (chunk->length - offset) * sizeof(fbchar));
    memcpy(chunk->converted + offset, s, n * sizeof(fbchar));
  }
  chunk->length += n;

  /* Update line positions within the chunk */

  for (i = chunk->nlines; i > 0; i--) {
    if (chunk->lineend[i - 1] <= offset)
      break;
    chunk->lineend[i - 1] += n;
    if (chunk->linestart[i - 1] <= offset)
      break;
    chunk->linestart[i - 1] += n;
  }

  /* Update book positions, if we're inserting within the body */

  if (chunk == fb->body) {
    for (i = 0; i < fb->nsections; i++) {
      if (fb->sections[i].pos_conv > offset)
	fb->sections[i].pos_conv += n;
    }

    for (i = fb->nblocks; i > 0; i--) {
      if (fb->blockend[i - 1] <= offset)
	break;
      fb->blockend[i - 1] += n;
      if (fb->blockstart[i - 1] <= offset)
	break;
      fb->blockstart[i - 1] += n;
    }
  }
}

/* Convert a chunk into byte encoding.  If system = 0, use the book
   encoding; otherwise use the system encoding. */
void fb_chunk_convert_to_ti(fbbook* fb, fbchunk* chunk, int system)
{
  unsigned char* p;
  fbsize i;
  fbunichar c;
  int j;
  const fbunichar* map;
  const fbchartable* tbl;

  fbfree(chunk->converted);
  chunk->length_a = chunk->length = chunk->nutf8chars;
  chunk->converted = fbnew(fbchar, chunk->length);

  if (system) {
    map = get_system_encoding(fb->encoding_id);
    tbl = NULL;
  }
  else {
    map = fb->encoding;
    tbl = &fb->encoding_table;
  }

  p = chunk->utf8;
  for (i = 0; i < chunk->nutf8chars; i++) {
    if (*p < 128 && map[*p] == *p) {
      chunk->converted[i] = *p;
      p++;
    }
    else if (tbl) {
      c = read_utf8(&p);
      j = fb_chartable_get_value(tbl, c);
      if (j)
	chunk->converted[i] = j;
      else
	chunk->converted[i] = '?';
    }
    else {
      c = read_utf8(&p);
      for (j = 1; j < 256; j++) {
        if (c == map[j]) {
          chunk->converted[i] = j;
          break;
        }
      }
      if (j == 256)
        chunk->converted[i] = '?';
    }
  }

  if (chunk == fb->body)
    for (i = 0; i < fb->nsections; i++)
      fb->sections[i].pos_conv = fb->sections[i].pos_utf8;
}

/* Convert a chunk into UTF-8.  If system = 0, use the book encoding;
   otherwise use the system encoding. */
void fb_chunk_convert_to_utf8(fbbook* fb, fbchunk* chunk, int system)
{
  const fbunichar* map;
  fbunichar c;
  fbsize i, j;

  if (!chunk->converted)
    return;

  chunk->utf8size = 0;
  chunk->nutf8chars = 0;

  if (system)
    map = get_system_encoding(fb->encoding_id);
  else
    map = fb->encoding;

  for (i = 0; i < chunk->length; i++) {
    if (chunk == fb->body) {
      for (j = 0; j < fb->nsections; j++) {
	if (fb->sections[j].pos_conv == i)
	  fb->sections[j].pos_utf8 = chunk->nutf8chars;
      }
    }

    if (chunk->utf8size + 4 >= chunk->utf8size_a) {
      chunk->utf8size_a = (chunk->utf8size + 4) * 2;
      chunk->utf8 = fbrenew(unsigned char, chunk->utf8,
                            chunk->utf8size_a);
    }
    c = map[chunk->converted[i]];
    if (c && c != 0xAD) {
      chunk->utf8size += write_utf8(&chunk->utf8[chunk->utf8size], c);
      chunk->nutf8chars++;
    }
  }
}

/* Return the specified base character value. */
fbunichar fb_book_get_base_char(const fbbook* fb, fbchar c)
{
  const fbunichar* map = get_base_encoding(fb->encoding_id);

  if (map)
    return map[c];
  else
    return 0;
}

/* Return the specified system character value. */
fbunichar fb_book_get_system_char(const fbbook* fb, fbchar c)
{
  const fbunichar* map = get_system_encoding(fb->encoding_id);

  return map[c];
}
