/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include "fb.h"
#include "compress.h"

/**************** Decompressing ****************/

int fb_compress_read_data(fbbook* fb, fbchunk* chunk, fbcompressctx* comp,
			  const unsigned char* data, fbsize bufsize)
{
  fbcompresstree* t = comp->roots[0];
  unsigned char bitbuf = 0;
  unsigned long v;
  fbsize offs, length, n;
  int i;
  int l = 0;

  while (1) {
    if (l == 0) {
      if (bufsize == 0)
	return fb_error(fb, ERR_RUNAWAY_DATA,
			"Compressed data is missing EOF indicator");

      bitbuf = *data;
      data++;
      bufsize--;
      l = 8;
    }

    if (bitbuf & 0x80)
      t = t->right;
    else
      t = t->left;

    bitbuf <<= 1;
    l--;

    if (!t->left) {
      if (t->value != LZ_MARKER) {
        chunk->length++;
        if (chunk->length >= chunk->length_a) {
          chunk->length_a = chunk->length * 2;
          chunk->converted = fbrenew(fbchar, chunk->converted,
                                     chunk->length_a);
        }
        chunk->converted[chunk->length-1] = t->value;
      }
      else {
        v = 0;
        for (i = 0; i < LZ_PARAM_BITS; i++) {
          if (l == 0) {
	    if (bufsize == 0)
	      return fb_error(fb, ERR_RUNAWAY_DATA,
			      "Compressed data is missing EOF indicator");

            bitbuf = *data;
            data++;
	    bufsize--;
            l = 8;
          }
          v = (v << 1) | (bitbuf >> 7);
          bitbuf <<= 1;
          l--;
        }

        if (!(v & LZ_OFFSET_MASK))
          return 0;

        offs = LZ_OFFSET(v);
        length = LZ_LENGTH(v);

        if ((chunk->length + length) >= chunk->length_a) {
          chunk->length_a = (chunk->length + length) * 2;
          chunk->converted = fbrenew(fbchar, chunk->converted,
                                     chunk->length_a);
        }

        for (n = 0; n < length; n++) {
          chunk->converted[chunk->length]
            = chunk->converted[chunk->length - offs];
          chunk->length++;
        }
      }

      t = comp->roots[0];
    }
  }

  return 0;
}

static int compress_read_tree(fbbook* fb, fbcompresstree** t,
			      const unsigned char** data,
			      unsigned char* bitbuf, int* l,
			      fbsize* bufsize, int* maxnodes)
{
  int i;

  *t = fbnew(fbcompresstree, 1);
  (*t)->value = 0;
  (*t)->left = (*t)->right = NULL;
  (*t)->count = 0;

  if (*l == 0) {
    if (*bufsize == 0)
      return fb_error(fb, ERR_INVALID_HUFFMAN_TREE,
		      "Corrupted Huffman tree");

    *bitbuf = **data;
    (*data)++;
    (*bufsize)--;
    *l = 8;
  }

  if (*bitbuf & 0x80) {
    if (*maxnodes <= 0)
      return fb_error(fb, ERR_INVALID_HUFFMAN_TREE,
		      "Corrupted Huffman tree");
    (*maxnodes)--;

    *bitbuf <<= 1;
    (*l)--;

    if (compress_read_tree(fb, &(*t)->left, data, bitbuf, l,
			   bufsize, maxnodes))
      return ERR_INVALID_HUFFMAN_TREE;

    return compress_read_tree(fb, &(*t)->right, data, bitbuf, l,
			      bufsize, maxnodes);
  }
  else {
    *bitbuf <<= 1;
    (*l)--;

    for (i = 0; i < 8; i++) {
      if (*l == 0) {
	if (*bufsize == 0)
	  return fb_error(fb, ERR_INVALID_HUFFMAN_TREE,
			  "Corrupted Huffman tree");

        *bitbuf = **data;
        (*data)++;
	(*bufsize)--;
        *l = 8;
      }
      (*t)->value = ((*t)->value << 1) | (*bitbuf >> 7);

      *bitbuf <<= 1;
      (*l)--;
    }

    return 0;
  }
}

int fb_compress_read_global_data(fbbook* fb, fbcompressctx* comp, const unsigned char* data, fbsize bufsize)
{
  unsigned char bitbuf;
  int l;
  int maxnodes = 256;

  comp->lz_offset_bits = data[0];
  comp->lz_length_bits = data[1];
  comp->lz_length_bias = data[2];

  data += 3;
  bufsize -= 3;

  bitbuf = 0x80;
  l = 1;
  return compress_read_tree(fb, &comp->roots[0], &data, &bitbuf, &l, &bufsize, &maxnodes);
}
