/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifndef _FB_H
#define _FB_H

#ifdef __cplusplus
extern "C" {
#endif

/* Macros */

#if __GNUC__ >= 3
# define FBATTR_PURE __attribute__((__pure__))
# define FBATTR_UNUSED __attribute__((__unused__))
# define FBATTR_MALLOC __attribute__((__malloc__))
# define FBATTR_PRINTF(x,y) __attribute__((__format__(__printf__,x,y)))
#else
# define FBATTR_PURE
# define FBATTR_UNUSED
# define FBATTR_MALLOC
# define FBATTR_PRINTF(x,y)
#endif

#define FB_DEFAULT_BLOCK_SIZE 765
#define FB_MAX_STATIC_BLOCK_SIZE 765

#define FB_APP_SIGNATURE "FBebk1"
#define FB_APP_SIGNATURE_SIZE 6

#define FB_MAX_FONT_HEIGHT 16

#define FB_MAX_CHARS 255

#define FB_DEFAULT_SMALL_FONT "fb-small.fbf"
#define FB_DEFAULT_LARGE_FONT "fb-large.fbf"
#define FB_DEFAULT_SOURCE_FILE "fb-reader.asm"
#define FB_DEFAULT_KEY_FILE "0104.key"

/* Error codes */
enum {
  SUCCESS = 0,

  ERR_MISSING_FILE,		/* Missing common file */

  ERR_FILE_IO,			/* File I/O error */

  /* Parsing errors */
  ERR_XML_PARSER,		/* Invalid XML */
  ERR_UNKNOWN_XML_ROOT,		/* Unrecognized root element */
  ERR_INVALID_NUMERIC_VALUE,	/* Bad numeric value */

  /* Packing errors */
  ERR_NO_APPNAME,	    	/* No application name specified */
  ERR_TOO_MANY_CHARS,	        /* Too many characters */
  ERR_BITMAPS_NOT_FOUND,        /* One or more font bitmaps not found */
  ERR_METADATA_TOO_LARGE,	/* Metadata overflows first page */
  ERR_ASSEMBLER,		/* Error in assembly */
  ERR_HYPHEN_INIT,		/* Error loading hyphenation file */

  /* Unpacking errors */
  ERR_NO_HEADER,		/* No header found */
  ERR_UNKNOWN_HEADER,		/* Book uses unknown
				   addressing/compression mode */
  ERR_INVALID_HEADER,		/* Header is invalid */
  ERR_INVALID_HUFFMAN_TREE,	/* Huffman tree corrupted */
  ERR_INVALID_POINTER,		/* Invalid pointer in header/block
				   table */
  ERR_RUNAWAY_DATA		/* Missing EOF indicator */
};


/* Types */

typedef unsigned char fbchar;   /* encoded character -- 8 bits */
typedef unsigned int fbunichar; /* Unicode character, at least 32 bits */
typedef unsigned long fbsize;   /* size of files/segments, at least 32 bits */

enum {
  ENCODING_UNKNOWN = 0,
  ENCODING_TI83P_LARGE = 0x83,
  ENCODING_TI83P_SMALL = 0x84,
  ENCODING_TI73_LARGE = 0x73,
  ENCODING_TI73_SMALL = 0x74
};

/* Table for mapping characters to values */
typedef struct _fbchartable {
  int nchars;			/* number of characters in table */
  int overflow : 1;		/* = 1 if we have tried to add more
				   than 255 characters to the table */
  fbunichar chars[FB_MAX_CHARS]; /* characters */
  fbsize values[FB_MAX_CHARS];	 /* corresponding values */
} fbchartable;

typedef struct _fbchunk {
  /* Text.  When reading from an XML file only the UTF-8 will be set
     initially; when unpacking an app only the converted text will be
     set initially. */

  unsigned char* utf8;		/* UTF-8 (source) text */
  fbsize utf8size;		/* Size of UTF-8 text in bytes */
  fbsize utf8size_a;		/* Size of "utf8" buffer allocated */
  fbsize nutf8chars;		/* Number of characters in UTF-8 */
  fbchar* converted;		/* Byte encoded text.  Note that this
				   is not an exact copy of the UTF-8
				   text, as some modifications
				   (hyphenation, forced line breaks)
				   may be done after conversion. */
  fbsize length;		/* Number of characters (size of
				   byte-encoded text) */
  fbsize length_a;		/* Size of "converted" buffer
				   allocated */

  /* Line information.  Set by fb_chunk_split_lines(). */

  fbsize* linestart;		/* Array of line starting offsets
				   (each entry is the offset to to the
				   first character in the line) */
  fbsize* lineend;		/* Array of line ending offsets (each
				   entry is the offset to the last
				   character in the line plus 1; if
				   the line ends with a newline or
				   breaking space, this is the offset
				   to that character) */
  fbsize nlines;		/* Number of lines */

  /* Character information.  This must be updated by any function that
     modifies the text. */

  fbchartable charcounts;	/* Table giving the current number of
				   each character */
} fbchunk;

typedef struct _fbsection {
  fbchunk* title;		/* Title of section (if empty, will be
				   omitted from TOC) */
  fbsize pos_utf8;		/* Offset to start of section (source
				   characters) */
  fbsize pos_conv;		/* Offset (final converted text) */
} fbsection;

typedef struct _fbfontchar {
  fbunichar value;		/* Character value */
  unsigned char width;		/* Logical width of character */
  unsigned char miny;		/* Actual minimum y value */
  unsigned char maxy;		/* Actual maximum y value */
  unsigned char bitmap[FB_MAX_FONT_HEIGHT];
  unsigned set : 1;		/* 1 if bitmap is valid */
  unsigned custom : 1;		/* 1 if a custom character, which must
				   be included in the app */
  unsigned allowover : 1;	/* 1 if vertical overlaps are
				   permitted */
} fbfontchar;

typedef struct _fbbook {
  /* Book contents */

  fbchunk* title;		/* Title of the book (if empty, no
				   title will be shown) */
  fbchunk* appname;		/* Application name */
  fbchunk* body;		/* Body text */
  fbsection* sections;		/* Array of sections */
  fbsize nsections, nsections_a; /* Number of sections / number allocated */

  /* Block information -- set by fb_book_split_blocks() */

  fbsize* blockstart;		/* Array of block starting offsets */
  fbsize* blockend;		/* Array of block ending offsets */
  fbsize nblocks;		/* Number of blocks */
  fbsize blocksize_limit;	/* User-specified block size limit */
  fbsize blocksize_actual;	/* Actual maximum block size */
  fbsize packedbitlen;		/* Total length, in bits, of all
				   compressed text in the book */

  /* Font information -- set by fb_book_assign_font() */

  char* fontfile;		/* Font file name */
  int encoding_id;		/* Base encoding ID */
  fbunichar encoding[256];	/* Table for converting byte values to
				   unichars */
  fbchartable encoding_table;	/* Table for converting unichars to
				   byte values */
  fbsize charcount[256];        /* Character counts for assigned chars */
  fbchar newlinechar;           /* Newline character */
  fbchar spacechar;             /* Space character */

  int fontheight;               /* Font cell height */
  int rowheight;                /* Font row height */
  int tocheight;                /* Height of highlights in TOC */
  fbfontchar fontchars[256];    /* Array of font bitmaps */

  /* Other stuff */

  int hyphenate : 1;		/* 1 to enable hyphenation */
  char* hyphlang;		/* Language to use for hyphenation */

  unsigned char markid[4];      /* Bookmark file ID (hopefully unique;
				   used to ensure bookmark files won't
				   conflict) */

  char* searchdir;              /* Directory containing source file if
                                   known */

  char* asmfile;		/* Assembly source file */
  FILE* lstfile;		/* Assembly listing file */

  /* Function used to report errors.  Caller may set a different
     function here; the default function prints errors to stderr */
  void (*errfunc)(struct _fbbook* fb, int errcode, const char* message);

  /* Arbitrary data (used by the GUI) */
  void* userdata;
} fbbook;


typedef struct _fbcompresstree {
  fbchar value;                  /* Byte value */
  struct _fbcompresstree* left;  /* Left (zero bit) branch */
  struct _fbcompresstree* right; /* Right (one bit) branch */
  fbsize count;                  /* Number of occurrences of this
                                    byte/subtree */
} fbcompresstree;

typedef struct _fbcompresscode fbcompresscode;

typedef struct _fbcompressctx {
  fbcompresstree* roots[256];	/* Array of possible compression trees
				   (roots[0] is the final compression
				   trees; the others are used
				   temporarily during the Huffman
				   algorithm) */
  int nroots;			/* Number of possible compression
				   trees */

  fbcompresscode* codes;	/* Compression codes, constructed
				   based on the Huffman tree (usually
				   from the previous iteration) */
  int have_codes;		/* 1 if codes have been computed */

  int lz_offset_bits;	        /* Number of bits used for LZ offset */
  int lz_length_bits;           /* Number of bits used for LZ length */
  int lz_length_bias;           /* LZ length bias */

  fbchar* cdata;                /* Temporary buffer used during
				   optimization */

  fbsize current_bitlen;        /* Total bit length obtained using the
				   current parameters */
  fbsize last_bitlen;           /* Total bit length in the previous
				   iteration */
  int iterations;               /* Number of iterations so far */
} fbcompressctx;

typedef struct _fboutbuf {
  unsigned char *data;
  size_t length;
  size_t length_a;
} fboutbuf;

typedef struct _fbhyphctx {
  void *dict;			/* pointer to hnj HyphenDict */
  unsigned char map[256];	/* Map for book characters
				   (letters -> lowercase, space -> ' ',
				   punct -> '.', others -> 0) */
  int widthtab[256];		/* Copy of book width table */
  fbchar space;			/* Book space character */
  fbchar newline;		/* Book newline character */
  fbchar softhyphen;		/* Book soft hyphen character */
} fbhyphctx;


/* Memory allocation -- mem.c */

#define fbnew(ttt, nnn) ((ttt *) fbrealloc(0, (nnn) * sizeof(ttt)))
#define fbrenew(ttt, ppp, nnn) ((ttt *) fbrealloc((ppp), (nnn) * sizeof(ttt)))
#define fbfree(ppp) fbrealloc(ppp, 0)

void* fbrealloc(void* p, unsigned long sz) FBATTR_MALLOC;
fbbook* fb_book_new(void) FBATTR_MALLOC;
void fb_book_free(fbbook* fb);
fbchunk* fb_chunk_new(void) FBATTR_MALLOC;
void fb_chunk_free(fbchunk* chunk);
fbcompressctx* fb_compressctx_new(fbsize blocksize) FBATTR_MALLOC;
void fb_compressctx_free(fbcompressctx* comp);
void fb_chunk_clear(fbchunk* chunk);
fbsection* fb_book_append_section(fbbook* fb);
fboutbuf* fb_outbuf_new(void) FBATTR_MALLOC;
void fb_outbuf_free(fboutbuf* buf);
void fb_outbuf_append_byte(fboutbuf* buf, unsigned char d);
void fb_outbuf_prealloc(fboutbuf* buf, size_t size);


/* Unicode conversion -- unicode.c */

fbunichar read_utf8(unsigned char** s);
int write_utf8(unsigned char* s, fbunichar c);


/* Character-indexed lookup tables -- chartbl.c */

#define fb_chartable_clear(ttt) ((ttt)->overflow = (ttt)->nchars = 0)
fbsize fb_chartable_get_value(const fbchartable* tbl, fbunichar c)
  FBATTR_PURE;
void fb_chartable_set_value(fbchartable* tbl, fbunichar c, fbsize value);
void fb_chartable_inc_value(fbchartable* tbl, fbunichar c);
void fb_chartable_add_table(fbchartable* dest, const fbchartable* src);


/* Character set conversion -- convert.c */

void fb_chunk_register_character(fbbook* fb, fbchunk* chunk, fbunichar c);
void fb_chunk_register_unpacked(fbbook* fb, fbchunk* chunk);
int fb_book_assign_characters(fbbook* fb);
void fb_chunk_append_text_utf8(fbbook* fb, fbchunk* chunk, const char* s, fbsize n);
void fb_chunk_insert_text_ti(fbbook* fb, fbchunk* chunk, fbsize offset, const fbchar* s, fbsize n);
void fb_chunk_convert_to_ti(fbbook* fb, fbchunk* chunk, int system);
void fb_chunk_convert_to_utf8(fbbook* fb, fbchunk* chunk, int system);
fbunichar fb_book_get_base_char(const fbbook* fb, fbchar c) FBATTR_PURE;
fbunichar fb_book_get_system_char(const fbbook* fb, fbchar c) FBATTR_PURE;


/* Font handling -- font.c */

int fb_book_assign_font(fbbook* fb);


/* Text splitting -- split.c */

int fb_chunk_split_lines(fbbook* fb, fbchunk* chunk);
void fb_book_split_blocks(fbbook* fb);
void fb_book_get_block_addr(fbbook* fb, fbsize pos, fbsize* offs,
                            fbsize* block);
void fb_book_generate_mark_id(fbbook* fb);


/* Hyphenation -- hyph.c */

fbhyphctx* fb_hyphctx_new(fbbook* fb) FBATTR_MALLOC;
void fb_hyphctx_free(fbhyphctx* hyph);
fbsize fb_hyphenate(const fbhyphctx* hyph, const fbchunk* chunk,
		    fbsize word_start, int maxchars, int maxwidth);


/* Compression -- compress.c */

void fb_compress_begin(fbcompressctx* comp);
void fb_compress_try_data(fbcompressctx* comp, const fbchar* data, fbsize size);
void fb_compress_finish(fbcompressctx* comp);
int fb_compress_again(fbcompressctx* comp);

void fb_compress_write_global_data(fboutbuf* buffer, fbcompressctx* comp);
void fb_compress_write_data(fboutbuf* buffer, fbcompressctx* comp, const fbchar* data, fbsize size);

/* Decompression -- decomp.c */

int fb_compress_read_global_data(fbbook* fb, fbcompressctx* comp, const unsigned char* data, fbsize bufsize);
int fb_compress_read_data(fbbook* fb, fbchunk* chunk, fbcompressctx* comp, const unsigned char* data, fbsize bufsize);


/* XML input -- parse.c */

int fb_book_read_xml(fbbook* fb, FILE* f, const char* filename);


/* XML output -- output.c */

int fb_book_write_xml(fbbook* fb, FILE* f);


/* Packing -- build.c */

int fb_book_write_app(fboutbuf* buffer, fbbook* fb);


/* Unpacking -- unpack.c */

int fb_book_read_app(fbbook* fb, const unsigned char* data, fbsize len, fbsize startaddr);


/* Shared files -- share.c */

char* fb_find_common_file(fbbook* fb, const char* name);
FILE* fb_open_common_file(fbbook* fb, const char* name, const char* mode);
char* fb_get_dirname(const char* name);


/* Miscellaneous -- util.c */

char* fb_escape_filename(const char* fname, int len);
char* fb_unescape_filename(const char* ename, int len);


/* Error handling -- error.c */

int fb_error(fbbook* fb, int errcode, const char* fmt, ...) FBATTR_PRINTF(3, 4);
void fb_error_default(fbbook* fb, int errcode, const char* message);


#ifdef __cplusplus
}
#endif

#endif  /* _FB_H */
