/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fb.h"

/**************** Font handling ****************/

static void read_font_bitmaps(FILE* fontfile, fbbook* fb,
			      int height, int baseline)
{
  char buf[256];
  char *p;
  fbunichar v;
  fbfontchar *result;
  unsigned int x;
  int n, i;

  while (fgets(buf, sizeof(buf), fontfile)) {
    if ((p = strchr(buf, '#')))
      *p = 0;

    /* parse character value */
    v = strtol(buf, &p, 16);
    if (p == buf)
      continue;

    /* check if character is used in book */
    if (!(i = fb_chartable_get_value(&fb->encoding_table, v)))
      continue;

    /* set character value and 'set' flag */
    result = &fb->fontchars[i];
    result->value = v;
    result->set = 1;

    /* parse flags */
    result->custom = 1;
    result->allowover = 0;
    while (*p && (*p < '0' || *p > '9')) {
      if (*p == 'd')
        result->custom = 0;
      if (*p == 'o')
        result->allowover = 1;
      p++;
    }

    /* parse width */
    result->width = strtol(p, &p, 10);

    while (*p == ' ')
      p++;

    result->miny = result->maxy = baseline < 0 ? 0 : baseline;

    /* parse the actual bitmap */
    for (i = 0; i < height; i++) {
      if (0 < sscanf(p, "%02X%n", &x, &n)) {
        result->bitmap[i] = x;
        p += n;
        if (x) {
          result->maxy = i;
          if (i < result->miny)
            result->miny = i;
        }
      }
      else
        break;
    }

    for (; i < FB_MAX_FONT_HEIGHT; i++)
      result->bitmap[i] = 0;
  }
}

static int parse_header(char* line, char** valptr)
{
  char* p;

  if ((p = strchr(line, '#')))
    *p = 0;

  p = line;
  while (*p && *p != ':' && *p != '\n' && *p != '\r')
    p++;

  if (*p != ':' || p == line)
    return 0;

  *p = 0;
  p++;
  while (*p == ' ' || *p == '\t')
    p++;
  *valptr = p;
  while (*p && *p != '\n' && *p != '\r')
    p++;
  *p = 0;

  return 1;
}

int fb_book_assign_font(fbbook* fb)
{
  int i, j;
  int err = 0;
  fbsize charcount;
  fbsize ycount[FB_MAX_FONT_HEIGHT];
  fbsize oycount[FB_MAX_FONT_HEIGHT];
  int height, minrheight, baseline;
  double ccount;
  FILE* inf;
  char buf[256];
  char* value;

  if (!fb->fontfile || !fb->fontfile[0])
    inf = fb_open_common_file(fb, FB_DEFAULT_SMALL_FONT, "rb");
  else
    inf = fb_open_common_file(fb, fb->fontfile, "rb");

  if (!inf)
    return ERR_MISSING_FILE;

  fb->encoding_id = ENCODING_UNKNOWN;
  height = 7;
  baseline = 5;
  minrheight = 0;

  while (fgets(buf, sizeof(buf), inf)) {
    if (buf[0] == 0 || buf[0] == '\n' || buf[0] == '\r')
      break;
    else if (parse_header(buf, &value)) {
      if (!strcmp(buf, "Default-Encoding")) {
        if (!strcmp(value, "TI83P-Large"))
          fb->encoding_id = ENCODING_TI83P_LARGE;
        else if (!strcmp(value, "TI83P-Small"))
          fb->encoding_id = ENCODING_TI83P_SMALL;
      }
      else if (!strcmp(buf, "Cell-Height"))
        height = strtol(value, &value, 10);
      else if (!strcmp(buf, "Min-Row-Height"))
        minrheight = strtol(value, &value, 10);
      else if (!strcmp(buf, "Baseline"))
        baseline = strtol(value, &value, 10);
    }
  }

  if (height < 1)
    height = 1;
  if (height > FB_MAX_FONT_HEIGHT)
    height = FB_MAX_FONT_HEIGHT;

  if (minrheight < 1)
    minrheight = 1;
  if (minrheight > FB_MAX_FONT_HEIGHT)
    minrheight = FB_MAX_FONT_HEIGHT;

  if (baseline < 0)
    baseline = 0;
  if (baseline > height - 1)
    baseline = height - 1;

  for (i = 0; i < 256; i++) {
    fb->fontchars[i].set = 0;
    fb->fontchars[i].custom = 0;
  }

  /* This must be done after we read the font header, as it depends on
     the selected base encoding */
  err = fb_book_assign_characters(fb);

  /* Read bitmap data */
  read_font_bitmaps(inf, fb, height, baseline);

  charcount = 0;
  for (j = 0; j < FB_MAX_FONT_HEIGHT; j++)
    ycount[j] = oycount[j] = 0;

  /* Check that all needed bitmaps are defined, compute total
     character count, and compute Y histogram */
  for (i = 1; i < 256; i++) {
    if (fb->encoding[i] && fb->encoding[i] != '\n') {
      if (!fb->fontchars[i].set) {
        write_utf8((unsigned char*) buf, fb->encoding[i]);
        fb_error(fb, ERR_BITMAPS_NOT_FOUND,
                 "No font bitmap defined for character U+%04X (%s)",
                 fb->encoding[i], buf);
        err = ERR_BITMAPS_NOT_FOUND;
        continue;
      }

      if (fb->encoding[i] != fb_book_get_base_char(fb, i))
	fb->fontchars[i].custom = 1;

      charcount += fb->charcount[i];
      for (j = fb->fontchars[i].miny;
           j <= fb->fontchars[i].maxy; j++) {
        ycount[j] += fb->charcount[i];
        if (!fb->fontchars[i].allowover)
          oycount[j] += fb->charcount[i];
      }
    }
  }

  /* Choose row height */
  for (i = minrheight; i < height; i++) {
    ccount = 0;
    for (j = 0; j < (height - i + 1); j++)
      ccount += (double) oycount[j] * oycount[i + j - 1];

    if (ccount < (double) charcount * charcount * 0.01)
      break;
  }
  fb->rowheight = i;

  /* Choose font height */
  for (i = height - 1; i > baseline; i--)
    if (ycount[i])
      break;
  fb->fontheight = i + 1;

  fb->tocheight = height;

  /*
  printf("yct:");
  for (i = 0; i < height; i++)
  printf(" %6lu", ycount[i]);
  printf("\noct:");
  for (i = 0; i < height; i++)
  printf(" %6lu", oycount[i]);
  printf("\nrh: %d\nfh: %d\n", fb->rowheight, fb->fontheight);
  */

  fclose(inf);
  return err;
}
