/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include "fb.h"

#ifdef HAVE_LIBHNJ_HYPHEN_H
# include <libhnj/hyphen.h>
#else
# include <hyphen.h>
#endif

#if !defined(strcasecmp) && !defined(HAVE_STRCASECMP) && defined(HAVE_STRICMP)
# define strcasecmp stricmp
#endif

#include "hyphmap.h"

#define MAX_HYPHENATED_WORD 100

/**************** Hyphenation ****************/

static const struct enc_info {
  const char* name;
  const fbunichar* value;
  const unsigned char* map;
} encodings[] = 
  {{ "iso88591", iso8859_1_chars, iso8859_1_map },
   { "iso88592", iso8859_2_chars, iso8859_2_map },
   { "iso88597", iso8859_7_chars, iso8859_7_map },
   { "iso885913", iso8859_13_chars, iso8859_13_map },
   { "iso885915", iso8859_15_chars, iso8859_15_map },
   { "koi8r", koi8_r_chars, koi8_r_map },
   { "koi8u", koi8_u_chars, koi8_u_map },
   { "microsoftcp1251", cp1251_chars, cp1251_map },
   { 0, 0, 0 }};

static const struct enc_info* get_enc_info(const char* name)
{
  int i;
  char nbuf[256];

  i = 0;
  while (i < 255 && *name) {
    if (*name != '-' && *name != '_')
      nbuf[i++] = *name;
    name++;
  }
  nbuf[i] = 0;

  for (i = 0; encodings[i].name; i++)
    if (!strcasecmp(nbuf, encodings[i].name))
      return &encodings[i];

  return NULL;
}

static char map_char(const struct enc_info* enc, const char* lang, fbunichar c)
{
  unsigned int i;

  /* FIXME: This is broken.

     Whether apostrophe is considered a letter depends on the
     language, and more to the point, it depends on the hyphenation
     file.  (For instance, the French hyphenation file assumes ' is a
     letter and has rules to address whether words can be hyphenated
     near apostrophes.  The English hyphenation file assumes ' is not
     a letter, but rather a word terminator, so it has no such rules.)

     In TeX, a character is a letter for the purposes of hyphenation
     if it has a nonzero lccode.  Hyphenation files such as frhyph.tex
     and ithyph.tex specify \lccode`\'=`\'.  However, this information
     is not included in the libhyphen-format files, so there is no way
     to find it out using libhyphen. */

  if (c == 0x27 || c == 0x2019) {
    if (!strncmp(lang, "el_", 3)
	|| !strncmp(lang, "fr_", 3)
	|| !strncmp(lang, "it_", 3)
	|| !strncmp(lang, "uk_", 3))
      return '\'';
    else
      return '.';
  }

  if (c < 128)
    return ascii_map[c];

  if (enc) {
    for (i = 0; i < 128; i++)
      if (c == enc->value[i])
	return enc->map[i];
  }

  for (i = 0; i < N_EXTRA_PUNCTUATION_CHARS; i++)
    if (c == extra_punctuation[i])
      return '.';

  for (i = 0; i < N_EXTRA_SPACE_CHARS; i++)
    if (c == extra_space[i])
      return ' ';

  return 0;
}

fbhyphctx* fb_hyphctx_new(fbbook* fb)
{
  fbhyphctx* hyph;
  char* fname;
  char* fpath;
  HyphenDict* dict;
  const struct enc_info* enc;
  int i;

  if (!fb->hyphlang)
    return NULL;

  fname = fbnew(char, strlen(fb->hyphlang) + 10);
  sprintf(fname, "hyph_%s.dic", fb->hyphlang);
  fpath = fb_find_common_file(fb, fname);
  if (!fpath) {
    fbfree(fname);
    return NULL;
  }
  fbfree(fname);

  if (!(dict = hnj_hyphen_load(fpath))) {
    fb_error(fb, ERR_HYPHEN_INIT, "Cannot parse hyphenation file");
    fbfree(fpath);
    return NULL;
  }
  fbfree(fpath);

  enc = get_enc_info(dict->cset);
  /*
  if (!enc) {
    fb_error(fb, ERR_HYPHEN_INIT, "Unknown hyphenation file encoding %s",
	     dict->cset);
    return NULL;
  }
  */

  hyph = fbnew(fbhyphctx, 1);
  hyph->dict = dict;
  hyph->space = fb->spacechar;
  hyph->newline = fb->newlinechar;
  hyph->softhyphen = 0;

  for (i = 0; i < 256; i++) {
    hyph->widthtab[i] = fb->fontchars[i].width;

    if (fb->encoding[i] == 0xAD)
      hyph->softhyphen = i;

    hyph->map[i] = map_char(enc, fb->hyphlang, fb->encoding[i]);
  }

  return hyph;
}

void fb_hyphctx_free(fbhyphctx* hyph)
{
  if (hyph->dict)
    hnj_hyphen_free(hyph->dict);
  fbfree(hyph);
}

fbsize fb_hyphenate(const fbhyphctx* hyph, const fbchunk* chunk,
		    fbsize word_start, int maxchars, int maxwidth)
{
  HyphenDict* dict = hyph->dict;
  char buf[MAX_HYPHENATED_WORD];
  char result[MAX_HYPHENATED_WORD];
  int i, skip, len, x, best;
  char c;

  /* Skip initial punctuation */

  for (skip = 0; skip < MAX_HYPHENATED_WORD; skip++) {
    if (word_start + skip >= chunk->length
	|| hyph->map[chunk->converted[word_start + skip]] == ' ')
      return 0;
    if (hyph->map[chunk->converted[word_start + skip]] != '.')
      break;
  }

  if (skip == MAX_HYPHENATED_WORD)
    return 0;

  for (i = 0; i < skip; i++)
    maxwidth -= hyph->widthtab[chunk->converted[word_start + i]];
  maxchars -= skip;
  word_start += skip;

  if (maxwidth < 1 || maxchars < 1)
    return 0;

  /* Find length of word */

  for (len = 0; len < MAX_HYPHENATED_WORD; len++)
    if (word_start + len >= chunk->length
	|| hyph->map[chunk->converted[word_start + len]] == ' '
	|| hyph->map[chunk->converted[word_start + len]] == '.')
      break;

  if (len < dict->lhmin + dict->rhmin || len == MAX_HYPHENATED_WORD)
    return 0;

  /* Check if any hyphenation is possible */

  maxwidth -= hyph->widthtab[hyph->softhyphen];

  x = 0;
  for (i = 0; i < dict->lhmin; i++)
    x += hyph->widthtab[chunk->converted[word_start + i]];
  if (x > maxwidth)
    return 0;

  /* Convert into dictionary encoding */

  for (i = 0; i < len; i++) {
    c = hyph->map[chunk->converted[word_start + i]];
    if (!c)
      return 0;
    buf[i] = c;
  }

  /* Search for hyphenation points */

  if (hnj_hyphen_hyphenate(dict, buf, len, result))
    return 0;

  for (i = 0; i < dict->lhmin - 1; i++)
    result[i] = '0';
  for (i = len - dict->rhmin; i < len; i++)
    result[i] = '0';

  /* Find best result */

  x = 0;
  best = 0;
  for (i = 0; i < len; i++) {
    x += hyph->widthtab[chunk->converted[word_start + i]];
    if (x > maxwidth || i + 2 > maxchars)
      break;
    else if ((result[i] - '0') & 1)
      best = i + 1;
  }
  if (best)
    return word_start + best;
  else
    return 0;
}
