/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <miniasm.h>
#include "fb.h"

/**************** Packing ****************/

#define LSB(x) ((x) & 0xff)
#define MSB(x) (((x) >> 8) & 0xff)
#define MSBA(x) ((((x) >> 8) & 0x3f) | 0x40)
#define PAGE(x) (((x) >> 14) & 0xff)


int fb_book_write_app(fboutbuf* buffer, fbbook* fb)
{
  static const char signature[FB_APP_SIGNATURE_SIZE] = FB_APP_SIGNATURE;
  int compressed_titles = 1;

  fbcompressctx* comp;
  fbsize titlepos, blocktabpos, fonttabpos, maptabpos, sectabpos,
    imgtabpos, sectab, blocktab;
  const char* asmfile;
  asctx* as;
  const unsigned char* asbuf;
  long int asstart, asend;
  fbsize offs, block;
  fbsize i, j;
  fbunichar c;
  int k;
  unsigned char v;
  fbsize oldbuflen, pagestart;
  int e;

  if (fb->appname->nutf8chars == 0 && fb->appname->length == 0)
    return fb_error(fb, ERR_NO_APPNAME, "No application name specified");


  /******** Step 0: Assign characters and font bitmaps ********/

  if ((e = fb_book_assign_font(fb)))
    return e;


  /******** Step 1: Convert text into output encoding ********/

  fb_chunk_convert_to_ti(fb, fb->title, 0);
  fb_chunk_convert_to_ti(fb, fb->body, 0);
  for (i = 0; i < fb->nsections; i++)
    fb_chunk_convert_to_ti(fb, fb->sections[i].title, 0);

  fb_chunk_convert_to_ti(fb, fb->appname, 1);


  /******** Step 2: Format titles ********/

  /* Add newline characters to the book title to split it into lines
     for display.  Note that the title may also contain hard newlines
     if the creator wants it to be formatted in a certain way.

     FIXME: we should also truncate the title if it gets too long
     (probably more than 3 lines is too long.)
  */
  if (fb->title->length) {
    if ((e = fb_chunk_split_lines(fb, fb->title)))
      return e;
    for (i = 0; i < fb->title->nlines - 1; i++) {
      if (fb->title->converted[fb->title->lineend[i]] == fb->spacechar)
	fb->title->converted[fb->title->lineend[i]] = fb->newlinechar;
      else {
	fb_chunk_insert_text_ti(fb, fb->title, fb->title->lineend[i],
				&fb->newlinechar, 1);
	fb->title->linestart[i+1]++;
      }
    }
  }

  /* Truncate all section titles to a single line */
  for (i = 0; i < fb->nsections; i++) {
    if ((e = fb_chunk_split_lines(fb, fb->sections[i].title)))
      return e;
    if (fb->sections[i].title->nlines > 0)
      fb->sections[i].title->length = fb->sections[i].title->lineend[0];
  }


  /******** Step 3: Format body text ********/

  while (fb->body->length > 0
         && fb->body->converted[fb->body->length-1] == fb->newlinechar)
    fb->body->length--;
  if ((e = fb_chunk_split_lines(fb, fb->body)))
    return e;
  fb_book_split_blocks(fb);
  fb_book_generate_mark_id(fb);


  /******** Step 4: Analyze data for compression ********/

  fb->blocksize_actual = 0;
  for (i = 0; i < fb->nblocks; i++)
    if (fb->blockend[i] - fb->blockstart[i] > fb->blocksize_actual)
      fb->blocksize_actual = fb->blockend[i] - fb->blockstart[i];

  comp = fb_compressctx_new(fb->blocksize_limit);

  do {
    fb_compress_begin(comp);
    for (i = 0; i < fb->nblocks; i++)
      fb_compress_try_data(comp, &fb->body->converted[fb->blockstart[i]],
                           fb->blockend[i] - fb->blockstart[i]);
    if (compressed_titles)
      for (i = 0; i < fb->nsections; i++)
        fb_compress_try_data(comp, fb->sections[i].title->converted,
                             fb->sections[i].title->length);
    fb_compress_finish(comp);
  } while (fb_compress_again(comp));

  fb->packedbitlen = comp->current_bitlen;


  /******** Step 5: Assemble reader ********/

  as = miniasm_init((fb->lstfile ? MINIASM_WRITE_LISTING : 0), fb);

  miniasm_define(as, "_FBBUILD", 1);
  miniasm_define(as, "TI83P", 1);
  miniasm_define(as, "BLOCK_SIZE", fb->blocksize_actual);
  if (fb->blocksize_actual > FB_MAX_STATIC_BLOCK_SIZE)
    miniasm_define(as, "DYNAMIC_TEXT_BUFFERS", 1);
  miniasm_define(as, "USE_BOOKMARKS", 1);
  if (compressed_titles)
    miniasm_define(as, "COMPRESSED_TITLES", 1);
  if (fb->title->length == 0)
    miniasm_define(as, "NO_BOOK_TITLE", 1);

  for (i = 0; i < fb->nsections; i++) {
    if (fb->sections[i].title->length) {
      miniasm_define(as, "USE_TOC", 1);
      break;
    }
  }

  for (i = 0; i < 256; i++) {
    if (fb->fontchars[i].custom) {
      miniasm_define(as, "CUSTOM_FONT", 1);
      break;
    }
  }

  for (i = 0; i < 256; i++) {
    if (fb->encoding[i] == 0x2010)
      miniasm_define(as, "HYPHEN", i);
    if (fb->encoding[i] == 0xAD)
      miniasm_define(as, "SOFT_HYPHEN", i);
    if (fb->encoding[i] == 0x200B)
      miniasm_define(as, "ZERO_WIDTH_SPACE", i);
  }

  miniasm_define(as, "BASE_CHARSET", fb->encoding_id);
  if (fb->encoding_id == ENCODING_TI83P_LARGE)
    miniasm_define(as, "LARGE_FONT", 1);

  miniasm_define(as, "OFFSET_BITS", comp->lz_offset_bits);
  miniasm_define(as, "LENGTH_BITS", comp->lz_length_bits);
  miniasm_define(as, "LENGTH_BIAS", comp->lz_length_bias);
  miniasm_define(as, "SPACE", fb->spacechar);
  miniasm_define(as, "NEWLINE", fb->newlinechar);
  miniasm_define(as, "LINEDELTA", fb->rowheight);
  miniasm_define(as, "NUMLINES",
                 (64 + fb->rowheight - fb->fontheight) / fb->rowheight);
  miniasm_define(as, "FONT_HEIGHT", fb->fontheight);
  miniasm_define(as, "HIGHLIGHT_HEIGHT", fb->tocheight);
  miniasm_define(as, "MULTI_PAGE", 1);
  miniasm_define(as, "MULTI_PAGE_BLOCKS", 1);

  if (fb->asmfile)
    asmfile = fb->asmfile;
  else
    asmfile = FB_DEFAULT_SOURCE_FILE;

  if (miniasm_assemble(as, asmfile, &asbuf, &asstart, &asend,
		       NULL, NULL, NULL)) {
    miniasm_exit(as);
    fb_compressctx_free(comp);
    return ERR_ASSEMBLER;
  }
  for (i = 0; i < (fbsize) asend - asstart; i++)
    fb_outbuf_append_byte(buffer, asbuf[i]);
  miniasm_exit(as);

  for (i = 0; i < buffer->length; i++)
    if (!strncmp((char*) &buffer->data[i], signature,
                 FB_APP_SIGNATURE_SIZE))
      break;
  if (i == buffer->length)
    return fb_error(fb, ERR_NO_HEADER,
		    "Reader (%s) does not contain a known Flashbook header",
		    asmfile);

  buffer->length = i;


  /******** Step 6: Add metadata ********/

  for (i = 0; i < FB_APP_SIGNATURE_SIZE; i++)
    fb_outbuf_append_byte(buffer, signature[i]);

  fb_outbuf_append_byte(buffer, 0);	     /* Header option flags (reserved) */
  fb_outbuf_append_byte(buffer, 0);
  fb_outbuf_append_byte(buffer, fb->encoding_id); /* Base character set */
  fb_outbuf_append_byte(buffer, fb->fontheight);  /* Font height */

  fb_outbuf_append_byte(buffer, LSB(fb->blocksize_limit));
  fb_outbuf_append_byte(buffer, MSB(fb->blocksize_limit));
  titlepos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Title */
  fb_outbuf_append_byte(buffer, 0);
  fb_outbuf_append_byte(buffer, LSB(fb->nblocks));
  fb_outbuf_append_byte(buffer, MSB(fb->nblocks));
  blocktabpos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Block table */
  fb_outbuf_append_byte(buffer, 0);
  fonttabpos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Small font table */
  fb_outbuf_append_byte(buffer, 0);
  maptabpos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Mapping table */
  fb_outbuf_append_byte(buffer, 0);
  sectabpos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Section table */
  fb_outbuf_append_byte(buffer, 0);
  imgtabpos = buffer->length;
  fb_outbuf_append_byte(buffer, 0);       /* Image table (NYI) */
  fb_outbuf_append_byte(buffer, 0);


  /* Compression info */

  fb_compress_write_global_data(buffer, comp);


  /* Title */

  if (fb->title->length) {
    buffer->data[titlepos] = LSB(buffer->length);
    buffer->data[titlepos+1] = MSBA(buffer->length);
    for (i = 0; i < fb->title->length; i++)
      fb_outbuf_append_byte(buffer, fb->title->converted[i]);
    fb_outbuf_append_byte(buffer, fb->newlinechar);
    /* NOTE: next byte is a zero (start of section table) */
  }


  /* Mapping and font tables (empty for now) */

  buffer->data[maptabpos] = LSB(buffer->length);
  buffer->data[maptabpos+1] = MSBA(buffer->length);

  buffer->data[fonttabpos] = LSB(buffer->length);
  buffer->data[fonttabpos+1] = MSBA(buffer->length);


  /* Section table */

  buffer->data[sectabpos] = LSB(buffer->length);
  buffer->data[sectabpos+1] = MSBA(buffer->length);

  if (!fb->nsections || fb->sections[0].pos_conv > 0) {
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
  }

  sectab = buffer->length;

  for (i = 0; i < fb->nsections; i++) {
    fb_book_get_block_addr(fb, fb->sections[i].pos_conv, &offs, &block);

    fb_outbuf_append_byte(buffer, LSB(offs));
    fb_outbuf_append_byte(buffer, MSB(offs));
    fb_outbuf_append_byte(buffer, LSB(block));
    fb_outbuf_append_byte(buffer, MSB(block));
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
  }

  fb_outbuf_append_byte(buffer, -1);
  fb_outbuf_append_byte(buffer, -1);
  fb_outbuf_append_byte(buffer, -1);
  fb_outbuf_append_byte(buffer, -1);


  /* Section titles */

  for (i = 0; i < fb->nsections; i++) {
    if (fb->sections[i].title->length) {
      j = sectab + 6 * i;
      buffer->data[j + 4] = LSB(buffer->length);
      buffer->data[j + 5] = MSBA(buffer->length);
      if (compressed_titles) {
        fb_compress_write_data(buffer, comp,
                               fb->sections[i].title->converted,
                               fb->sections[i].title->length);
      }
      else {
        for (i = 0; i < fb->sections[i].title->length; i++)
          fb_outbuf_append_byte(buffer,
                            fb->sections[i].title->converted[i]);
        fb_outbuf_append_byte(buffer, 0);
      }
    }
  }


  /* Mapping table */

  k = 0;
  i = 0;
  while (i < 256) {
    j = i;
    while (j < 256
           && fb->encoding[j]
           && fb->encoding[j] != fb_book_get_base_char(fb, j))
      j++;

    if (i == j)
      i++;
    else {
      if (!k) {
        buffer->data[maptabpos] = LSB(buffer->length);
        buffer->data[maptabpos+1] = MSBA(buffer->length);
      }
      k = 1;
      fb_outbuf_append_byte(buffer, i);
      fb_outbuf_append_byte(buffer, j - i);
      while (i < j) {
        c = fb->encoding[i];
        if (c < 0x10000) {
          fb_outbuf_append_byte(buffer, c & 0xff);
          fb_outbuf_append_byte(buffer, (c >> 8) & 0xff);
        }
        else {
          c -= 0x10000;
          fb_outbuf_append_byte(buffer, (c >> 10) & 0xff);
          fb_outbuf_append_byte(buffer, 0xd8 | ((c >> 18) & 3));
          fb_outbuf_append_byte(buffer, c & 0xff);
          fb_outbuf_append_byte(buffer, 0xdc | ((c >> 8) & 3));
        }
        i++;
      }
    }
  }
  if (k)
    fb_outbuf_append_byte(buffer, 0);


  /* Font table */

  k = 0;
  i = 0;
  while (i < 256) {
    j = i;
    while (j < 256 && fb->fontchars[j].custom)
      j++;

    if (i == j)
      i++;
    else {
      if (!k) {
        buffer->data[fonttabpos] = LSB(buffer->length);
        buffer->data[fonttabpos+1] = MSBA(buffer->length);
      }
      k = 1;

      fb_outbuf_append_byte(buffer, i);
      fb_outbuf_append_byte(buffer, j - i);

      while (i < j) {
        fb_outbuf_append_byte(buffer, fb->fontchars[i].width);
        for (k = 0; k < fb->fontheight; k++)
          fb_outbuf_append_byte(buffer, fb->fontchars[i].bitmap[k]);
        i++;
      }
    }
  }
  if (k)
    fb_outbuf_append_byte(buffer, 0);


  /* Block table */

  buffer->data[blocktabpos] = LSB(buffer->length);
  buffer->data[blocktabpos+1] = MSBA(buffer->length);

  blocktab = buffer->length;

  for (i = 0; i < fb->nblocks; i++) {
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
    fb_outbuf_append_byte(buffer, 0);
  }

  if (buffer->length > 0x4000) {
    fb_error(fb, ERR_METADATA_TOO_LARGE,
             "Metadata overflows first app page");
    fb_compressctx_free(comp);
    return ERR_METADATA_TOO_LARGE;
  }


  /******** Step 7: Add compressed text ********/

  /* Actual block data */

  /* Pre-allocate the output buffer now, since we know almost exactly
     how big the compressed data will be. */
  fb_outbuf_prealloc(buffer, comp->current_bitlen * 8 + fb->nblocks);

  for (i = 0; i < fb->nblocks; i++) {
    v = MSB(buffer->length) & 0x3f;
    if (i + 1 < fb->nblocks) {
      /* Block splitting state stored in 2 MS bits of block logical
         address:

         0 = split at newline character
         1 = split at space character
         2 = split in midword (including after hyphen)
         3 = reserved
      */
      if (fb->blockstart[i + 1] == fb->blockend[i])
        v |= 0x80;
      else if (fb->body->converted[fb->blockend[i]] == fb->spacechar)
        v |= 0x40;
    }

    buffer->data[blocktab + 3 * i] = LSB(buffer->length);
    buffer->data[blocktab + 3 * i + 1] = v;
    buffer->data[blocktab + 3 * i + 2] = PAGE(buffer->length);

    oldbuflen = buffer->length;
    fb_compress_write_data(buffer, comp,
                           &fb->body->converted[fb->blockstart[i]],
                           fb->blockend[i] - fb->blockstart[i]);

    /* We need to avoid placing an FF byte at the start of any page,
       to work around a bug in the OS's defragmenting code. */
    pagestart = (buffer->length - 1) & ~0x3fff;
    if (pagestart >= oldbuflen && buffer->data[pagestart] == 0xff) {
      buffer->length = oldbuflen;
      fb_outbuf_append_byte(buffer, 0);
      i--;
    }
  }
  fb_compressctx_free(comp);

  return 0;
}
