/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include "fb.h"

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#else
# ifdef HAVE_IO_H
#  include <io.h>
# endif
#endif

#ifdef USE_GLIB_GSTDIO
# include <glib.h>
# include <glib/gstdio.h>
# define getenv g_getenv
# define fopen g_fopen
# define exists(name) g_file_test((name), G_FILE_TEST_EXISTS)
# define SEPARATOR G_DIR_SEPARATOR
# define IS_SEPARATOR G_IS_DIR_SEPARATOR
#else
# define exists(name) (!access((name), R_OK))
# define SEPARATOR '/'
# define IS_SEPARATOR(c) ((c) == '/')
#endif

#ifdef HAVE_LIBIGE_MAC_INTEGRATION
# include <ige-mac-integration.h>
#endif

/**************** Shared files ****************/

/* Check if given file exists and is readable */
static char* check_file(const char* prefix, const char* name)
{
  char* path;
  int n, m;

  n = strlen(prefix);
  m = strlen(name);
  path = fbnew(char, n + m + 2);

  strcpy(path, prefix);

  if (n > 0 && !IS_SEPARATOR(path[n-1])) {
    path[n] = SEPARATOR;
    n++;
  }

  strcpy(path + n, name);

  if (exists(path))
    return path;

  fbfree(path);
  return NULL;
}

/* Find a common file */
static char* find_common_file(fbbook* fb, const char* name)
{
  char* path;
  const char* e;
  char* instdir FBATTR_UNUSED;

  if (fb && fb->searchdir && (path = check_file(fb->searchdir, name)))
    return path;
  else if ((path = check_file("", name)))
    return path;

  if ((e = getenv("FLASHBOOK_DATA_DIR"))
      && (path = check_file(e, name)))
    return path;

#ifdef G_OS_WIN32
  instdir = g_win32_get_package_installation_directory(NULL, NULL);
  if (instdir && (path = check_file(instdir, name))) {
    g_free(instdir);
    return path;
  }
  g_free(instdir);
#endif

#ifdef HAVE_LIBIGE_MAC_INTEGRATION
  e = ige_mac_bundle_get_path(ige_mac_bundle_get_default());
  if (e) {
    instdir = g_build_filename(e, "Contents", "Resources", NULL);
    if ((path = check_file(instdir, name))) {
      g_free(instdir);
      return path;
    }
    g_free(instdir);
  }
#endif

#ifdef SHARE_DIR
  if ((path = check_file(SHARE_DIR, name)))
    return path;
#endif

#ifdef UNINSTALLED_SHARE_DIR
  if ((path = check_file(UNINSTALLED_SHARE_DIR, name)))
    return path;
#endif

  if (fb)
    fb_error(fb, ERR_MISSING_FILE, "Cannot find %s", name);
  return NULL;
}

/* Find a common file */
char* fb_find_common_file(fbbook* fb, const char* name)
{
#ifdef G_OS_WIN32
  char* path = find_common_file(fb, name);
  char* locpath;

  if (!path)
    return NULL;

  locpath = g_win32_locale_filename_from_utf8(path);
  if (locpath) {
    if (strlen(locpath) > strlen(path)) {
      fbfree(path);
      path = fbnew(char, strlen(locpath) + 1);
    }
    strcpy(path, locpath);
    g_free(locpath);
  }
  return path;

#else
  return find_common_file(fb, name);
#endif
}

/* Find and open a common file */
FILE* fb_open_common_file(fbbook* fb, const char* name, const char* mode)
{
  FILE* f;
  char* path = find_common_file(fb, name);

  if (!path)
    return NULL;

  f = fopen(path, mode);
  if (!f && fb)
    fb_error(fb, ERR_MISSING_FILE, "Unable to read %s", name);
  fbfree(path);

  return f;
}

/* Get directory component of a file name */
char* fb_get_dirname(const char* name)
{
  int i;
  char* s;

  for (i = strlen(name) - 1; i >= 0; i--) {
    if (IS_SEPARATOR(name[i])) {
      s = fbnew(char, i + 1);
      strncpy(s, name, i);
      s[i] = 0;
      return s;
    }
  }

  return NULL;
}
