/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include "fb.h"

/**************** Unicode format conversion ****************/

/* Print out an error about invalid UTF-8 (this is mostly for
   debugging.) */
static void utf8error(const char* desc, const unsigned char* s)
{
  int i;

  while (*s >= 0x80 || (*s && *s < 0x20))
    s++;

  fprintf(stderr, "Invalid UTF-8 (%s)", desc);
  if (*s) {
    fprintf(stderr, " near '");
    for (i = 0; i < 32 && s[i] >= 0x20 && s[i] <= 0x7f; i++)
      fputc(s[i], stderr);
    fprintf(stderr, "...'\n");
  }
  else
    fprintf(stderr, " near end of string\n");
}

/* Read a UTF-8 character and advance the input pointer.  Return FFFF
   (which is an invalid Unicode character) if the input is not
   well-formed UTF-8. */
fbunichar read_utf8(unsigned char** s)
{
  int n;
  fbunichar v, min;

  if (**s < 0x80) {
    (*s)++;
    return (*s)[-1];
  }
  if (**s < 0xc0) {
    (*s)++;
    utf8error("no prefix", *s);
    return 0xffff;
  }

  if (**s < 0xe0) {
    n = 1;
    min = 0x80;
    v = (**s & 0x1f);
  }
  else if (**s < 0xf0) {
    n = 2;
    min = 0x800;
    v = (**s & 0x0f);
  }
  else if (**s < 0xf8) {
    n = 3;
    min = 0x10000;
    v = (**s & 0x07);
  }
  else if (**s < 0xfc) {
    n = 4;
    min = 0x200000;
    v = (**s & 0x03);
  }
  else if (**s < 0xfe) {
    n = 5;
    min = 0x4000000;
    v = (**s & 0x01);
  }
  else {
    (*s)++;
    utf8error("bad prefix", *s);
    return 0xffff;
  }

  while (n > 0) {
    (*s)++;
    if (**s < 0x80) {
      utf8error("not enough bits for prefix", *s);
      return 0xffff;
    }
    else if (**s >= 0xc0)
      return 0xffff;

    v <<= 6;
    v |= (**s & 0x3f);
    n--;
  }
  (*s)++;

  if (v < min) {
    utf8error("overlong form", *s);
    return 0xffff;
  }
  else
    return v;
}

/* Write a character as UTF-8; return the length of the output
   string. */
int write_utf8(unsigned char* s, fbunichar c)
{
  if (c > 0x1fffff)
    c = 0xfffd;

  if (c < 0x80) {
    s[0] = c;
    s[1] = 0;
    return 1;
  }
  else if (c < 0x800) {
    s[0] = (0xc0 | (c >> 6));
    s[1] = (0x80 | (c & 0x3f));
    s[2] = 0;
    return 2;
  }
  else if (c < 0x10000) {
    s[0] = (0xe0 | (c >> 12));
    s[1] = (0x80 | ((c >> 6) & 0x3f));
    s[2] = (0x80 | (c & 0x3f));
    s[3] = 0;
    return 3;
  }
  else {
    s[0] = (0xf0 | (c >> 18));
    s[1] = (0x80 | ((c >> 12) & 0x3f));
    s[2] = (0x80 | ((c >> 6) & 0x3f));
    s[3] = (0x80 | (c & 0x3f));
    s[4] = 0;
    return 4;
  }
}
