/*
 * libflashbook - Library for creating eBooks for the TI-83/84 Plus
 * Copyright (c) 2008 Benjamin Moody
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 3 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <rabbitsign.h>
#include "fb.h"

/**************** Unpacking ****************/

static void register_unpacked(fbbook* fb, fbchunk* chunk)
{
  fbsize i;
  for (i = 0; i < chunk->length; i++)
    fb_chunk_register_character(fb, chunk, chunk->converted[i]);
}


int fb_book_read_app(fbbook* fb, const unsigned char* data,
                     fbsize len, fbsize startaddr)
{
  static const char signature[FB_APP_SIGNATURE_SIZE] = FB_APP_SIGNATURE;
  fbsize i, n, pos;
  fbsize hdrpos, titlepos, nblocks, blocktabpos, fonttabpos, maptabpos,
    sectabpos, imgtabpos;
  fbsize maxblocksize;
  fbcompressctx* comp;
  fbsection* sect;
  fbsize blocknum;
  fbchar c, nc;
  unsigned long namepos, namelength;
  int e;

  /* Read app name */

  if (data[0] == 0x80 && data[1] == 0x0f
      && !rs_find_app_field(0x8040, data + 6, 122, NULL, &namepos, &namelength)) {
    fb->appname->length = fb->appname->length_a = namelength;
    fb->appname->converted = fbrenew(fbchar, fb->appname->converted,
				     namelength);
    for (i = 0; i < namelength; i++)
      fb->appname->converted[i] = data[6 + namepos + i];
  }

  /* Read FB header */

  for (hdrpos = 0; hdrpos < len - FB_APP_SIGNATURE_SIZE; hdrpos++)
    if (!strncmp((const char*) data + hdrpos, signature,
                 FB_APP_SIGNATURE_SIZE))
      break;

  if (hdrpos == len - FB_APP_SIGNATURE_SIZE)
    return fb_error(fb, ERR_NO_HEADER,
                    "No compatible Flashbook data found");

  hdrpos += FB_APP_SIGNATURE_SIZE;

  if (hdrpos > len - 24)
    return fb_error(fb, ERR_INVALID_HEADER,
                    "App has a corrupted Flashbook header"
                    " (or missing data.)");

  if ((data[hdrpos + 2] != ENCODING_UNKNOWN
       && data[hdrpos + 2] != ENCODING_TI83P_LARGE
       && data[hdrpos + 2] != ENCODING_TI83P_SMALL))
    return fb_error(fb, ERR_UNKNOWN_HEADER,
                    "App uses unknown character set %x",
                    data[hdrpos + 2]);

  fb->encoding_id = data[hdrpos + 2];
  fb->fontheight = data[hdrpos + 3];

  maxblocksize =  data[hdrpos + 4] | (data[hdrpos + 5] << 8);
  titlepos     =  data[hdrpos + 6] | (data[hdrpos + 7] << 8);
  nblocks      =  data[hdrpos + 8] | (data[hdrpos + 9] << 8);
  blocktabpos  = data[hdrpos + 10] | (data[hdrpos + 11] << 8);
  fonttabpos   = data[hdrpos + 12] | (data[hdrpos + 13] << 8);
  maptabpos    = data[hdrpos + 14] | (data[hdrpos + 15] << 8);
  sectabpos    = data[hdrpos + 16] | (data[hdrpos + 17] << 8);
  imgtabpos    = data[hdrpos + 18] | (data[hdrpos + 19] << 8);

  fb->blocksize_limit = maxblocksize;

  if (titlepos) titlepos -= startaddr;
  blocktabpos -= startaddr;
  fonttabpos -= startaddr;
  maptabpos -= startaddr;
  sectabpos -= startaddr;
  imgtabpos -= startaddr;

  if (titlepos >= len || blocktabpos + 3 * nblocks > len
      || fonttabpos >= len || maptabpos >= len || sectabpos >= len)
    return fb_error(fb, ERR_INVALID_HEADER,
                    "App has a corrupted Flashbook header"
                    " (or missing data.)");

  /* Read mapping table */

  for (i = 1; i < 256; i++)
    fb->encoding[i] = fb_book_get_base_char(fb, i);

  i = maptabpos;
  while (i < len - 1 && data[i]) {
    c = data[i];
    nc = data[i + 1];
    i += 2;
    while (nc > 0) {
      if (data[i + 1] >= 0xd8 && data[i + 1] <= 0xdb) {
        fb->encoding[c] = (((data[i] << 10)
                            | ((data[i + 1] & 3) << 18)
                            | data[i + 2]
                            | ((data[i + 3] & 3) << 8))
                           + 0x10000);
        i += 4;
      }
      else {
        fb->encoding[c] = data[i] | (data[i + 1] << 8);
        i += 2;
      }
      c++;
      nc--;
    }
  }

  for (i = 1; i < 256; i++) {
    if (fb->encoding[i] == '\n')
      fb->newlinechar = i;
    else if (fb->encoding[i] == ' ')
      fb->spacechar = i;
  }

  /* Read title */

  if (titlepos) {
    for (n = 0; titlepos + n < len; n++)
      if (!data[titlepos + n])
	break;

    if (titlepos + n >= len)
      return fb_error(fb, ERR_INVALID_HEADER, "App has an invalid title");

    fb->title->length = fb->title->length_a = n;
    fb->title->converted = fbrenew(fbchar, fb->title->converted, n);
    for (i = 0; i < n; i++)
      fb->title->converted[i] = data[titlepos + i];
    if (fb->title->converted[n - 1] == fb->newlinechar)
      fb->title->length--;
    for (i = 0; i < fb->title->length; i++)
      if (fb->title->converted[i] == fb->newlinechar)
        fb->title->converted[i] = fb->spacechar;
    register_unpacked(fb, fb->title);
  }

  /* Initialize decompressor */

  comp = fb_compressctx_new(maxblocksize);
  pos = hdrpos + 20;
  if ((e = fb_compress_read_global_data(fb, comp, data + pos, len - pos)))
    return e;

  /* Read body text */

  for (i = 0; i < nblocks; i++) {
    pos = (data[blocktabpos + 3 * i]
           | ((data[blocktabpos + 3 * i + 1] & 0x3f) << 8)
           | (data[blocktabpos + 3 * i + 2] << 14));

    if (pos >= len) {
      return fb_error(fb, ERR_INVALID_POINTER,
		      "App has a corrupted block table"
		      " (or missing data.)");
    }

    fb->nblocks++;
    fb->blockstart = fbrenew(fbsize, fb->blockstart, fb->nblocks);
    fb->blockend = fbrenew(fbsize, fb->blockend, fb->nblocks);
    fb->blockstart[fb->nblocks-1] = fb->body->length;
    if ((e = fb_compress_read_data(fb, fb->body, comp,
				   data + pos, len - pos)))
      return e;
    fb->blockend[fb->nblocks-1] = fb->body->length;

    if ((fb->body->length + 1) >= fb->body->length_a) {
      fb->body->length_a = (fb->body->length + 1) * 2;
      fb->body->converted = fbrenew(fbchar, fb->body->converted,
                                    fb->body->length_a);
    }

    switch (data[blocktabpos + 3 * i + 1] & 0xc0) {
    case 0x00:
      fb->body->converted[fb->body->length] = fb->newlinechar;
      fb->body->length++;
      break;

    case 0x40:
      fb->body->converted[fb->body->length] = fb->spacechar;
      fb->body->length++;
      break;

    case 0x80:
    case 0xc0:
      break;
    }
  }

  register_unpacked(fb, fb->body);

  /* Read section table and titles */

  for (i = sectabpos;
       i + 5 < len && (data[i] & data[i+1] & data[i+2] & data[i+3]) != 0xff;
       i += 6) {
    /* omit the null section (address = 0 and no title) */
    if (data[i] || data[i+1] || data[i+2]
        || data[i+3] || data[i+4] || data[i+5]) {

      sect = fb_book_append_section(fb);

      blocknum = data[i+2] + (data[i+3] << 8);
      if (blocknum >= nblocks)
	return fb_error(fb, ERR_INVALID_POINTER,
			"App has a corrupted section table"
			" (or missing data.)");

      sect->pos_conv = (fb->blockstart[blocknum]
			+ data[i] + (data[i+1] << 8));
      if (sect->pos_conv > fb->blockend[blocknum])
	return fb_error(fb, ERR_INVALID_POINTER,
			"App has a corrupted section table"
			" (or missing data.)");

      if (data[i+4] || data[i+5]) {
        pos = data[i+4] + (data[i+5] << 8) - startaddr;
	if (pos >= len)
	  return fb_error(fb, ERR_INVALID_POINTER,
			  "App has a corrupted section table"
			  " (or missing data.)");

        if ((e = fb_compress_read_data(fb, sect->title, comp,
				       data + pos, len - pos)))
	  return e;
      }

      register_unpacked(fb, sect->title);
    }
  }

  fb_compressctx_free(comp);

  return 0;
}
