/*
 * Miniasm
 *
 * Copyright (c) 2007 Benjamin Moody <floppusmaximus@users.sf.net>
 *
 * This program is free software: you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation, either version 2 of the
 * License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include "miniasm.h"

#ifdef __GNUC__
# define UNUSED          __attribute__((unused))
#else
# define UNUSED
#endif

/**************** Memory allocation ****************/

#define xmalloc(sss) xrealloc(0, sss)
#define xfree(ppp) xrealloc(ppp, 0)

static void* xrealloc(void* p, unsigned long sz)
{
  if (sz) {
    if (p)
      p = realloc(p, sz);
    else
      p = malloc(sz);
    if (!p) {
      fprintf(stderr, "Out of memory (need %lu bytes)\n", sz);
      abort();
    }
  }
  else {
    if (p)
      free(p);
    p = 0;
  }
  return p;
}


/**************** I/O Callback Functions ****************/

struct _asfile {
  FILE* fp;
};

asfile* miniasm_open(const char* filename, void* data UNUSED)
{
  asfile* f = (asfile*) xmalloc(sizeof(asfile));

  if (strcmp(filename, "-")) {
    if (!(f->fp = fopen(filename, "rt"))) {
      xfree(f);
      return NULL;
    }
  }
  else
    f->fp = stdin;

  return f;
}

void miniasm_close(asfile* filep, void* data UNUSED)
{
  if (filep->fp != stdin)
    fclose(filep->fp);
  xfree(filep);
}

char* miniasm_gets(asfile* filep, void* data UNUSED)
{
  int len = 0, len_a = 256;
  int c;
  char* str;

  if (feof(filep->fp) || ferror(filep->fp)) {
    return NULL;
  }

  str = (char*) xmalloc(len_a * sizeof(char));

  do {
    c = fgetc(filep->fp);

    if (c != EOF) {
      if (len >= len_a) {
	len_a *= 2;
	str = (char*) xrealloc(str, len_a * sizeof(char));
      }
      str[len] = c;
      len++;
    }
    else
      break;
  } while (c != '\n' && c != '\r');

  if (c == '\r') {
    c = fgetc(filep->fp);
    if (c == '\n') {
      if (len >= len_a) {
	len_a *= 2;
	str = (char*) xrealloc(str, len_a * sizeof(char));
      }
      str[len] = c;
      len++;
    }
    else
      ungetc(c, filep->fp);
  }

  if (len) {
    if (len >= len_a) {
      len_a = len + 1;
      str = (char*) xrealloc(str, len_a * sizeof(char));
    }
    str[len] = 0;
    return str;
  }

  xfree(str);
  return NULL;
}

void miniasm_gets_free(char* str)
{
  xfree(str);
}

int miniasm_rewind(asfile* filep, void* data UNUSED)
{
  return fseek(filep->fp, 0L, SEEK_SET);
}

int miniasm_get_bin_size(const char* filename, long int* lengthp,
			   void* data UNUSED)
{
  struct stat st;

  if (stat(filename, &st))
    return 1;

  *lengthp = st.st_size;
  return 0;
}

int miniasm_read_bin(const char* filename, unsigned char* buf, long int n,
		       void* data UNUSED)
{
  FILE* f;

  f = fopen(filename, "rb");
  if (!f)
    return 1;

  if (n == (long) fread(buf, 1, n, f)) {
    fclose(f);
    return 0;
  }
  else {
    fclose(f);
    return 1;
  }
}

void miniasm_write_message(const char* string, void* data UNUSED)
{
  fprintf(stderr, "%s\n", string);
}

void miniasm_write_listing(int linenum, const char* text,
			   long int lineaddr, long int linecount,
			   const unsigned char* linebytes, void* data)
{
  FILE* lstfile = (FILE*) data;
  int i;
  const char* p;

  if (!lstfile)
    return;

  fprintf(lstfile, "%5d %08X", linenum, (unsigned int) lineaddr);

  for (i = 0; i < linecount && i < 5; i++)
    fprintf(lstfile, " %02X", (unsigned int) linebytes[i]);
  for (; i < 5; i++)
    fputs("   ", lstfile);
  fputs("   ", lstfile);

  for (p = text; *p; p++)
    if (*p != '\n' && *p != '\r')
      fputc(*p, lstfile);

  for (; i < linecount; i++) {
    if (!(i % 5))
      fputs("\n              ", lstfile);
    fprintf(lstfile, " %02X", (unsigned int) linebytes[i]);
  }

  fputc('\n', lstfile);
}


/**************** Standalone Assembler ****************/

int main(int argc, char** argv)
{
  char* infilename = NULL;
  char* outfilename = NULL;
  char* lstfilename = NULL;
  FILE* outfile = NULL;
  FILE* lstfile = NULL;
  asctx* ctx;
  const unsigned char* buf;
  long int start;
  long int end;
  int npass;
  int nerrs;
  int nwarn;

  int i;
  char *p, *q;
  long int v;

  for (i = 1; i < argc; i++) {
    if (argv[i][0] != '-' || argv[i][1] == 0)
      infilename = argv[i];
    else {
      if (argv[i][1] == 'o') {
	if (argv[i][2])
	  outfilename = &argv[i][2];
	else {
	  outfilename = argv[++i];
	}
      }
      else if (argv[i][1] == 'l') {
	if (argv[i][2])
	  lstfilename = &argv[i][2];
	else
	  lstfilename = argv[++i];
      }
      else if (argv[i][1] == 'D') {
	if (!argv[i][2])
	  i++;
      }
      else {
	fprintf(stderr, "%s: unknown option %s\n", argv[0], argv[i]);
	return 1;
      }
    }
  }

  if (!infilename) {
    fprintf(stderr,
	    "Usage: %s asmfile [-o binfile] [-l lstfile]\n"
	    "         [-D symbol[=value]] ...\n",
	    argv[0]);
    return 1;
  }

  if (lstfilename) {
    if (strcmp(lstfilename, "-"))
      lstfile = fopen(lstfilename, "wt");
    else
      lstfile = stdout;
    if (!lstfile) {
      perror(lstfilename);
      return 1;
    }

    fprintf(lstfile, "Miniasm standalone assembler (pre-release)\n");
    fprintf(lstfile, "Source File: %s\n\n", infilename);

    ctx = miniasm_init(MINIASM_WRITE_LISTING, lstfile);
  }
  else {
    ctx = miniasm_init(0, 0);
  }

  for (i = 1; i < argc; i++) {
    if (argv[i][0] == '-') {
      if (argv[i][1] == 'D') {
	if (argv[i][2])
	  p = &argv[i][2];
	else
	  p = argv[++i];

	v = 1;
	q = strchr(p, '=');
	if (q) {
	  *q = 0;
	  sscanf(q + 1, "%li", &v);
	}
	miniasm_define(ctx, p, v);
      }
      else if (argv[i][1] && !argv[i][2]) i++;
    }
  }

  miniasm_assemble(ctx, infilename, &buf, &start, &end, &npass, &nerrs, &nwarn);

  if (lstfile && lstfile != stdout)
    fclose(lstfile);

  if (outfilename) {
    if (strcmp(outfilename, "-"))
      outfile = fopen(outfilename, "wb");
    else
      outfile = stdout;
    if (!outfile) {
      perror(outfilename);
      miniasm_exit(ctx);
      return 1;
    }

    fwrite(buf, 1, end - start, outfile);
    fclose(outfile);
  }

  fprintf(stderr, "Assembled in %d passes (%d errors, %d warnings)\n",
	  npass, nerrs, nwarn);

  miniasm_exit(ctx);
  return nerrs ? 1 : 0;
}
